/* hp3000_adcc.c: HP 3000 31264A Asynchronous Data Communications Channel simulator

   Copyright (c) 2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   ADCC         31264A Asynchronous Data Communications Channel Main
                31265A Asynchronous Data Communications Channel Extender

   17-Jul-24    JDB     RESET now disconnects dataset-connected sessions
                        Revised tracing messages for better clarity
   16-Apr-24    JDB     First release
   01-Jan-24    JDB     Allow output to an unconnected port if logging is enabled
   05-Aug-23    JDB     ADCCDIAG 1.32 passes
   29-Jul-23    JDB     ADCCDIAG 1.31 passes
   12-Jul-23    JDB     Created

   References:
     - HP 3000 Series 40/44 Computer Systems Reference/Training Manual
         (30090-90001, July 1981) [pp. 8-1 to 8-14]
     - HP 3000 Series 39/40/42/44/48 Computer Systems Engineering Diagrams Set
         (30090-90034, October 1984) [pp. 3-36 to 3-49]
     - "A Computer Input/Output System Based on the HP Interface Bus"
         (HP Journal, July 1979) [pp. 9 to 13]
     - ADCC - Main and ADCC - Extended Diagnostic Manual
         (30070-90037, September 1978)


   The 31264A Asynchronous Data Communications Channel (ADCC) is a 4-port
   terminal multiplexer used with IMB systems.  It supports 14 programmable baud
   rates from 50 to 9600 bits per second.  Character sizes are also programmable
   from 5 to 8 bits in length, excluding the start and one or two stop bits.
   Parity is optional and may be configured for even or odd parity generation
   and checking.  Each port can be independently configured, including for
   separate send and receive rates, and the number of ports may be expanded to
   eight with the addition of an ADCC extender card.

   Ports may be directly connected to terminals or may be connected through data
   sets.  The ADCC main and extender cards each connect via a cable to a
   four-place RS-232 connector panel, populated by female DB-25 connectors wired
   as DCE.

   Each port is controlled by an on-board UART, which contains a two-character
   input buffer and a two-character output buffer, so while one character is
   being received or sent, a second character may be unloaded from or loaded
   into the UART.  In addition, each port may specify that certain characters
   are "special," and when one of these is encountered in an input data stream,
   a read operation is terminated, and the special character is tagged to
   indicate the reason for the termination.

   A system may have one or more ADCC main and extender cards, with each pair
   residing on the Intermodule Bus at an assigned channel address from 1-15.
   Port 0 of the ADCC residing at channel address 1 is reserved for the system
   console.

   The ADCC is controlled by channel programs executed by the Channel Program
   Processor.  Because the ADCC lacks DMA capability, all reading and writing
   must be done in burst mode with a burst length of 1.  This configures the CPP
   to do byte-at-a-time I/O to the ADCC.

   The ADCC decodes 13 IMB commands:

     Mnemonic  Opcode  Command  Operation
     --------  ------  -------  ----------------------------
       RIOC     100     0000    Read I/O Data
       OBII     100     0010    Obtain Interrupt Information
       OBSI     100     0100    Obtain Service Information
       IPOLL    100     1000    Interrupt Poll
       ROCL     100     1010    Roll Call
       SPOL1    100     1100    Service Poll 1
       SPOL2    100     1110    Service Poll 2
       WIOC     110     0000    Write I/O Data
       INIT     110     0010    Initialize Channel
       SIOP     110     0100    Start I/O Program
       HIOP     110     0110    Halt I/O Program
       SMSK     110     1000    Set Interrupt Mask
       IOCL     110     1010    I/O Clear

   INIT, IOCL, and the IMB SRST, PFW, and ~PON signals all cause the ADCC to
   reset its internal state.

   SMSK clocks the data bus bit corresponding to the channel address into the
   Mask flip-flop.  The output of the flip-flop gates the IRQ line, so that IRQ
   assertion is allowed only when the corresponding mask bit is set.

   The Channel Program Processor views the ADCC as an HP-IB controller, even
   though it is not.  The ADCC responds only to the HP-IB Listen, Talk, and
   Secondary commands.  Listen or Talk addresses 0-7 select the ADCC port number
   for a read or write operation designated by the associated Secondary address.

   In order to function with the CPP, the ADCC must appear as a GIC, although
   only these four of the sixteen GIC registers are decoded (register numbers
   below are in hex):

     0 : FIFO (read/write)
     2 : Interrupt Conditions (read only)
     C : Interrupt Register (write only)
     E : Channel Configuration (read only)

   Register 0 on the ADCC is not a FIFO but is instead used to route HP-IB
   commands to the device address and modifier latches and data to and from the
   UARTs and other ADCC-specific registers.  However, the FIFO designation is
   used below to maintain the familiar GIC register nomenclature.

   The diagnostic tests that Registers 1, 3-D, and F return zero values when
   read and that writing into registers 1-B and D-F is ignored.

   OBSI and OBII are decoded from their commands and are not accessible via a
   register read.  Their values are:

   OBII -- Obtain Interrupt Information:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   0   0   0   0   0 | 0 |    Channel    |  Device   | DATA BUS
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Obtain Interrupt Information is issued to the highest-priority channel that
   is asserting IRQ, as determined by the Interrupt Poll command.  The channel
   returns the number of the highest-priority device that has an interrupt
   pending.  The returned word is used to determine the DRT entry for the
   external interrupt.  For the ADCC, the Not Valid bit (bit 8) is always clear.


   OBSI -- Obtain Service Information:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   0   0   0 | D | 0   0 |    Channel    |  Device   | DATA BUS
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     D = Device Request

   The response indicates the device requesting service from the CPP and the
   reason for the request.  For the ADCC, the Device Request bit reflects the
   state of the CSRQ signal, and the bits for Channel Request (bit 7) and Not
   Valid (bit 8) are always clear.


   Reading and writing the ADCC registers is accomplished by specifying the port
   number (0-7) in an HP-IB Talk or Listen command, a modifier number that
   selects the operation to be performed in a Secondary command, and then
   retrieving or supplying the register data byte.

   The registers are structured as follows:

   Register 0 - FIFO:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   0 | -   -   -   -   -   - |       regular data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   1 | -   -   -   -   -   - |       special data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           data byte           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 |   bus interface command   | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   While the CPP writes the full HP-IB protocol to Register 0, the ADCC responds
   only to the above formats.

   The bus interface commands to which the ADCC responds are:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 0 | 1 | D | 0 |  address  | Listen
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 0 | D | 0 |  address  | Talk
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 1 | D | 0 | modifier  | Secondary
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 1 | 1   0   0   0   0 | DSJ
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     D = Disable CSRQ if address is even and secondary is odd

   To be recognized as a Talk or Listen, either bit 12 must be zero or bit 15
   must be 1.  The address and the D bit are stored in a latch and are used to
   select the multiplexer port.

   The modifier and the D bit are stored in a latch and are used to determine
   the action to be taken.  Interpretation of the data byte depends on the
   received modifier.

   The D bit is used to inhibit CSRQ for the diagnostic.  For Section 12 of the
   diagnostic, the manual says:

    "The ADCC logic is capable of suppressing the assertion of CSRQ on the IMB.
     This is done by sending an address to talk or listen for any of the
     following device numbers: 16, 18, 20, or 22, followed by any one of the
     following secondary addresses: 17, 19, 21, 23, 25, 27, 29, or 31.

    "Any reset (PON, SRST, etc.) of the ADCC will re-enable CSRQ assertion.
     Also, CSRQ assertion can be re-enabled by sending an address to talk or
     listen for any of the following device numbers: 0, 1, 2, 3, 4, 5, 6, 7, or
     31, followed by any of the following secondary addresses: 0, 1, 2, 3, 4, 5,
     6, 7, or 16."


   Register 2 - Interrupt Conditions:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   0   0   0   0   0   0   0   0   0 | R | 0   1   0   1   0 | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R = Parallel poll response is present (ADCC is asserting CSRQ)

   Bits 11-15 have no meaning on an ADCC, but are defined for use by the Channel
   Program Processor microcode.  They indicate "Outbound FIFO is not full" and
   "Outbound FIFO is empty."  Bit 0, the OR of the other bits, is always 1 to
   indicate "An interrupt is pending."


   Register C - Interrupt Register:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   -   -   -   -   -   -   - | S |  device   | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     S = Set interrupt request for specified device number

   Setting a device IRQ bit will cause the ADCC to assert the IMB IRQ signal if
   the interrupt mask bit is set.  Device 0 has the highest priority, with
   priority decreasing as device address increases.


   Register E - Channel Configuration:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1 | 0 | C | 0   0   0   0   0   0   0   0 | E |  channel ID   | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     C = CPU/CPP (switch S1 position 0/1 = CPU/CPP)
     E = An ADCC extender is present

   Bits 12-15 supply the channel identifier; the ADCC identifier is 1.


   The Channel Program Processor controls the ADCC by writing HP-IB protocol
   commands and data to Register 0.  Because the ADCC does not have DMA, CPP
   Read and Write commands specify Burst Mode transfers with a burst size of 1.
   Therefore, the reads and writes are performed by the CPP.

   Each sequence starts with a Talk or Listen for the controller address (30),
   followed by a Listen or Talk for the ADCC addresses (0-7), followed by a
   Secondary of 00H-07H or 10H, followed by data read(s) or write(s).  The
   primary and secondary addresses are stored by ADCC registers and used to
   indicate the port number and operation, respectively.  The allowed operations
   are as follows:

   Secondary 00H - UART Data:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   0 | -   -   -   -   -   - |       regular data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   1 | -   -   -   -   -   - |       special data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           data byte           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   A read asserts the DRR signal to the UART to reset the Data Received
   condition and permit reception of the next data byte.  If the character read
   is a special character, bit 1 is set to indicate a tagged data byte.
   Otherwise, is is clear to indicate a final data byte.

   A write asserts the THRL signal to the UART, entering the data byte into the
   transmitter holding register.  Bit 0 (EOI) is ignored on writes.


   Secondary 01H - UART Status/Control:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   B | 0   0   0   0   0   0   0   0   0   0 | B | O | P | F | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - | R | -   - |  len  | S | E | I | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     B = Break detected
     O = Overrun error (OE * DR)
     P = Parity error (PE * DR)
     F = Transmit buffer empty

     R = Master reset
     S = Stop bits (0/1 = one bit/two bits)
     E = Even parity
     I = Inhibit parity generation and checking

    Word Length:

     00 = 5 bits
     01 = 6 bits
     10 = 7 bits
     11 = 8 bits

   The Transmit Buffer Empty condition is true when both the Transmitter Holding
   Register and the Transmitter Register are empty.  This status value is masked
   according to the Service Condition Mask settings.

   If the status read includes the Break Detected bit, then bit 1 is set to
   represent a tagged data byte.  Otherwise, is is clear to represent a final
   data byte.


   Secondary 02H - Modem Status/Interface Control:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1 | 1 | 0   0   0   0   0   0   0   0   0 | B | C | E | F | S | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   - | R | A | D | H | X | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     B = CB (CTS) state
     C = CC (DSR) state
     E = CE (RI) state
     F = CF (DCD) state
     S = SCF (SDCD) state

     R = Enable received character echo
     A = CA (RTS) state (0/1 = off/on)
     D = CD (DTR) state (0/1 = off/on)
     H = CH (DRS) state (0/1 = off/on)
     X = SCA (SRTS) state (0/1 = off/on)

   Reading the Modem Status register returns with bits 0 and 1 set to indicate a
   tagged data byte.

   The RS-232 circuits named above are the DCE signals present at the ADCC PCA.

   The RI (Ring Indicator) and DRS (Data Rate Selector) bits are only valid when
   the ADCC extender is installed, as all eight receivers for the CE signals
   and transmitters for the CH signals are located on the extender PCA.  If the
   extender is absent, the E bit is always zero, and the H bit is ignored.

   Reset clears the Interface Control register.


   Secondary 03H - Service Condition Masks:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   -   -   - | M | cond  | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     M = Mask bit (0/1 = disable/enable)

   Condition:

     00 = Transmit buffer not full (THRE)
     01 = Transmit buffer is empty (THRE * TRE)
     10 = Receive buffer is not empty (DR)
     11 = not used

   Mask condition 2 enables a read of Register 0 to pulse DRR (UART Data
   Received Reset), which clears the Receiver Holding Register occupancy.

   Reset clears this register.


   Secondary 04H - Clear Special Character RAM:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           character           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The specified character is no longer a special character.


   Secondary 05H - Set Special Character RAM:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           character           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The specified character becomes a special character.


   Secondary 06H - Set Baud Rates:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   - | S |   baud rate   | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     S = Rate selector (0/1 = transmitter/receiver)

   Baud Rate:

      0 = external
      1 = external
      2 = 50
      3 = 75
      4 = 134.5
      5 = 200
      6 = 600
      7 = 2400
      8 = 9600
      9 = 4800
     10 = 1800
     11 = 1200
     12 = 2400
     13 = 300
     14 = 150
     15 = 110


   Secondary 07H - Status Reference/Mask:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   - | M | B | C | E | F | S | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     M = Status selector (0/1 = reference/mask)
     B = Set CB (CTS) value
     C = Set CC (DSR) value
     E = Set CE (RI) value
     F = Set CF (DCD) value
     S = Set SCF (SDCD) value

   The RS-232 circuits named above are the DCE signals present at the ADCC PCA.

   Setting a given status mask bit enables a channel request when the selected
   status line asserts if the corresponding mask bit is clear or denies if the
   mask bit is set.

   Setting the reference and mask states for the RI (Ring Indicator) status is
   available only when the ADCC extender is installed.  Otherwise, the E bit is
   ignored.


   Secondary 10H - Device Specified Jump:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   -   -   - | 0   0   0   0   0   0 | jump  | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Jump:

     0 = UART receive buffer contains a character
     1 = UART transmit buffer is empty
     2 = UART status has changed (overrun error, parity error, break, or modem line)
     3 = device is not requesting service

   If multiple conditions are present simultaneously, the presentation priority
   is 2, 0, 1, 3.


   Identify:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 0 | 1   1   1   1   1 | Untalk
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 1 | D | 0 |  address  | Secondary
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   An Untalk followed by a secondary address 0-7 will return the Amigo ID of the
   ADCC, which is 4080H.

   If the address is 4-7 and the extender is not present, the ADCC will assert
   DNV (Data Not Valid).


   This simulation provides one ADCC main and one ADCC extender, presented as a
   single ADCC device.  The device may be cloned up to 14 times if additional
   ADCCs are desired.  Each ADCC may be configured with or without the extender.

   Each of the ADCC ports may be connected either to a Telnet session or a
   serial port on the host machine.  Port 0 of the ADCC configured for IMB
   channel 1 is connected to the simulation console, which initially performs
   I/O to the controlling window but may be rerouted instead to a Telnet session
   or serial port, if desired.  Additional port configuration options select the
   input mode (upshifted or normal), output mode (8-bit, 7-bit, printable, or
   upshifted), and whether the HP-standard ENQ/ACK handshaking is done by the
   external device or internally by the simulator.

   The ADCC supports multiplexer connection groups.  If more than one ADCC
   instance exists, then each instance may be attached to the same listening
   port as the original instance, and all instances so connected will form one
   large connection group.  After the ports on the first instance are filled,
   additional connections will be assigned to ports on the other instances in
   the order in which they were attached.  More than one connection group may be
   configured within a set of instances, with, e.g., two instances on one
   listening port, three more instances on a second listening port, and one
   instance on its own listening port.

   A device mode specifies whether terminals or the diagnostic loopback
   connector is connected to the ADCC.  Enabling the diagnostic mode simulates
   the installation of the HP 5060-5563 "ADCC TEST HOOD," as required by the
   multiplexer diagnostics.  In this mode, the transmitter on a given port is
   looped back to the receiver of the same port, and the same is done for the
   modem signal drivers and receivers.  In addition, all Telnet and serial
   sessions are disconnected, and the ADCC is detached from the listening port.
   While in diagnostic mode, the ATTACH command is not allowed.  Enabling
   terminal mode allows the ADCC to be attached to accept incoming connections
   again.

   Another device mode specifies whether the ADCC operates in real-time or
   optimized ("fast") timing mode.  In the former, character send and receive
   events occur at approximately the same rate (in machine instructions) as in
   hardware.  The latter mode increases the rate to the maximum value consistent
   with correct operation in MPE.

   If MPE is configured to use a data set on a given port, then logging off will
   cause the port to disconnect the Telnet session or drop the Data Terminal
   Ready signal on the serial port.  A port so configured by MPE will be marked
   as "data set" in a unit listing; ports not controlled will be marked as
   "direct".


   A note on terminology: HP gave the full name of the ADCC as "Asynchronous
   Data Communication Channel," "Asynchronous Data Communications Channel," and
   "Asynchronous Data Communications Controller."  The first two terms are used
   within the same HP Journal article given in the References above, while the
   September 1979 HP Journal article on the Series 33 uses the latter two terms
   within the same paragraph.  All of the terms are considered to be synonymous.


   Implementation notes:

    1. The UART register occupancy bit is selected to correspond to the Final
       Data Byte mode for the return value.

    2. Parity generation and checking is not performed, as it serves no purpose.
       Telnet doesn't support parity, physical serial ports perform parity
       operations locally, and the simulated UART sets parity generation and
       checking equivalently, so loopback errors are impossible.  While the
       serial port can be configured to support a particular parity scheme for
       communication with the connected terminal, the SIMH Telnet library has no
       way of communicating received parity errors back to our simulator.

    3. MPE supports the connection of a serial printer to an ADCC port.  This is
       supported in simulation by attaching a log file to the port.  Output to
       the port does not require a Telnet connection.
*/



#include <ctype.h>

#include "sim_defs.h"
#include "sim_tmxr.h"

#include "hp3000_defs.h"
#include "hp3000_io.h"
#include "hp3000_imb.h"
#include "hp3000_gic.h"
#include "hp3000_trace.h"



/* Program limits */

#define TERM_COUNT          8                           /* number of terminal units */
#define MAIN_COUNT          4                           /* number of main terminal units */
#define POLL_COUNT          1                           /* number of poll units */

#define UNIT_COUNT          (TERM_COUNT + POLL_COUNT)   /* number of units */
#define POLL                8                           /* the poll unit number */


/* Program constants */

#define CHAR_COUNT          256                 /* the number of character in the character set */

#define CHANNEL_COUNT       (CHAN_MAX + 1)      /* number of separate IMB channel addresses */
#define MAX_ADCC_COUNT      (CHANNEL_COUNT - 2) /* count of potential ADCC assignments */

#define ADCC_CHANNEL        1                   /* default ADCC channel number */

#define AMIGO_ID            0x4080              /* Amigo Identify value */

#define FAST_IO_TIME        uS (900)            /* initial fast receive/send time */

#define ENQ                 '\005'              /* enquire */
#define ACK                 '\006'              /* acknowledge */
#define DEL                 '\177'              /* delete */

#define DSJ                 0x10                /* Device Specified Jump secondary */

#define OCCUPIED            BIT (15)            /* UART register occupancy bit */


/* Common per-unit state variables */

#define Instance            u3                  /* unit instance identifier */


/* Device flags and accessors.

   DEV_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
             | - | - | - | - | - | - | - | - | - | - | - | - | R | P | D | E |
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R = use realistic timing
     P = S2 Processor switch (0/1 = CPU/CPP)
     D = diagnostic connector is connected
     E = ADCC extender is installed
*/

#define DEV_REALTIME        DF_BIT (3)          /* (R) use realistic timing */
#define DEV_CPP             DF_BIT (2)          /* (P) processor is CPP */
#define DEV_DIAG            DF_BIT (1)          /* (D) diagnostic connector is connected */
#define DEV_EXT             DF_BIT (0)          /* (E) ADCC extender is installed */


/* Unit flags and accessors.

   UNIT_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
   TTUF_V_UF + 11  10   9   8   7   6   5   4   3   2   1   0
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
              | - | - | - | - | - | - | - | - | - | M | L | C |  (reserved)   |
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     M = port is connected to a data set
     L = generate ACK locally
     C = CAPS LOCK is down
*/

#define UNIT_MODEM          TF_BIT (2)          /* (M) port connects to a data set */
#define UNIT_LOCALACK       TF_BIT (1)          /* (L) ENQ/ACK mode is local */
#define UNIT_CAPSLOCK       TF_BIT (0)          /* (C) caps lock is down */


/* Register definitions */

/* Register 0 - FIFO:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | - | 0 | 1 | D | 0 |  address  | Listen
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | - | 1 | 0 | D | 0 |  address  | Talk
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | - | 1 | 1 | D | - | modifier  | Secondary
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1 | -   -   -   -   -   - | 0 | 1 | 1 | 1   0   0   0   0 | DSJ
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   To be stored in the address latch, a Talk or Listen command requires either
   bit 12 to be zero or bit 15 to be 1.

   Secondaries are always stored in the modifier latch, but modifier actions
   are disabled if bit 11 is 1 (i.e., if the secondary is DSJ).

   When the D bit is set, the bus address is >= 16.
*/

#define R0_QUALIFIER        (HP_BIT (12) | HP_BIT (15)) /* address latch qualifier */
#define R0_INHIBITED        HP_BIT (12)                 /* address latch disabler */

#define R0_ADDRESS(w)       HP_BITS_TO (13, 15, w)
#define R0_MODIFIER(w)      HP_BITS_TO (13, 15, w)


/* Register E - Channel Configuration:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1 | 0 | C | 0   0   0   0   0   0   0   0 | E |  channel ID   | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define RE_EXTENDER         HP_BIT (11)         /* (E) extender is present */


/* Modifier accessors */

/* Secondary 00 - UART Data:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   0 | -   -   -   -   -   - |       regular data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   1 | -   -   -   -   -   - |       special data byte       | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           data byte           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S0_DATA(w)          HP_BITS_TO (8, 15, w)

#define S0_REGULAR          TO_HP_BITS (0, 1, R0_MODE_FINAL)
#define S0_SPECIAL          TO_HP_BITS (0, 1, R0_MODE_TAGGED)


/* Secondary 01 - UART Control/Status:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   B | 0   0   0   0   0   0   0   0   0   0 | B | O | P | F | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - | R | -   - |  len  | S | E | I | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S1_BREAK            HP_BIT (12)         /* (B) Break detected */
#define S1_OVERRUN          HP_BIT (13)         /* (O) Overrun error (OE * DR) */
#define S1_PARITY           HP_BIT (14)         /* (P) Parity error (PE * DR) */
#define S1_EMPTY            HP_BIT (15)         /* (F) Transmit buffer empty */

#define S1_REGULAR          TO_HP_BITS (0, 1, R0_MODE_FINAL)
#define S1_TAGGED           TO_HP_BITS (0, 1, R0_MODE_TAGGED)

#define S1_RESET            HP_BIT ( 8)         /* (R) Reset UART */
#define S1_STOP_BITS        HP_BIT (13)         /* (S) Stop bits (0/1 = one bit/two bits) */
#define S1_EVEN_PARITY      HP_BIT (14)         /* (E) Even parity */
#define S1_INHIBIT_PARITY   HP_BIT (15)         /* (I) Inhibit parity generation and checking */

#define S1_LENGTH(w)        HP_BITS_TO (11, 12, w)

#define S1_DATA_MASK(w)     HP_BITS (11 - S1_LENGTH (w), 15)



/* Secondary 02 - Port Control/Port Status:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1 | 1 | 0   0   0   0   0   0   0   0   0 | B | C | E | F | S | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   - | R | A | D | H | X | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     B = CB (CTS) state
     C = CC (DSR) state
     E = CE (RI) state
     F = CF (DCD) state
     S = SCF (SDCD) state

     R = Enable received character echo
     A = Assert CA (RTS)
     D = Assert CD (DTR)
     H = Assert CH (DRS)
     X = Assert SCA (SRTS)
*/

#define S2_CB               HP_BIT (11)         /* (B) CB (CTS) state */
#define S2_CC               HP_BIT (12)         /* (C) CC (DSR) state */
#define S2_CE               HP_BIT (13)         /* (E) CE (RI) state */
#define S2_CF               HP_BIT (14)         /* (F) CF (DCD) state */
#define S2_SCF              HP_BIT (15)         /* (S) SCF (SDCD) state */

#define S2_ECHO             HP_BIT (11)         /* (R) Enable received character echo */
#define S2_CA               HP_BIT (12)         /* (A) Assert CA (RTS) */
#define S2_CD               HP_BIT (13)         /* (D) Assert CD (DTR) */
#define S2_CH               HP_BIT (14)         /* (H) Assert CH (DRS) */
#define S2_SCA              HP_BIT (15)         /* (X) Assert SCA (SRTS) */

#define S2_MODEM            (S2_CA | S2_CD | S2_CH | S2_SCA)    /* the set of modem control signals */


/* Secondary 03 - Service Condition Masks:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   -   -   - | M | cond  | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S3_ENABLE           HP_BIT (13)         /* (M) Mask bit (0/1 = disable/enable) */

#define S3_CONDITION(w)     HP_BITS_TO (14, 15, w)

#define S3_THRE             BIT (0)             /* condition 0 = transmitter buffer is empty */
#define S3_TE               BIT (1)             /* condition 1 = transmitter is empty */
#define S3_DR               BIT (2)             /* condition 2 = receiver buffer is occupied */


/* Secondary 04 - Clear Special Character RAM:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           character           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S4_DATA(w)          HP_BITS_TO (8, 15, w)


/* Secondary 05 - Set Special Character RAM:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   - |           character           | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S5_DATA(w)          HP_BITS_TO (8, 15, w)


/* Secondary 06 - Set Baud Rates:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   -   - | S |   baud rate   | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define S6_RECEIVER         HP_BIT (11)         /* (S) Rate selector (0/1 = transmitter/receiver) */

#define S6_MODE(w)          HP_BITS_TO (11, 11, w)
#define S6_RATE(w)          HP_BITS_TO (12, 15, w)


/* Secondary 07 - Status Reference/Mask:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | 0 | -   -   -   -   -   -   -   - | M | B | C | E | F | S | write
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Setting the reference and mask states for the RI (Ring Indicator) status is
   available only when the ADCC extender is installed.  Otherwise, the E bit is
   ignored.
*/

#define S7_SET_MASK         HP_BIT (10)         /* (M) Status selector (0/1 = reference/mask) */
#define S7_CE               HP_BIT (13)         /* (E) Set CE (RI) value */

#define S7_VALUES(w)        HP_BITS_TO (11, 15, w)


/* Secondary 10 - Device Specified Jump:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | -   -   -   -   -   -   -   - | 0   0   0   0   0   0 | jump  | read
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define DSJ_RECEIVE         0                   /* 0 = receiver register is occupied */
#define DSJ_TRANSMIT        1                   /* 1 = transmitter buffer is empty */
#define DSJ_STATUS          2                   /* 2 = UART status has changed */
#define DSJ_IDLE            3                   /* 3 = service not requested */


/* Terminal baud rate service time delays.

   Rates correspond to the baud rate code specified in the Set Baud Rates
   command.  Baud rate settings 0 and 1 mean "external clock" and is determined
   by assuming that the connected terminal clock is set for 9600 baud, as most
   real terminals were set to their maximum rates.
*/

static const int32 delays [16] = {
/*   external    external       50          75        134.5        200         600         2400    */
/*     9600        4800        1800        1200       2400         300         150         110     */
/*  ----------  ----------  ----------  ---------- ----------- ----------- -----------  ---------- */
    mS (1.042), mS (1.042), mS (200.0), mS (133.3), mS (74.39), mS (50.00), mS (16.67), mS (4.166),
    mS (1.042), mS (2.084), mS (5.556), mS (8.334), mS (4.166), mS (33.34), mS (66.67), mS (90.91)
    };


/* Activation reasons */

typedef enum {
    Receive,                                    /* port is waiting for a character reception */
    Send,                                       /* port is waiting for a character transmission */
    Handshake,                                  /* port is waiting for completion of a handshake */
    Stall                                       /* port is waiting for an output buffer stall */
    } ACTIVATOR;


/* Trace structure declarations */

typedef enum {                                  /* register format index selector */
    Read  = 0,                                  /*   read format */
    Write = 1                                   /*   write format */
    } FORMAT_MODE;

static const BITSET_NAME R2_names [] = {        /* Register 2 bit names */
    "IRQ",                                      /*   bit  0 */
    NULL,                                       /*   bit  1 */
    NULL,                                       /*   bit  2 */
    NULL,                                       /*   bit  3 */
    NULL,                                       /*   bit  4 */
    NULL,                                       /*   bit  5 */
    NULL,                                       /*   bit  6 */
    NULL,                                       /*   bit  7 */
    NULL,                                       /*   bit  8 */
    NULL,                                       /*   bit  9 */
    "poll response",                            /*   bit 10 */
    NULL,                                       /*   bit 11 */
    "available",                                /*   bit 12 */
    NULL,                                       /*   bit 13 */
    "empty"                                     /*   bit 14 */
    };

static const BITSET_FORMAT R2_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (R2_names, 1, msb_first, no_alt, no_bar) };


static const BITSET_NAME RC_control_names [] = {    /* Register C control bit names */
    "\1set\0clear"                                  /*   bit 12 */
    };

static const BITSET_FORMAT RC_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (RC_control_names, 3, msb_first, has_alt, append_bar) };


static const BITSET_NAME RE_status_names [] = { /* Register E status bit names */
    "\1S1 CPP\0S1 CPU",                         /*   bit  2 */
    NULL,                                       /*   bit  3 */
    NULL,                                       /*   bit  4 */
    NULL,                                       /*   bit  5 */
    NULL,                                       /*   bit  6 */
    NULL,                                       /*   bit  7 */
    NULL,                                       /*   bit  8 */
    NULL,                                       /*   bit  9 */
    NULL,                                       /*   bit 10 */
    "extender"                                  /*   bit 11 */
    };

static const BITSET_FORMAT RE_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (RE_status_names, 4, msb_first, has_alt, append_bar) };


static const BITSET_NAME S1_control_names [] = {    /* Secondary 01 bit names */
    "reset",                                        /*   bit  8 */
    NULL,                                           /*   bit  9 */
    NULL,                                           /*   bit 10 */
    NULL,                                           /*   bit 11 */
    NULL,                                           /*   bit 12 */
    "\1two stop bits\0one stop bit",                /*   bit 13 */
    "\1even parity\0odd parity",                    /*   bit 14 */
    "inhibit parity"                                /*   bit 15 */
    };

static const BITSET_NAME S1_status_names [] = { /* Secondary 01 bit names */
    "break detected",                           /*   bit 12 */
    "overrun error",                            /*   bit 13 */
    "parity error",                             /*   bit 14 */
    "transmit buffer empty"                     /*   bit 15 */
    };

static const BITSET_FORMAT S1_control_format =  /* names, offset, direction, alternates, bar */
    { FMT_INIT (S1_control_names, 0, msb_first, has_alt, no_bar) };

static const BITSET_FORMAT S1_status_format =   /* names, offset, direction, alternates, bar */
    { FMT_INIT (S1_status_names, 0, msb_first, no_alt, no_bar) };


static const BITSET_NAME S2_control_names [] = {    /* Secondary 02 control bit names */
    "echo",                                         /*   bit 11 */
    "CA (RTS)",                                     /*   bit 12 */
    "CD (DTR)",                                     /*   bit 13 */
    "CH (DRS)",                                     /*   bit 14 */
    "SCA (SRTS)"                                    /*   bit 15 */
    };

static const BITSET_NAME S2_status_names [] = { /* Secondary 02 status bit names */
    "CB (CTS)",                                 /*   bit 11 */
    "CC (DSR)",                                 /*   bit 12 */
    "CE (RI)",                                  /*   bit 13 */
    "CF (DCD)",                                 /*   bit 14 */
    "SCF (SDCD)"                                /*   bit 15 */
    };

static const BITSET_FORMAT S2_control_format =  /* names, offset, direction, alternates, bar */
    { FMT_INIT (S2_control_names, 0, msb_first, no_alt, no_bar) };

static const BITSET_FORMAT S2_status_format =   /* names, offset, direction, alternates, bar */
    { FMT_INIT (S2_status_names, 0, msb_first, no_alt, no_bar) };


static const BITSET_NAME S3_names [] = {        /* Secondary 03 service condition names */
    "RR occupied",                              /*   bit 13 */
    "TR and THR empty",                         /*   bit 14 */
    "THR empty"                                 /*   bit 15 */
    };

static const BITSET_FORMAT S3_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (S3_names, 0, msb_first, no_alt, no_bar) };


#define S7_names            S2_status_names     /* Secondary 07 bit names (same as Secondary 02 status names) */

static const BITSET_FORMAT S7_format =          /* names, offset, direction, alternates, bar */
    { FMT_INIT (S7_names, 0, msb_first, no_alt, no_bar) };


static const char *mode_names [2] = {           /* set baud rate mode, indexed by S6_MODE */
    "transmitter",                              /*   0 = set transmitter */
    "receiver"                                  /*   1 = set receiver */
    };

static const char *baud_names [16] = {          /* baud rate, indexed by S6_RATE */
    "external clock",                           /*   0 = external */
    "external clock",                           /*   1 = external */
    "50 baud",                                  /*   2 = 50 */
    "75 baud",                                  /*   3 = 75 */
    "134.5 baud",                               /*   4 = 134.5 */
    "200 baud",                                 /*   5 = 200 */
    "600 baud",                                 /*   6 = 600 */
    "2400 baud",                                /*   7 = 2400 */
    "9600 baud",                                /*   8 = 9600 */
    "4800 baud",                                /*   9 = 4800 */
    "1800 baud",                                /*  10 = 1800 */
    "1200 baud",                                /*  11 = 1200 */
    "2400 baud",                                /*  12 = 2400 */
    "300 baud",                                 /*  13 = 300 */
    "150 baud",                                 /*  14 = 150 */
    "110 baud"                                  /*  15 = 110 */
    };

static const char *dsj_names [4] = {            /* DSJ values, indexed by DSJ */
    "RR occupied",                              /*   0 = receiver register occupied */
    "THR empty or TR and THR empty",            /*   1 = transmitter register(s) empty */
    "modem change",                             /*   2 = status has changed */
    "idle"                                      /*   3 = service not requested */
    };


/* Trace identifier definitions.

   The WINDOW_ON macro tests if the supplied port number is to be included or
   excluded from the trace stream.  It returns TRUE if the bit in the window set
   corresponding to the port number is set and FALSE if it is clear.  If the bit
   is set, the trace will be allowed; otherwise, it will be suppressed.
*/

typedef enum {
    Default_Mode      = 0,
    Trigger_On_Input  = BIT ( 0),
    Trigger_On_Output = BIT ( 1)
    } TRIGGER_MODE;

typedef enum {
    Nothing,
    Input,
    Output,
    Port
    } SET_OPTION;

typedef struct {
    TRIGGER_MODE  modes;
    uint8         input [CHAR_COUNT];
    uint8         output [CHAR_COUNT];
    uint32        input_ports;
    uint32        output_ports;
    SET_OPTION    last_option;
    } TRIGGER_SET;

typedef enum {
// initial
    Trace_DTR_Drop,
    Trace_Connected,
    Trace_DSR_Raise,
    Trace_DSR_Drop,
    Trace_Port_Service,
    Trace_Overrun,
    Trace_Break,
    Trace_ENQ_Absorbed,
    Trace_Poll_Schedule,
    Trace_Poll_Service,

// 0
    Trace_Poll_Stopped,

// 1
    Trace_OBSI_Status,
    Trace_OBII_Status,

// 2
    Trace_Modifier_0_Read,
    Trace_Modifier_1_Read,
    Trace_Modifier_2_Read,
    Trace_DSJ,
    Trace_Modifier_0_Write,
    Trace_Modifier_1_Write,
    Trace_Modifier_2_Write,
    Trace_Modifier_4_Write,
    Trace_Modifier_5_Write,
    Trace_Modifier_6_Write,
    Trace_Modifier_7_Write,
    Trace_CSRQ_Assert,
    Trace_Reception,
    Trace_Port_Send,
    Trace_Port_Rescheduled,
    Trace_Port_Wait,
    Trace_Loopback,
    Trace_Discard_Output,
    Trace_Connection_Loss,
    Trace_Stall,
    Trace_Port_Scheduled,

// 3
    Trace_Modifier_3_Write,
    Trace_Register,

// 5
    Trace_Service_Condition
    } TRACE_ID;


#define TRIGGER_ON_INPUT(test_address, test_data) \
    adcc->trigger_modes & Trigger_On_Input \
      && adcc->trigger_input [test_data] \
      && adcc->trigger_input_ports & BIT (test_address)

#define TRIGGER_ON_OUTPUT(test_address, test_data) \
    adcc->trigger_modes & Trigger_On_Input \
      && adcc->trigger_output [test_data] \
      && adcc->trigger_output_ports & BIT (test_address)

#define WINDOW_ON(test_address) \
    BIT (test_address) & adcc->window_ports


/* Port state structure declarations.

   The port state structure collects all of the port-specific state.  This
   structure is included as part of the ADCC state structure.


   Implementation notes:

    1. A special character RAM element is addressed by the character value and
       contains 1 if the character is special and 0 if not.  In hardware, the
       RAM is organized as 256 x 1.  In simulation, the elements are bytes, as
       they are the smallest units that are efficiently addressed.
*/

typedef struct {                                /* per-port state */
    uint32   dsj;                               /*   Device Specified Jump value */
    uint32   uart_thr;                          /*   UART transmitter holding register */
    uint32   uart_tr;                           /*   UART transmitter register */
    uint32   uart_rhr;                          /*   UART receiver holding register */
    uint32   uart_rr;                           /*   UART receiver register */
    uint32   uart_data_mask;                    /*   UART byte length mask */
    HP_WORD  uart_control;                      /*   UART control lines */
    HP_WORD  uart_status;                       /*   UART status lines */
    HP_WORD  modem_control;                     /*   modem control lines */
    HP_WORD  modem_status;                      /*   modem status lines */
    uint32   service_masks;                     /*   service condition masks */
    int32    transmit_delay;                    /*   realtime transmit baud rate delay */
    int32    receive_delay;                     /*   realtime receive baud rate delay */
    HP_WORD  status_mask;                       /*   modem status lines mask */
    HP_WORD  status_ref;                        /*   modem status lines reference */
    HP_BYTE  special_ram [CHAR_COUNT];          /*   special character RAM */
    } PORT_STATE;


/* ADCC state structure declarations.

   The ADCC state structure collects all of the card-specific state.  This
   structure is replicated when multiple ADCC instances are created.


   Implementation notes:

    1. CSRQ may be asserted for a UART service request and also for a new status
       request, i.e., when the channel program is starting or halting.  Multiple
       requests are prioritized, recognizing UART requests before new status
       requests for each port, and requests for port 0 before port 1, etc.  In
       simulation, per-port requests are interleaved in the "service_requests"
       field, with the UART requests encoded as bit number (busaddr * 2) and new
       status requests encoded as bit number (busaddr * 2 + 1), producing this
       layout:

          15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
         | N | U | N | U | N | U | N | U | N | U | N | U | N | U | N | U |
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
         | port  | port  | port  | port  | port  | port  | port  | port  |
         |   7   |   6   |   5   |   4   |   3   |   2   |   1   |   0   |

       Applying the LSB_NUMBER macro then obtains the correctly prioritized
       request.

    2. Nominally, the HP-IB primary addresses are 0-3 for the main PCA ports and
       4-7 for the extender PCA ports, and the secondary indicates the function
       to perform.  However, the primary address field is only partially
       decoded, and some "out of range" addresses are used by the ADCC
       diagnostic to suppress CSRQ assertion.  Specifically, CSRQ will be
       asserted only if the primary address is 0-15, or the primary address is
       odd, or the secondary address is 0-15, or the secondary address is even.

       These primary and secondary address conditions affect the CSRQ enable
       independently.  So, for example, a disabled primary followed by a
       disabled secondary followed by an enabled secondary must enable CSRQ
       assertion.  Therefore, the "csrq_enable" field uses two separate enable
       bits as follows:

          31  30  29    ...    3   2   1   0
         +---+---+---+---+---+---+---+---+---+
         | 0   0   0   0   0   0   0 | S | P |
         +---+---+---+---+---+---+---+---+---+

       CSRQ assertion is enabled if either the primary (P) condition or the
       secondary (S) condition is met.
*/

#define UART_BIT(a)         BIT ((a) * 2)       /* CSRQ assertion bit for UART service */
#define NEW_STATUS_BIT(a)   BIT ((a) * 2 + 1)   /* CSRQ assertion bit for new status */

#define CSRQ_PRI_EN(d)      ((BUS_ADDRESS (d) < 16 ||      (d) & 1 ) << 0)  /* address < 16 or is odd */
#define CSRQ_SEC_EN(d)      ((BUS_ADDRESS (d) < 16 || not ((d) & 1)) << 1)  /* address < 16 or is even */

typedef enum {                                  /* Amigo Identify sequencer */
    Untalk,                                     /*   0 = waiting for UNT */
    My_Secondary,                               /*   1 = waiting for MSA */
    ID_Byte_1,                                  /*   2 = waiting for ID byte 1 read */
    ID_Byte_2                                   /*   3 = waiting for ID byte 2 read */
    } ID_STATE;

typedef struct {                                /* ADCC state */
    PORT_STATE    ports [TERM_COUNT];           /*   port states */
    uint32        csrq_enable;                  /*   set of two CSRQ assertion enables */
    HP_WORD       address_group;                /*   addressing state (talk or listen) */
    HP_WORD       address;                      /*   port number (bus address) */
    HP_WORD       modifier;                     /*   port modifier (secondary address) */
    uint32        interrupts;                   /*   set of device interrupt requests */
    uint32        irq_mask;                     /*   interrupt request mask channel bit */
    uint32        service_requests;             /*   set of CSRQ assertions */
    uint32        running;                      /*   set of active channel programs */
    ID_STATE      identify;                     /*   Amigo Identify current state */
    int32         fast_time;                    /*   fast transmit/receive scheduling time */
    TRIGGER_MODE  trigger_modes;                    //
    uint8         trigger_input [CHAR_COUNT];      //
    uint8         trigger_output [CHAR_COUNT];     //
    uint32        trigger_input_ports;          //
    uint32        trigger_output_ports;         //
    uint32        window_ports;                 /*   the set of ports that are within the trace window */
    } ADCC_STATE;


/* Multiple instance structure declarations */

typedef struct {                                /* ADCC card instance structure */
    DEVICE      *dptr;                          /*   a pointer to the device instance */
    ADCC_STATE  *sptr;                          /*   a pointer to the state instance */
    TMXR        *mptr;                          /*   a pointer to the multiplexer instance */
    } CARD_ENTRY;


/* Interface local state declarations.

   The local state consists of the ADCC state structure.  The original device's
   state structure is used as a template to create a new state instance for each
   new device instance.
*/

static ADCC_STATE adcc_state;                   /* the original device instance state structure */


/* Interface local SCP support routine declarations */

static IMB_INTERFACE adcc_imb_interface;
static t_stat        adcc_reset       (DEVICE *dptr);
static t_stat        adcc_attach      (UNIT *uptr, char *cptr);
static t_stat        adcc_detach      (UNIT *uptr);
static uint32        adcc_adjust      (DEVICE **dvptr, int32 delta);
static t_stat        adcc_connect     (UNIT *uptr, int32 line,  char *cptr, void *desc);
static t_stat        adcc_extend      (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat        adcc_diag        (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat        adcc_set_endis   (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat        adcc_show_status (FILE *st, UNIT *uptr, int32 value, void *desc);


/* Interface local utility routine declarations */

static IMB_OUTBOUND read_register  (uint32 adcc_id, REGISTER Register);
static IMB_OUTBOUND write_register (uint32 adcc_id, REGISTER Register, IMB_DATA data);
static IMB_DATA     read_data      (uint32 adcc_id);
static void         write_data     (uint32 adcc_id, IMB_DATA data);
static t_bool       check_service  (uint32 adcc_id, uint32 address);
static void         update_status  (uint32 adcc_id, uint32 address);
static t_stat       line_service   (UNIT *uptr);
static t_stat       poll_service   (UNIT *uptr);
static int32        activate_unit  (UNIT *uptr, ACTIVATOR reason);


/* ADCC local trace routine declarations */

static t_stat        format_register (TRACE_RECORD *eptr);
static TRACE_PARSER  parse_trigger;
static TRACE_HANDLER trace_adcc;


/* ADCC local trace data structures */

static TRACE_PARSER_TABLE trigger_table [] = {
    { "INPUT",  Input,  parse_trigger },
    { "OUTPUT", Output, parse_trigger },
    { "PORT",   Port,   parse_trigger },
    { NULL,     0,      NULL          }
    };


/* Interface SCP data structure declarations */


DEVICE adcc_dev;                                /* incomplete device structure */


/* Terminal multiplexer library structures.

   The ADCC uses the connection line order feature to bypass port 0 of the
   instance that is configured for IMB channel 1, which is dedicated to the
   system console.  For convenience, the system console is connected to the
   simulation console.  As such, it calls the console I/O routines instead of
   the terminal multiplexer routines.
*/

static int32 adcc_order [TERM_COUNT] = {        /* line connection order */
    1,  2,  3,  4,  5,  6,  7,  -1
    };

static TMLN adcc_ldsc [TERM_COUNT] = {          /* line descriptors */
    { 0 }
    };

static TMXR adcc_mdsc = {                       /* multiplexer descriptor */
    TERM_COUNT,                                 /*   number of terminal lines */
    0,                                          /*   listening port (reserved) */
    0,                                          /*   master socket  (reserved) */
    adcc_ldsc,                                  /*   line descriptors pointer */
    adcc_order,                                 /*   line connection order pointer */
    &adcc_dev                                   /*   multiplexer device */
    };


/* Device information block */

static DIB adcc_dib = {
    NULL,                                       /* the device's IOP interface function pointer */
    NULL,                                       /* the execution initialization function pointer */
    DEVNO_UNUSED,                               /* the device number */
    SRNO_UNUSED,                                /* the service request number */
    INTPRI_UNUSED,                              /* the interrupt priority */
    INTMASK_UNUSED,                             /* the interrupt mask */
    0,                                          /* the card index */
    CLEAR,                                      /* the interrupt request flip-flop */
    CLEAR,                                      /* the interrupt active flip-flop */
    FALSE,                                      /* TRUE if service has been requested */
    &adcc_imb_interface,                        /* the device's IMB interface function pointer */
    ADCC_CHANNEL,                               /* the IMB channel number */
    SLOT_UNUSED,                                /* the IMB slot number */
    CPU_IS_IMB,                                 /* the supported system configurations */
    trace_adcc                                  /* the trace handler function pointer */
    };


/* Unit lists.

   The first eight ADCC units correspond to the eight multiplexer ports provided
   by the ADCC main and extender cards.  These handle character I/O via the
   Telnet library.  A ninth unit is responsible for polling for connections and
   socket I/O.  It also holds the master socket.

   Port 0 of the ADCC configured for IMB channel 1 is reserved for the system
   console and is connected to the simulation console.  As such, it is not
   likely to be using an HP terminal emulator, so the default is CAPSLOCK input
   mode and 7P output mode.  The remainder of the ports default to NOCAPSLOCK
   and 7B, as they're likely to be connected to HP terminals or terminal
   emulators.  All ports initially omit the UNIT_MODEM flag to allow the MPE
   terminal subtype configuration to determine which ports support data sets and
   which do not.

   The ADCC line service routine runs only when there are characters to read or
   write.  It is scheduled either at a realistic rate corresponding to the
   programmed baud rate of the port to be serviced, or at a somewhat faster
   optimized rate.  The multiplexer connection and input poll must run
   continuously, but it may operate much more slowly, as the only requirement is
   that it must not present a perceptible lag to human input.  It is coscheduled
   with the process clock to permit idling.  The poll unit is hidden by
   disabling it, so as to present a logical picture of the multiplexer to the
   user.


   Implementation notes:

    1. If an input character is being echoed at the same time as an output
       character is being transmitted, a collision occurs.  In hardware, the
       logical AND of the two characters would appear on the TX line.  In
       simulation, this condition is indicated by the substitution of a DEL
       character.  Note, however, that by default Mode 7P filters out DEL unless
       it has been changed with a SET CONSOLE PCHAR command.

    2. If additional ADCC instances are created with the SET ADCC COUNT command,
       the configuration for port 0 is copied from port 1, as the special
       configuration for the system console pertains only to the ADCC assigned
       to IMB channel 1.
*/

static UNIT adcc_unit [UNIT_COUNT] = {
    { UDATA (&line_service, TT_MODE_7P | UNIT_LOCALACK | UNIT_CAPSLOCK, 0) },   /* port 0 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 1 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 2 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 3 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 4 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 5 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 6 */
    { UDATA (&line_service, TT_MODE_7B | UNIT_LOCALACK,                 0) },   /* port 7 */
    { UDATA (&poll_service, UNIT_ATTABLE | UNIT_DIS | UNIT_IDLE,        0) }    /* input poll */
    };


/* Register list */

static REG adcc_reg [] = {
/*    Macro   Name      Location          (Field)        Radix  Width  Offset     Depth      Flags          */
/*    ------  --------  -------------------------------  -----  -----  ------  -----------  --------------- */
    { DRDATA (ADDRGRP,  adcc_state.address_group,                 3),                       PV_LEFT         },
    { DRDATA (PRIADDR,  adcc_state.address,                       5),                       PV_LEFT         },
    { DRDATA (SECADDR,  adcc_state.modifier,                      5),                       PV_LEFT         },
    { ORDATA (IRQMASK,  adcc_state.irq_mask,                     16),                       PV_RZRO         },
    { GRDATA (IRQ,      adcc_state.interrupts,             2,     8,     0),                PV_RZRO         },
    { GRDATA (CSRQ,     adcc_state.service_requests,       2,    16,     0),                PV_RZRO         },
    { GRDATA (CPRUN,    adcc_state.running,                2,     8,     0),                PV_RZRO         },
    { DRDATA (FTIME,    adcc_state.fast_time,                    24),                       PV_LEFT         },

    { SRDATA (THR,      adcc_state.ports, uart_thr,        8,     8,     0,    TERM_COUNT), PV_RZRO | REG_A },
    { SRDATA (TR,       adcc_state.ports, uart_tr,         8,     8,     0,    TERM_COUNT), PV_RZRO | REG_A },
    { SRDATA (RHR,      adcc_state.ports, uart_rhr,        8,     8,     0,    TERM_COUNT), PV_RZRO | REG_A },
    { SRDATA (RR,       adcc_state.ports, uart_rr,         8,     8,     0,    TERM_COUNT), PV_RZRO | REG_A },
    { SRDATA (DSJ,      adcc_state.ports, dsj,            10,     2,     0,    TERM_COUNT), PV_LEFT         },
    { SRDATA (MCONTROL, adcc_state.ports, modem_control,   2,     5,     0,    TERM_COUNT), PV_RZRO         },
    { SRDATA (MSTATUS,  adcc_state.ports, modem_status,    2,     5,     0,    TERM_COUNT), PV_RZRO         },
    { SRDATA (SVMASK,   adcc_state.ports, service_masks,   2,     3,     0,    TERM_COUNT), PV_RZRO         },
    { SRDATA (STMASK,   adcc_state.ports, status_mask,     2,     5,     0,    TERM_COUNT), PV_RZRO         },
    { SRDATA (STREF,    adcc_state.ports, status_ref,      2,     5,     0,    TERM_COUNT), PV_RZRO         },

    { SVDATA (STATE,    adcc_state)                                                                         },
    { NULL }
    };


/* Modifier list */

#define ETAB_DEVICE         (MTAB_EDV)
#define ETAB_CHANNEL        (MTAB_EDV | MTAB_ERX (10))

static MTAB adcc_mod [] = {
/*    Mask Value     Match Value    Print String      Match String  Validation  Display  Descriptor */
/*    -------------  -------------  ----------------  ------------  ----------  -------  ---------- */
    { UNIT_MODEM,    UNIT_MODEM,    "data set",       NULL,         NULL,       NULL,    NULL       },
    { UNIT_MODEM,    0,             "direct",         NULL,         NULL,       NULL,    NULL       },

    { UNIT_LOCALACK, UNIT_LOCALACK, "local ENQ/ACK",  "LOCALACK",   NULL,       NULL,    NULL       },
    { UNIT_LOCALACK, 0,             "remote ENQ/ACK", "REMOTEACK",  NULL,       NULL,    NULL       },

    { UNIT_CAPSLOCK, UNIT_CAPSLOCK, "CAPS LOCK down", "CAPSLOCK",   NULL,       NULL,    NULL       },
    { UNIT_CAPSLOCK, 0,             "CAPS LOCK up",   "NOCAPSLOCK", NULL,       NULL,    NULL       },

    { TT_MODE,       TT_MODE_UC,    "UC output",      "UC",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_7B,    "7b output",      "7B",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_7P,    "7p output",      "7P",         NULL,       NULL,    NULL       },
    { TT_MODE,       TT_MODE_8B,    "8b output",      "8B",         NULL,       NULL,    NULL       },

/*    Entry Flags          Value  Print String   Match String  Validation        Display          Descriptor          */
/*    -------------------  -----  -------------  ------------  ----------------  ---------------  ------------------- */
    { MTAB_XUN | MTAB_NC,    0,   "LOG",         "LOG",        tmxr_set_log,    &tmxr_show_log,  (void *) &adcc_mdsc },
    { MTAB_XUN | MTAB_NC,    0,   NULL,          "NOLOG",      tmxr_set_nolog,  NULL,            (void *) &adcc_mdsc },
    { MTAB_XUN,              0,   NULL,          "DISCONNECT", tmxr_dscln,      NULL,            (void *) &adcc_mdsc },
    { MTAB_XDV,              0,   NULL,          "CONNECT",    adcc_connect,    NULL,            (void *) &adcc_mdsc },

/*    Mask Value     Match/Max    Print String         Match String  Validation   Display  Descr  4.x   Expansion Flags */
/*    ------------  ------------  -------------------  ------------  -----------  -------  -----  ----  --------------- */
    { CHAN_MASK,    CHAN_MAX,     "channel ",          "CHANNEL",    NULL,        NULL,    &adcc_dib.channel_number, \
                                                                                                  NULL, ETAB_CHANNEL    },
    { DEV_REALTIME, 0,            "fast timing",       "FASTTIME",   NULL,        NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_REALTIME, DEV_REALTIME, "real timing",       "REALTIME",   NULL,        NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_DIAG,     0,            "terminal cable",    "TERMINAL",   adcc_diag,   NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_DIAG,     DEV_DIAG,     "diagnostic hood",   "DIAGNOSTIC", adcc_diag,   NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_CPP,      0,            "CPU",               "CPU",        NULL,        NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_CPP,      DEV_CPP,      "CPP",               "CPP",        NULL,        NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_EXT,      0,            "no extender",       "NOEXTENDER", adcc_extend, NULL,    NULL,  NULL, ETAB_DEVICE     },
    { DEV_EXT,      DEV_EXT,      "extender",          "EXTENDER",   adcc_extend, NULL,    NULL,  NULL, ETAB_DEVICE     },

/*    Entry Flags           Value  Print String   Match String  Validation        Display            Descriptor          */
/*    --------------------  -----  -------------  ------------  ----------------  -----------------  ------------------- */
    { MTAB_XDV,               0,   "",            NULL,         NULL,             adcc_show_status,  NULL                },
    { MTAB_XDV | MTAB_NMO,    1,   "CONNECTIONS", NULL,         NULL,             tmxr_show_cstat,   (void *) &adcc_mdsc },
    { MTAB_XDV | MTAB_NMO,    0,   "STATISTICS",  NULL,         NULL,             tmxr_show_cstat,   (void *) &adcc_mdsc },
    { MTAB_XDV | MTAB_NMO,    1,   "LINEORDER",   "LINEORDER",  tmxr_set_lnorder, tmxr_show_lnorder, (void *) &adcc_mdsc },

    { MTAB_XDV,               1,   NULL,          "ENABLED",    adcc_set_endis,   NULL,              (void *) &adcc_mdsc },
    { MTAB_XDV,               0,   NULL,          "DISABLED",   adcc_set_endis,   NULL,              (void *) &adcc_mdsc },
    { 0 }
    };


/* Debugging trace list */

static DEBTAB adcc_deb [] = {                   /* debugging trace list */
    { "CMD",   TRACE_CMD   },                   /*   trace interface commands */
    { "CSIRQ", TRACE_CSIRQ },                   /*   trace channel service and interrupt requests */
    { "CSRW",  TRACE_CSRW  },                   /*   trace interface control, status, read, and write actions */
    { "SERV",  TRACE_SERV  },                   /*   trace port unit service scheduling */
    { "PSERV", TRACE_PSERV },                   /*   trace poll unit service scheduling */
    { "XFER",  TRACE_XFER  },                   /*   trace data transmissions and receptions */
    { "IMBRQ", TRACE_IMBRQ },                   /*   trace IMB channel service and interrupt request actions */
    { "IMBUS", TRACE_IMBUS },                   /*   trace IMB signals and data words */
    { NULL,    0           }
    };

#define TRACE_COMMON        (TRACE_CMD | TRACE_CSIRQ | TRACE_XFER | TRACE_IMBRQ)
#define TRACE_ALL           (TRACE_COMMON | TRACE_CSRW | TRACE_SERV | TRACE_PSERV | TRACE_IMBUS)


/* Device descriptor */

DEVICE adcc_dev = {
    "ADCC",                                     /* device name */
    adcc_unit,                                  /* unit array */
    adcc_reg,                                   /* register array */
    adcc_mod,                                   /* modifier array */
    UNIT_COUNT,                                 /* number of units */
    10,                                         /* address radix */
    PA_WIDTH,                                   /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    8,                                          /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    adcc_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    adcc_attach,                                /* attach routine */
    adcc_detach,                                /* detach routine */
    &adcc_dib,                                  /* device information block pointer */
    DEV_DISABLE | DEV_DEBUG | DEV_EXT,          /* device flags */
    0,                                          /* debug control flags */
    adcc_deb,                                   /* debug flag name array */
    NULL,                                       /* memory size change routine */
    NULL,                                       /* logical device name */
    NULL,                                       /* (4.0 dummy) help routine */
    NULL,                                       /* (4.0 dummy) help attach routine */
    NULL,                                       /* (4.0 dummy) help context */
    NULL,                                       /* (4.0 dummy) description */
    adcc_adjust                                 /* the device adjustment function pointer */
    };


/* Interface local allocation state declarations */

static uint32 adcc_count = 1;                   /* current count of ADCC device instances */

static CARD_ENTRY adccs [MAX_ADCC_COUNT] = {    /* ADCC card entries, indexed by the instance ID */
    { &adcc_dev, &adcc_state, &adcc_mdsc }      /*   instance 0 (original) */
    };



/****************************************
 *                                      *
 * Interface local SCP support routines *
 *                                      *
 ****************************************/


/* ADCC interface.

   The Asynchronous Data Communications Channel is installed on the Intermodule
   Bus and provides terminal multiplexing for four (main only) or eight (main
   plus extender) ports.  Serial conversion and data management is provided by
   AMI S1602P (a single-supply version of the Western Digital TR1602) UARTs for
   each port.

   IMB commands are defined by an operation code asserted on OP0-OP2.  For
   memory reads and writes, ADRE8-E1 and ADR0-15 give the 24-bit address, and
   DATA0-15 supply or accept the 16-bit data, respectively.  For I/O commands,
   ADR0-15 give the command code, register number, and channel number, and
   DATA0-15 supply or accept the 16-bit data, respectively.  The command format
   is:

   IMB Command Code Format:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |    Command    |   Register    | - |    Channel    | -   -   - |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The ADCC decodes 13 IMB commands.

   In hardware, the ADCC has two user-settable switches that are modeled in
   simulation as user device flags.  The switch functions are:

     Switch   Choices   Setting  Action                Alternate Action
     ------  ---------  -------  --------------------  -------------------
       S1      0-15       11     Sets channel number   (not applicable)
       S2     CPU/CPP     CPU    Asserts CSRQ1         Asserts CSRQ2


   Implementation notes:

    1. In the absence of a service or interrupt request, the default device
       number reported is 3 for OBSI and 7 for OBII.  These values are
       independent of the presence or absence of the ADCC extender PCA.

    2. SPOL returns a positive response when a service request is present.  The
       response does not depend on the CSRQ disabler, as that only affects
       assertion of the CSRQ line on the IMB.
*/

static IMB_OUTBOUND adcc_imb_interface (DIB *dibptr, IMB_INBOUND inbound)
{
const uint32      channel  = dibptr->channel_number;    /* the IMB channel number */
const uint32      adcc_id  = dibptr->card_index;        /* the instance identifier */
DEVICE     *const dptr     = adccs [adcc_id].dptr;      /* the per-instance device pointer */
ADCC_STATE *const adcc     = adccs [adcc_id].sptr;      /* the per-instance ADCC state pointer */
uint32            address;
IMB_IO_COMMAND    command;
IMB_OUTBOUND      outbound = { 0, ADN | DDN | PRO };    /* the initial outbound value and signal set */

switch (inbound.opcode) {                               /* dispatch the IMB opcode */

    case IMB_IO_Write:                                  /* I/O Write */
    case IMB_IO_Read:                                   /* I/O Read */
        command  = IMB_COMMAND (inbound.address,        /* get the IMB command */
                                inbound.opcode);

        switch (command) {                              /* dispatch the IMB I/O command */

            case SPOL1:                                 /* Service Poll 1 */
                if (adcc->service_requests              /* if a service request is pending */
                  && not (dptr->flags & DEV_CPP))       /*   and S2 is in the CPU position */
                    outbound.data = HP_BIT (channel);   /*     then respond with our channel bit */
                break;


            case SPOL2:                                 /* Service Poll 2 */
                if (adcc->service_requests              /* if a service request is pending */
                  && dptr->flags & DEV_CPP)             /*   and S2 is in the CPP position */
                    outbound.data = HP_BIT (channel);   /*     then respond with our channel bit */
                break;


            case IOCL:                                  /* I/O Clear */
            case INIT:                                  /* Initialize Channel */
                adcc_reset (dptr);                      /* reset the ADCC */
                break;


            case ROCL:                                  /* Roll Call */
                outbound.data = HP_BIT (channel);       /* respond with our channel bit */
                break;


            case SIOP:                                  /* Start I/O Program */
            case HIOP:                                  /* Halt I/O Program */
                address = SIOP_DEVICE (inbound.data);   /* get the device address */

                if (inbound.data & SIOP_CLEAR_STATUS)                       /* if the S bit is set */
                    adcc->service_requests &= ~NEW_STATUS_BIT (address);    /*   then clear the new status bit */
                else                                                        /* otherwise */
                    adcc->service_requests |= NEW_STATUS_BIT (address);     /*   set the new status bit */


                if (command == SIOP)                    /* if the program is starting */
                    adcc->running |= BIT (address);     /*   then set the port's run bit */
                else                                    /* otherwise */
                    adcc->running &= ~BIT (address);    /*   clear the port's run bit */

                check_service (adcc_id, address);       /* check to see if CSRQ should be asserted */
                break;


            case RIOC:                                  /* Read I/O Data */
                outbound = read_register (adcc_id,      /* read the specified register */
                                          IMB_REGISTER (inbound.address));
                break;


            case WIOC:                                  /* Write I/O Data */
                outbound = write_register (adcc_id,     /* write the specified register */
                                           IMB_REGISTER (inbound.address),
                                           inbound.data);
                break;


            case OBSI:                                      /* Obtain Service Information */
                outbound.data = RF_TO_CHANNEL (channel);    /* set the channel number */

                if (adcc->service_requests) {                           /* if a service request is present */
                    address = LSB_NUMBER (adcc->service_requests) / 2;  /*   then get the highest priority device */
                    outbound.data |= RF_DEVRQ | RF_TO_DEVICE (address); /*     and merge it into the service info */
                    }

                else                                    /* otherwise no request is present */
                    outbound.data |= RF_TO_DEVICE (3);  /*   so merge in just the default device number */

                trace_1W (dptr, TRACE_CSIRQ, Trace_OBSI_Status,
                          WINDOW_ON (RD_DEVICE (outbound.data)),
                          outbound.data);
                break;


            case OBII:                                      /* Obtain Interrupt Information */
                outbound.data = RD_TO_CHANNEL (channel);    /* set the channel number */

                if (adcc->interrupts)                   /* if interrupts are present */
                    outbound.data |=                    /*   then get the highest priority device number */
                      RD_TO_DEVICE (LSB_NUMBER (adcc->interrupts));
                else                                    /* otherwise */
                    outbound.data |= RD_TO_DEVICE (7);  /*   merge in just the default device number */

                trace_1W (dptr, TRACE_CSIRQ, Trace_OBII_Status,
                          WINDOW_ON (RD_DEVICE (outbound.data)),
                          outbound.data);
                break;


            case IPOLL:                                 /* Interrupt Poll */
                if (adcc->interrupts && adcc->irq_mask) /* if the request is still pending and enabled */
                    outbound.data = HP_BIT (channel);   /*   then assert the channel bit */
                break;


            case SMSK:                                  /* Set Interrupt Mask */
                adcc->irq_mask = inbound.data           /* mask to our channel bit */
                                  & HP_BIT (channel);   /*   and save the result */
                break;
            }

        break;                                          /* end of I/O Read and Write cases */


    case IMB_Read:                                      /* not used by this interface */
    case IMB_Write:                                     /* not used by this interface */
    case IMB_RW_Ones:                                   /* not used by this interface */
    case IMB_MCS:                                       /* not used by this interface */
        break;
    }                                                   /* all defined cases are handled */

if (adcc->interrupts && adcc->irq_mask)                 /* if an interrupt request is pending and not masked */
    outbound.signals |= IRQ;                            /*   then assert IRQ on the IMB */

if (adcc->service_requests && adcc->csrq_enable)        /* if a service request is pending and enabled */
    if (dptr->flags & DEV_CPP)                          /*   then if switch S2 is in the CPP position */
        outbound.signals |= CSRQ2;                      /*     then assert CSRQ2 on the IMB */
    else                                                /*   otherwise */
        outbound.signals |= CSRQ1;                      /*     assert CSRQ1 on the IMB */

return outbound;                                        /* return the outbound signals and data */
}


/* ADCC reset.

   This routine is called for a RESET or RESET ADCC user command, when an IOCL
   or INIT command is received over the Intermodule Bus, or when the SRST, ~PON,
   or PFW IMB signals are asserted.  A reset performs the following hardware
   actions:

     - Clears Register C (Interrupt Register)
     - Clears the Interface Control (modifier 2) register
     - Clears the Service Condition Mask (modifier 3) register
     - Clears the New Status register
     - Clears the Program Run State register
     - Clears the modifier latch (but not the address latch)
     - Clears the Interrupt Mask flip-flop

   Three additional indirect actions are performed as a result of a reset:

     - CSRQ assertion is enabled when the modifier latch is cleared, because an
       even modifier value enables the signal

     - The occupancy flag (DR) for the UART Receiver Holding Register is cleared
       when the Service Condition Mask register is cleared, because Condition 2
       (receiver buffer is occupied) qualifies the active-low Data Register
       Reset signal.

     - Dataset ports are disconnected, because clearing the Interface Control
       register drops DTR to the datasets, causing them to hang up.

   The presence of the "-P" switch in a RESET command indicates that this is a
   power-on reset.  This returns the simulation back to its initial startup
   state.  It performs these simulation actions:

     - Clears the ADCC state structure, including the state of all eight ports
     - Restores the original FASTTIME setting
     - Saves the ADCC instance ID in each of the associated units
     - Initializes each multiplexer port
     - Cancels all scheduled port events
     - Resets each port's data mask
     - Disables explicit modem control on each multiplexer port
     - Clears the UNIT_MODEM flag from each unit

   Finally, if the ADCC is enabled and it hosts the system console or has a
   listening port attached, the poll unit is resynchronized with the keyboard
   poll.  Otherwise, the poll unit event service is canceled.


   Implementation notes:

    1. To synchronize polling, the event must be activated absolutely, as a
       service event may already be scheduled, and normal activation will not
       disturb an existing event.

    2. Because port 0 of an enabled ADCC assigned to IMB channel 1 handles the
       system console, we must schedule the poll service regardless of whether
       the poll unit is attached to a listening port.  This also necessitates
       the use of a local device ENABLED/DISABLED handler, because the global
       SCP handler will reject the disable call while the poll unit is
       scheduled.

    3. We initialize each multiplexer port by calling tmxr_init_line, which sets
       the line into the idle state.  It also enables transmission on the line,
       which is irrelevant for all ports except port 0 when it is being used for
       the system console.  In that case, it is necessary because console I/O is
       not using the mux library, and so the value won't be set automatically
       when the line is connected.  If this is not done, the line service
       routine would consider console transmission to be stalled, and the output
       buffer would be flushed after every character in a vain attempt to
       restore transmission.
*/

static t_stat adcc_reset (DEVICE *dptr)
{
DIB        *const dibptr  = (DIB *) dptr->ctxt;         /* the pointer to the device context */
const uint32      adcc_id = dibptr->card_index;         /* the instance ID */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the pointer to the ADCC state */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
UNIT       *const pptr    = &dptr->units [POLL];        /* the per-instance poll unit pointer */
uint32            unit;
UNIT              *uptr;
PORT_STATE        *port;

if (sim_switches & SWMASK ('P')) {                      /* if this a power-on reset */
    memset (adcc, 0, sizeof (ADCC_STATE));              /* clear the ADCC local state */
    adcc->fast_time = FAST_IO_TIME;                     /*   and reset the fast scheduling time */

    adcc->trigger_modes = Default_Mode;                 /* clear the trigger modes */

    adcc->trigger_input_ports  = BUS_MASK;
    adcc->trigger_output_ports = BUS_MASK;
    adcc->window_ports         = BUS_MASK;              /*   and open the trace window to all ports */

    for (unit = 0; unit < TERM_COUNT; unit++) {         /* loop through the multiplexer ports */
        uptr = &dptr->units [unit];                     /*   and get the unit pointer */
        port = &adcc->ports [unit];                     /*     and the port state pointer */

        uptr->Instance = adcc_id;                       /* save the instance ID */

        tmxr_init_line (&mptr->ldsc [unit]);            /* initialize the port */
        port->uart_data_mask = BITS (4, 0);             /*   and reset the data mask to its initial value (5 bits) */

        sim_cancel (uptr);                              /* cancel any scheduled event */
        uptr->wait = 0;                                 /*   and any pending ACK wait */

        tmxr_control_line (&mptr->ldsc [unit], Reset_Control);  /* disable modem control */
        uptr->flags &= ~UNIT_MODEM;                             /*   and clear the data set flag */
        }

    pptr->Instance = adcc_id;                           /* set the poll unit instance ID */
    }

adcc->interrupts       = 0;                             /* clear Register C (Interrupt Register) */
adcc->running          = 0;                             /* clears the Program Run State register */
adcc->modifier         = 0;                             /* clear the modifier latch */
adcc->irq_mask         = 0;                             /* clear the Interrupt Mask flip-flop */
adcc->service_requests = 0;                             /* clear the New Status register */
adcc->csrq_enable      = CSRQ_SEC_EN (0);               /*   which enables CSRQ assertion */

for (unit = 0; unit < TERM_COUNT; unit++) {             /* loop through the multiplexer ports */
    port = &adcc->ports [unit];                         /*   and get the port state pointer */

    if (dptr->units [unit].flags & UNIT_MODEM) {        /* if this is a modem port */
        if (port->modem_control & S2_CD)                /*   then setting DTR down will disconnect the port */
            trace_1W (dptr, TRACE_CMD, Trace_DTR_Drop, WINDOW_ON (unit), unit);

        tmxr_control_line (&mptr->ldsc [unit], No_Signals); /* drop the modem control signals */
        }

    port->modem_control = 0;                            /* clear the Interface Control (modifier 2) register */
    port->service_masks = 0;                            /* clear the Service Condition Mask (modifier 3) register */
    port->uart_rhr &= ~OCCUPIED;                        /*   which clears the RHR occupancy */
    }

if (not (dptr->flags & DEV_DIS)                                 /* if the device is enabled */
  && (dibptr->channel_number == 1 || pptr->flags & UNIT_ATT)) { /*   and the ADCC hosts the console or is attached */
    pptr->wait = hp_sync_poll (Initial);                        /*     then set up the poll */
    sim_activate_abs (pptr, pptr->wait);                        /*       and restart the poll timer */
    }

else                                                    /* otherwise */
    sim_cancel (pptr);                                  /*   cancel the poll */

return SCPE_OK;                                         /* reset always succeeds */
}


/* Attach the ADCC to a Telnet listening port or serial port.

   This routine is called by the ATTACH ADCC <netport> command to attach the
   listening port indicated by <netport> and by ATTACH ADCC<n> <serport> to
   attach line <n> to serial port <serport>.  For the former, it is logically
   the ADCC device that is attached; however, SIMH only allows units to be
   attached.  This makes sense for devices such as tape drives, where the
   attached media is a property of a specific drive.  In our case, though, the
   listening port is a property of the ADCC card, not of any given terminal
   line.  As ATTACH ADCC is equivalent to ATTACH ADCC0, the listening port
   would, by default, be attached to the first line and be reported there in a
   SHOW ADCC command.

   To preserve the logical picture, we attach the listening port to the poll
   unit (unit 8), which is normally disabled to inhibit its display.  Serial
   ports are attached to line units 0-7 normally, except that line 0 of the ADCC
   at IMB channel 1 is dedicated to the system console, which is attached by
   default to the simulation console.  Attachment is reported by the
   "adcc_show_status" routine.

   Attaching is not permitted if the ADCC is in diagnostic mode.  This prevents
   interference while the diagnostic is executing.


   Implementation notes:

    1. If we are being called as part of RESTORE processing, we may see a
       request to attach the poll unit (unit 8).  This will occur if the
       listening port was attached when the SAVE was done.  If the user attempts
       to attach unit 8 directly, the command will be rejected by SCP because
       the unit is disabled.

    2. If the poll unit is attached during a SAVE, it will be enabled as part of
       RESTORE processing.  We always unilaterally disable this unit to ensure
       that it remains hidden.

    3. ATTACH ADCC will pass a pointer to unit 0.  This is because SCP treats
       ATTACH ADCC as equivalent to ATTACH ADCC0.  The "tmxr_attach_unit"
       routine differentiates these cases by examining the "sim_ref_type" global
       to see if a device was referenced and, if so, applying the request to the
       poll unit.
*/

static t_stat adcc_attach (UNIT *uptr, char *cptr)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated adcc */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
UNIT       *const pptr    = &dptr->units [POLL];        /* the per-instance poll unit pointer */
t_stat            status;

if (dptr->flags & DEV_DIAG)                             /* if the multiplexer is in diagnostic mode */
    status = SCPE_NOFNC;                                /*   then the command is not allowed */
else                                                    /* otherwise */
    status = tmxr_attach_unit (mptr, pptr, uptr, cptr); /*   attempt to attach the designated unit */

if (status == SCPE_OK) {                                /* if the attach succeeded */
    pptr->wait = hp_sync_poll (Initial);                /*   then set up the poll */
    sim_activate_abs (pptr, pptr->wait);                /*     and start it */

    trace_1 (dptr, TRACE_PSERV, Trace_Poll_Schedule, pptr->wait);

    pptr->flags |= UNIT_DIS;                            /* ensure that the poll unit remains disabled */
    }

return status;                                          /* return the result of the attach */
}


/* Detach the ADCC.

   We are called by the DETACH ADCC command to detach the listening port and all
   Telnet sessions and by the DETACH ADCC<n> to detach a serial port from line
   <n>.  We will also be called by DETACH ALL, RESTORE, and during simulator
   shutdown.  For DETACH ALL and RESTORE, we must not fail the call, or
   processing of other units will cease.

   Detaching the listening port will disconnect all active Telnet connections
   but will not disturb any lines connected to serial ports.  Therefore, we look
   for the newly freed lines and clean them up.


   Implementation notes:

    1. During simulator shutdown, we will be called for units 0-7 (detach_all
       in scp.c calls the detach routines of all units that do NOT have
       UNIT_ATTABLE), as well as for unit 8 if it is attached.

    2. A RESTORE command will call us for unit 8 if it is attached.  In this
       case, the terminal channels will have already been rescheduled as
       appropriate, so canceling them is skipped.

    3. Because DETACH ADCC will pass unit 0, the "tmxr_detach_unit" routine
       examines the "sim_ref_type" global to see if a device was referenced and,
       if so, applies the request to the poll unit.

    4. Detaching the listening port resets each line first, which flushes the
       output buffer.  This ensures that buffered data that has not been output
       via a poll_service entry before the DETACH command was issued is written
       before the line is disconnected.
*/

static t_stat adcc_detach (UNIT *uptr)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated adcc */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
UNIT       *const pptr    = &dptr->units [POLL];        /* the per-instance poll unit pointer */
uint32            port;
t_stat            status;

status = tmxr_detach_unit (mptr, pptr, uptr);           /* detach the specified unit */

if (status == SCPE_OK) {                                /* if the detach succeeded */
    for (port = 0; port < TERM_COUNT; port++)           /*   then look for Telnet connections */
        if (tmxr_line_free (&mptr->ldsc [port])) {      /* if the line is now free */
            mptr->ldsc [port].rcve = 0;                 /*   then disable reception */
            sim_cancel (&dptr->units [port]);           /*     and cancel any transfer in progress */
            }

    if (tmxr_mux_free (mptr)) {                         /* if all lines are now free */
        sim_cancel (pptr);                              /*   then stop polling */

        trace_0 (dptr, TRACE_PSERV, Trace_Poll_Stopped);
        }
    }

return status;                                          /* return the status of the detach */
}


/* Adjust the device instances.

   This routine is called by SCP to adjust device instances that are newly
   allocated or pending deletion.  On entry, "dvptr" points at the "sim_devices"
   element that contains a pointer to the first of potentially several device
   instances to be adjusted, and "delta" contains the number of instances just
   created (if positive) or to be deleted (if negative).

   When the "dvptr" parameter is NULL, SCP expects the return value to be
   formatted as follows:

      31  30  ... 25  24 |23  22  ..  17  16 |15  14 13  ...  2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+
     | clone identifier  |  instance limit   |     sim_devices size     |
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+

   The fields have the following meanings:

     sim_devices size      The total number of elements defined in the
                           "sim_devices" array, including the NULL element at
                           the end.  The difference between this number and the
                           number of initialized elements gives the number of
                           new device instances that can be added by the user.

     instance limit        The maximum number of instances allowed.  The count
                           includes the original instance.

     clone identifier      A character used to produce unique names for new
                           device instances.  The ID will be appended to the
                           original device name and incremented by the device
                           instance number.

   When the adjust function is called with a non-NULL pointer parameter, the
   routine performs the actions are necessary to prepare the indicated device
   instances for use or destruction.  The return value is either SCPE_OK if all
   adjustments succeeded, or an SCPE error code indicating the nature of the
   failure.  In the latter case, any allocations made by the routine prior to
   the failure must be freed before returning.


   Implementation notes:

    1. The deallocation part of the routine is entered either for a deletion
       request or for an addition request when a required allocation fails.

    2. The ADCC requires per-instance allocation of the "adcc_order,"
       "adcc_ldsc," and "adcc_mod" arrays, and the "adcc_mdsc," "adcc_dib," and
       "adcc_state" structures.

    3. Before allocating any of the required structures, each new device's MTAB
       pointer is set to NULL.  This is so that failed allocations can be
       distinguished from the original value that points to a static array.

    4. The ADCC supports multiplexer connection groups by linking each of the
       TMXR structures in a circular linked list through the expansion pointer
       "exptr".  The list is built during allocation and torn down during
       deallocation.  When only the original device exists, its expansion
       pointer is NULL.

    5. The terminal configuration flags for port 0 are reset to those of port 1,
       as the port is not used for the system console on additional instances.
*/

static uint32 adcc_adjust (DEVICE **dvptr, int32 delta)
{
ADCC_STATE *nsptr;                              /* the pointer to the new ADCC state */
DEVICE     *dptr;                               /* the pointer to the new device structure */
DIB        *dibptr;                             /* the pointer to the new DIB structure */
TMXR       *mxptr;                              /* the pointer to the new multiplexer state */
TMLN       *lptr;                               /* the pointer to the new multiplexer line array */
int32      *optr;                               /* the pointer to the new multiplexer line order array */
REG        *rptr;                               /* the pointer to the new register array */
MTAB       *mptr;                               /* the pointer to the new modifier array */
char       *nptr;                               /* the pointer to the new device name */
int32      count;
uint32     index;
t_stat     status = SCPE_OK;

if (dvptr == NULL)                                      /* if this is a limits request */
    return TO_DWORD (TO_WORD ('A', MAX_ADCC_COUNT),     /*   then return the clone ID and instance limit */
                     SIM_DEVICE_SIZE);                  /*     and device array size */

else if (delta > 0) {                                   /* otherwise if device clones are being added */
    for (count = 0; count < delta; count++)             /*   then for each new device clone */
        dvptr [count]->modifiers = NULL;                /*     clear its modifier table pointer */

    for (count = 0; count < delta; count++) {           /* for each new device clone */
        dptr = dvptr [count];                           /*   get a pointer to the new device */

        status = SCPE_INCOMP;                           /* preset for name duplication error recovery */

        nptr = dptr->name;                              /* save the newly assigned name */
        dptr->name = "";                                /*   and temporarily remove it */

        if (find_dev (nptr) != NULL) {                      /* if the device name is duplicated */
            cprintf ("Duplicate device name %s\n", nptr);   /*   then report it to the user */

            dptr->name = nptr;                          /* restore the assigned name */
            break;                                      /*   and fail at this point */
            }

        else                                            /* otherwise the name is unique */
            dptr->name = nptr;                          /*   so restore it to the device */

        status = SCPE_MEM;                              /* preset for memory error recovery */

        dptr->units [0].flags = dptr->units [1].flags;  /* set port 0 to the set of flags used by port 1 */


        dibptr = malloc (sizeof (DIB));                 /* allocate a new DIB structure */

        if (dibptr == NULL)                             /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        dptr->ctxt = dibptr;                            /* save the new DIB pointer */

        memcpy (dibptr, &adcc_dib, sizeof (DIB));       /* copy the original DIB structure */
        dibptr->card_index = adcc_count + count;        /*   and set the new card index */


        mxptr = calloc (1, sizeof (TMXR));              /* allocate and zero a new TMXR structure */

        if (mxptr == NULL)                              /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        mxptr->dptr = dptr;                             /* set the device pointer */
        mxptr->lines = TERM_COUNT;                      /*   and the count of mux ports */

        mxptr->exptr = adccs [0].mptr;                      /* point my mux pointer at the first entry */
        adccs [adcc_count + count - 1].mptr->exptr = mxptr; /*   and the prior entry's pointer at me */

        adccs [adcc_count + count].mptr = mxptr;        /* save the new multiplexer pointer */


        lptr = calloc (1, sizeof adcc_ldsc);            /* allocate and zero a new TMLN array */

        if (lptr == NULL)                               /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        mxptr->ldsc = lptr;                             /* save the new line array pointer */


        optr = malloc (sizeof adcc_order);              /* allocate a new line order array */

        if (optr == NULL)                               /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        mxptr->lnorder = optr;                          /* save the new line order array pointer */

        for (index = 0; index < TERM_COUNT; index++)    /* initialize the line order array */
            *optr++ = index;                            /*   to 0-7 linearly */


        nsptr = calloc (1, sizeof (ADCC_STATE));        /* allocate and zero a new ADCC state structure */

        if (nsptr == NULL)                              /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        adccs [adcc_count + count].sptr = nsptr;        /* save the new state pointer */
        adccs [adcc_count + count].dptr = dptr;         /*   and device pointer in the table */


        mptr = malloc (sizeof adcc_mod);                /* allocate a new modifier table */

        if (mptr == NULL)                               /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        dptr->modifiers = mptr;                         /* save the new modifiers pointer */

        memcpy (mptr, adcc_mod, sizeof adcc_mod);       /* copy the original modifier array */

        do {                                                                /* loop through the modifiers */
            if (mptr->desc != NULL)                                         /* if the descriptor is a pointer */
                if (mptr->expander > 0)                                     /*   then if the entry is expanded */
                    mptr->desc = (char *) mptr->desc - (char *) &adcc_dib   /*     then relocate it to point */
                                   + (char *) dibptr;                       /*       into the new DIB */
                else                                                        /*   otherwise */
                    mptr->desc = (char *) mptr->desc - (char *) &adcc_mdsc  /*     relocate it to point */
                                   + (char *) mxptr;                        /*       into the new TMXR */

            if (mptr->valid == tmxr_set_lnorder)        /* if this is the LINEORDER entry */
                mptr->match = 0;                        /*   then set the line lower bound to 0 */
            }
        while ((++mptr)->mask != 0);


        for (rptr = dptr->registers; rptr->name != NULL; rptr++)    /* loop through the register entries */
            rptr->loc = (char *) rptr->loc - (char *) &adcc_state   /*   to get the location offsets */
                          + (char *) nsptr;                         /*     and relocate them to the new state */

        status = SCPE_OK;                               /* this pass succeeded */
        }                                               /* end of the for loop */

    adcc_count = adcc_count + delta;                    /* bump the count of device instances */
    }


if (delta < 0 || status != SCPE_OK) {                   /* if device clones are being deallocated */
    if (delta < 0) {                                    /*   then if existing entries are to be deleted */
        delta = - delta;                                /*     then make the deletion count positive */

        if (trace.status != Trace_Unallocated           /* if trace memory */
          && trace.status != Trace_Empty)               /*   is populated */
            for (count = 0; count < delta; count++)     /*     then for each device clone to be deleted */
                if (dvptr [count]->dctrl != 0)          /*       if it is configured for storing */
                    return SCPE_NOFNC;                  /*         then deallocation is not allowed */
        }

    adcc_count = adcc_count - delta;                    /* drop the count of device instances */

    if (adcc_count == 1)                                        /* if only the original instance remains */
        adccs [0].mptr->exptr = NULL;                           /*   then clear the connection group pointer */
    else                                                        /* otherwise */
        adccs [adcc_count - 1].mptr->exptr = adccs [0].mptr;    /*   set the prior entry's pointer to the first entry */

    for (count = 0; count < delta; count++) {           /* for each device clone to be deleted */
        dptr = dvptr [count];                           /*   get the device pointer */

        if (dptr != NULL) {                             /* if it is defined */
            free (dptr->ctxt);                          /*   then free the device context */
            free (dptr->modifiers);                     /*     and the modifier array */
            }

        if (adccs [adcc_count + count].sptr != NULL)    /* if the ADCC state was allocated */
            free (adccs [adcc_count + count].sptr);     /*   then free it */

        if (adccs [adcc_count + count].mptr != NULL) {  /* if the mux state pointer is valid */
            mxptr = adccs [adcc_count + count].mptr;    /*   then get it */

            free (mxptr->ldsc);                         /* free the line descriptor array */
            free (mxptr->lnorder);                      /*   and the line order array */
            free (mxptr);                               /*     and the mux state structure */
            }

        adccs [adcc_count + count].dptr = NULL;         /* clear the */
        adccs [adcc_count + count].sptr = NULL;         /*   table entries */
        adccs [adcc_count + count].mptr = NULL;         /*     for the deleted instance */
        }
    }

return (uint32) status;                                 /* return the adjustment status */
}


/* Set up a new connection.

   This validation routine is called to initialize a new multiplexer port
   connection, optionally waiting for the line to connect.  On entry, "uptr"
   points at any UNIT structure of the device to connect (the actual UNIT
   specified is irrelevant), and "desc" is non-NULL if a connection wait is
   desired or is NULL if the connection has already been established.  For the
   former case, the "line" parameter is ignored.  For the latter, "line"
   indicates the number of the multiplexer line to be initialized.  The "cptr"
   parameter is not used.

   A connection wait is typically requested with a SET <dev> CONNECT command
   whose MTAB entry specifies this validation routine.  In this case, the
   routine waits until a connection to the listening socket is established.  The
   user may abort the wait by entering the SCP interrupt character, which
   defaults to CTRL+E.  Connections that are established via the poll service
   routine pass NULL for the "desc" parameter to indicate that the wait is not
   needed.
*/

static t_stat adcc_connect (UNIT *uptr, int32 line, char *cptr, void *desc)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated adcc */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the pointer to the ADCC state */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
t_stat            status;

if (desc != NULL)                                       /* if a connection wait is requested */
    status = tmxr_connect (mptr, &line);                /*   then wait for the incoming request */
else                                                    /* otherwise the "line" parameter */
    status = SCPE_OK;                                   /*   already contains the new line number */

if (status == SCPE_OK) {                                /* if connection is established */
    mptr->ldsc [line].rcve = 1;                         /*   then enable the line to receive */

    trace_1W (dptr, TRACE_CMD, Trace_Connected, WINDOW_ON (line), line);
    }

return status;                                          /* return the connection status */
}


/* Add or remove the extender.

   For an expanded MTAB entry, test that the target location can be set to
   value, which will be either the value from the match field of the entry or
   the value parsed from the command line after any specified biasing.  Either
   uptr or desc will point at the associated UNIT or DEVICE structure,
   respectively, depending on whether a unit or device was specified in the
   command; the unused parameter will be set to NULL.  cptr points to the value
   portion of the parameter string; if cptr is NULL, no value was given.


   Implementation notes:

    1. We reset the connection line order to the new range of ports by calling
       the "tmxr_set_lnorder" routine and passing the command line "ALL".  This
       must be a variable array rather than a constant string because the
       routine appends a semicolon to the end of the string for ease in parsing.
*/

static t_stat adcc_extend (UNIT *uptr, int32 value, char *cptr, void *desc)
{
DEVICE *const dptr    = (DEVICE *) desc;        /* the pointer to the device structure */
DIB    *const dibptr  = (DIB *) dptr->ctxt;     /* the pointer to the device context */
const uint32  adcc_id = dibptr->card_index;     /* the instance ID */
TMXR   *const mptr    = adccs [adcc_id].mptr;   /* the per-instance multiplexer pointer */
char          all [8] = "ALL";                  /* keyword + room for ';' + room for NUL */
uint32        port;
int32         lower_bound;

if (value == DEV_EXT) {                                 /* if the extender is being added */
    for (port = MAIN_COUNT; port < TERM_COUNT; port++)  /*   then enable the ports */
        dptr->units [port].flags &= ~UNIT_DIS;          /*     that reside on the extender */

    mptr->lines = 8;                                    /* set the new number of mux ports */
    }

else {
    for (port = MAIN_COUNT; port < TERM_COUNT; port++) {    /* otherwise */
        uptr = &dptr->units [port];                         /*   get a pointer to the next extender unit */

        sim_cancel (uptr);                              /* cancel any port activity */

        if (uptr->flags & UNIT_ATT)                     /* if a serial port is attached to the unit */
            tmxr_detach_line (mptr, uptr);              /*   then detach it */
        else                                            /* otherwise */
            tmxr_disconnect_line (&mptr->ldsc [port]);  /*   disconnect the Telnet connection */

        uptr->flags |= UNIT_DIS;                        /* disable the extender port */
        }

    mptr->lines = 4;                                    /* set the smaller number of ports */
    }

if (dibptr->channel_number == 1)                        /* if this is the ADCC hosting the console */
    lower_bound = 1;                                    /*   then the first available port is 1 */
else                                                    /* otherwise */
    lower_bound = 0;                                    /*   all ports are available for connection */

tmxr_set_lnorder (uptr, lower_bound, all, mptr);        /* change the line order to reflect the port count */

return SCPE_OK;                                         /* return with success */
}


/* Set the diagnostic mode.

   This validation routine is called to install or remove the diagnostic
   loopback connector on the ADCC card.  On entry, "value" is either DEV_DIAG to
   install the connector and 0 to remove it.  The "uptr", "cptr", and "desc"
   parameters are not used.

   The diagnostic connector replaces the connector and cable to the RS-232
   terminal connectors, effectively disconnecting all ports.  In simulation,
   setting the diagnostic mode disconnects all Telnet and serial connections to
   avoid interference during the diagnostic run.


   Implementation notes:

    1. To detach the listening port, we fake a DETACH ADCC by setting the
       reference type to "device" and passing unit 0.  This disconnects all
       Telnet sessions, but any existing serial connections are undisturbed.
       So we then detach each line individually, as any still-connected ports
       must be serial connections.
*/

static t_stat adcc_diag (UNIT *uptr, int32 value, char *cptr, void *desc)
{
DEVICE *const dptr    = (DEVICE *) desc;        /* the pointer to the device structure */
DIB    *const dibptr  = (DIB *) dptr->ctxt;     /* the pointer to the device context */
const uint32  adcc_id = dibptr->card_index;     /* the instance ID */
TMXR   *const mptr    = adccs [adcc_id].mptr;   /* the per-instance multiplexer pointer */
uint32 port;

if (value == DEV_DIAG) {                                /* if we are installing the connector */
    sim_ref_type = REF_DEVICE;                          /*   then detach the poll unit */
    adcc_detach (dptr->units);                          /*     and any existing Telnet connections */

    for (port = 0; port < TERM_COUNT; port++) {         /* loop through the ports */
        tmxr_detach_line (mptr, &dptr->units [port]);   /*   and detach any remaining serial connections */

        mptr->ldsc [port].rcve = 0;                     /* disable reception on the port */
        sim_cancel (&dptr->units [port]);               /*   and cancel any transfer in progress */
        }
    }

return SCPE_OK;                                         /* the routine always succeeds */
}


/* Enable or disable the ADCC.

   This validation routine is entered with "value" set to 1 for an ENABLE and 0
   for a DISABLE, and "cptr" pointing to the next character after the keyword.
   If the ADCC is already enabled or disabled, respectively, the routine returns
   with no further action.  Otherwise, if "value" is 1, the device is enabled by
   clearing the DEV_DIS flag.  If "value" is 0, a check is made to see if the
   ADCC is listening for connections or has a port connected to a serial line.
   If either is TRUE, the disable request is rejected; the Telnet or serial
   client must be disconnected first.  Otherwise, the device is disabled by
   setting the DEV_DIS flag.

   In either case, the device is reset, which will restart or cancel the poll,
   as appropriate.
*/

static t_stat adcc_set_endis (UNIT *uptr, int32 value, char *cptr, void *desc)
{
TMXR   *const mptr = (TMXR *) desc;             /* the per-instance multiplexer pointer */
DEVICE *const dptr = mptr->dptr;                /* the pointer to the device structure */
uint32        port;

if (value)                                              /* if this is an ENABLE request */
    if (dptr->flags & DEV_DIS)                          /*   then if the device is disabled */
        dptr->flags &= ~DEV_DIS;                        /*     then reenable it */
    else                                                /*   otherwise the device is already enabled */
        return SCPE_OK;                                 /*     so there's nothing to do */

else                                                    /* otherwise this is a DISABLE request */
    if (dptr->flags & DEV_DIS)                          /*   so if the device is already disabled */
        return SCPE_OK;                                 /*     so there's nothing to do */

    else if (dptr->units [POLL].flags & UNIT_ATT)       /*   otherwise if the poll unit is still attached */
        return SCPE_NOFNC;                              /*     then report that the command failed */

    else {                                              /*   otherwise */
        for (port = 0; port < TERM_COUNT; port++)       /*     loop through the multiplexer ports */
            if (dptr->units [port].flags & UNIT_ATT)    /*       and if there's a serial connection */
                return SCPE_NOFNC;                      /*         then reject the command */

        dptr->flags |= DEV_DIS;                         /* all clear, so disable the device */
        }

return adcc_reset (dptr);                               /* reset the ADCC and restart or cancel polling */
}


/* Show the ADCC status.

   The attachment condition and connection count are printed to the stream
   specified by "st" as part of the ADCC device display.  The unit pointer,
   value, and descriptor pointer parameters are not used.
*/

static t_stat adcc_show_status (FILE *st, UNIT *uptr, int32 value, void *desc)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated adcc */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */

if (dptr->units [POLL].flags & UNIT_ATT)                /* if the poll unit is attached */
    fprintf (st, "attached to port %s, ",               /*   then report it */
             dptr->units [POLL].filename);              /*     with the listening port number */
else                                                    /* otherwise */
    fprintf (st, "not attached, ");                     /*   report the condition */

tmxr_show_summ (st, uptr, value, mptr);                 /* also report the count of connections */

return SCPE_OK;
}



/************************************
 *                                  *
 * Interface local utility routines *
 *                                  *
 ************************************/


/* Read an ADCC register.

   This routine reads a specified register from the ADCC and returns the result
   as a structure containing outbound data and signals fields.  On entry,
   "adcc_id" contains the ADCC instance identifier and "Register" contains the
   identifier of the register to read.

   Only Registers 0, 2, and E of the GIC set are implemented.  The others return
   a zero value.


   Implementation notes:

    1. A read of Register 0 returns the UART receiver holding register if the
       port is addressed to talk or a dummy zero value if the port is addressed
       to listen.  If an Amigo Identify sequence is in progress, then the read
       returns ID byte 1 and then ID byte 2 on successive calls.
*/

static IMB_OUTBOUND read_register (uint32 adcc_id, REGISTER Register)
{
DEVICE     *const dptr     = adccs [adcc_id].dptr;      /* the per-instance device pointer */
ADCC_STATE *const adcc     = adccs [adcc_id].sptr;      /* the per-instance ADCC state pointer */
IMB_OUTBOUND      outbound = { 0, ADN | DDN | PRO };    /* the initial outbound value and signal set */

switch (Register) {                                     /* dispatch the register read */

    case Reg_0:                                         /* FIFO Inbound */
        if (adcc->identify == Untalk)                   /* if an Identify sequence is not in progress */
            outbound.data = read_data (adcc_id);        /*   then return the UART data */

        else if (adcc->identify == ID_Byte_1) {                 /* otherwise if ID byte 1 is read */
            if (adcc->modifier <= 3 || dptr->flags & DEV_EXT)   /*   then if a main port or the extender is present */
                outbound.data = UPPER_BYTE (AMIGO_ID);          /*     then return the ID byte */
            else                                                /*   otherwise */
                outbound.signals |= DNV;                        /*     assert DNV for the absent port */

            adcc->identify = ID_Byte_2;                 /* read byte 2 next */
            }

        else if (adcc->identify == ID_Byte_2) {                 /* otherwise if ID byte 2 is read */
            if (adcc->modifier <= 3 || dptr->flags & DEV_EXT)   /*   then if a main port or the extender is present */
                outbound.data = LOWER_BYTE (AMIGO_ID);          /*     then return the ID byte */
            else                                                /*   otherwise */
                outbound.signals |= DNV;                        /*     assert DNV for the absent port */

            adcc->identify = ID_Byte_1;                 /* read ID byte 1 next if asked again */
            }
        break;


    case Reg_2:                                             /* Interrupt Conditions */
        outbound.data = R2_PENDING | R2_AVAIL | R2_EMPTY;   /* set the interrupt conditions */

        if (adcc->service_requests)                     /* if CSRQ is asserted */
            outbound.data |= R2_POLL;                   /*   then set the parallel poll assertion */
        break;


    case Reg_E:                                         /* Channel Configuration */
        outbound.data = HP_BIT (0) | ADCC_ID;           /* set the channel type to 1 (ADCC) */

        if (dptr->flags & DEV_EXT)                      /* if the extender is present */
            outbound.data |= RE_EXTENDER;               /*   then set the extender bit */

        if (dptr->flags & DEV_CPP)                      /* if switch S2 is set to the CPP position */
            outbound.data |= RE_CPP;                    /*   then set the CPP bit */
        break;


    default:                                            /* unimplemented registers */
        outbound.data = 0;                              /*   read as zero */
        break;
    }

trace_3 (dptr, TRACE_CSRW, Trace_Register,
         Read, Register, outbound.data);

return outbound;                                        /* return the register data and signals */
}


/* Write an ADCC register.

   This routine writes a specified ADCC register and returns the result as a
   structure containing an outbound data field, which is always zero, and a
   signals field.  On entry, "adcc_id" contains the ADCC instance
   identifier, "Register" contains the identifier of the register to write, and
   "data" contains the data value to be written.

   Only Registers 0 and C of the GIC set are implemented.  Writes to the other
   registers are ignored.

   Register 0 decodes the HP-IB Talk, Listen, and Secondary commands.
   Nominally, the primary addresses are 0-3 for the main and 4-7 for the
   extender, and the secondary indicates the function to perform.  However, the
   primary address field is only partially decoded, and some "out of range"
   addresses are used by the ADCC diagnostic to suppress CSRQ assertion.
   Specifically, CSRQ will be asserted only if the primary address is 0-15, or
   the primary address is odd, or the secondary address is 0-15, or the
   secondary address is even.

   To be stored in the address latch, a Talk or Listen command requires either
   bit 12 to be zero or bit 15 to be 1.

   Secondaries are always stored in the modifier latch, but modifier actions
   are disabled if bit 11 is 1 (i.e., if the secondary is DSJ).


   Implementation notes:

    1. The primary and secondary address conditions affect the CSRQ enable
       independently.  So, for example, a disabled primary followed by a
       disabled secondary followed by an enabled secondary must enable CSRQ
       assertion.  Therefore, separate primary and secondary enable bits are
       used, with either one enabling assertion.

    2. Because the ADCC is treated as a GIC by the CPP, it will receive PHI
       "byte transfer enable" commands when it is addressed to talk.  These
       appear as single-byte data writes to Register 0.  The GIC absorbs these
       writes internally, but the ADCC must ignore them, as otherwise they will
       corrupt the UART output.  We detect them by setting an "ignore" flag when
       a Talk is done as part of channel program execution and then skipping the
       next value written.  We cannot generally ignore data being written to the
       ADCC while it is in Talk mode because the diagnostic uses this invalid
       sequence during testing.

       The "ignore" flag can be static, rather than per-instance, because the
       byte transfer enable is sent as part of a single HP-IB transaction, so
       only a single instance and port is involved.
*/

static IMB_OUTBOUND write_register (uint32 adcc_id, REGISTER Register, IMB_DATA data)
{
static t_bool     ignore = FALSE;                   /* TRUE if ignoring a byte transfer enable */
DEVICE     *const dptr   = adccs [adcc_id].dptr;    /* the per-instance device pointer */
ADCC_STATE *const adcc   = adccs [adcc_id].sptr;    /* the per-instance adcc state pointer */
uint32            device;
IMB_OUTBOUND      outbound = { 0, ADN | DDN | PRO };    /* the initial outbound value and signal set */

trace_3 (dptr, TRACE_CSRW, Trace_Register,
         Write, Register, data);

switch (Register) {                                     /* dispatch the register write */

    case Reg_0:                                         /* FIFO Outbound */
        if ((data & R0_MODE_MASK) == R0_MODE_COMMAND) { /* if a bus command is present */
            switch (BUS_GROUP (data)) {                 /*   then dispatch on the command group */

                case BUS_TAG:                                           /* Talk Address Group */
                    ignore = adcc->running & BIT (BUS_ADDRESS (data));  /* set a flag if a channel program is running */

                /* fall through into Listen Address Group case */

                case BUS_LAG:                                       /* Listen Address Group */
                    if ((data & R0_QUALIFIER) != R0_INHIBITED) {    /* if the latch store is not inhibited */
                        adcc->address = R0_ADDRESS (data);          /*   then save the port address */
                        adcc->csrq_enable = CSRQ_PRI_EN (data);     /*     and conditionally set the primary enable bit */

                        adcc->address_group = BUS_GROUP (data); /* save the address group */

                        if (BUS_COMMAND (data) == BUS_UNTALK)   /* if this is an Untalk command */
                            adcc->identify = My_Secondary;      /*   then it may begin an Identify sequence */
                        else                                    /* otherwise */
                            adcc->identify = Untalk;            /*   Identify is waiting for an Untalk */
                        }
                    break;

                case BUS_SCG:                           /* Secondary Command Group */
                    if (BUS_ADDRESS (data) == DSJ) {    /* if the command is Device Specified Jump */
                        adcc->modifier = DSJ;           /*   then save the modifier */
                        adcc->csrq_enable |= BIT (1);   /*     and set the secondary enable bit */
                        }

                    else {                                      /* otherwise */
                        adcc->modifier = R0_MODIFIER (data);    /*   save the modifier */

                        adcc->csrq_enable &= ~BIT (1);              /* clear the secondary enable bit */
                        adcc->csrq_enable |= CSRQ_SEC_EN (data);    /*     and then conditionally set it */

                        if (adcc->identify == My_Secondary) {   /* if an Identify sequence is in progress */
                            adcc->identify = ID_Byte_1;         /*   then the ID is ready for reading */
                            ignore = TRUE;                      /*     and ignore any "byte transfer enable" next */
                            }
                        }
                    break;

                case BUS_PCG:                           /* Primary Command Group */
                    break;                              /* all primary commands are ignored */
                }
            }

        else {                                          /* otherwise this is a data write */
            if (not (ignore && data & HP_BIT (0)))      /*   so if it's not a "byte transfer enable" command */
                write_data (adcc_id, data);             /*     then write the data as directed by the secondary */

            ignore = FALSE;                             /* additional writes are accepted unconditionally */
            }
        break;


    case Reg_C:                                         /* Interrupt Register */
        device = RC_DEVICE (data);                      /* get the interrupting device number */

        if (data & RC_SET_IRQ)                          /* if an interrupt is requested */
            adcc->interrupts |= BIT (device);           /*   then set the device bit in C bit order */
        else                                            /* otherwise */
            adcc->interrupts &= ~BIT (device);          /*   clear the device bit */
        break;


    default:                                            /* unimplemented registers */
        break;                                          /*   are treated as read-only */
    }

return outbound;                                        /* return the result of the write */
}


/* Read a data word.

   This routine reads and returns the data word indicated by the previously set
   address and modifier latches.  On entry, "adcc_id" contains the instance
   identifier.  Only modifiers 0 (UART data), 1 (UART status), 2 (modem status),
   and %20 (DSJ) return values.  The other modifiers do not have implementations
   and return zero values.


   Implementation notes:

    1. Normally, the port would have to be addressed to talk to return data to
       the CPP.  However, the diagnostic ignores this protocol and reads while
       the port is addressed to listen.  So we test the "program is running" bit
       before enforcing this condition.

    2. The DRR (Data Received Reset) signal to the UART is asserted by reading
       with modifier 0, but only when the Receive Buffer Full service condition
       is enabled, due to an AND gate in the signal path.  If the condition is
       disabled, then reading does not clear the Receiver Holding Register, and
       the next character received will cause a data overrun error.
*/

static IMB_DATA read_data (uint32 adcc_id)
{
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the per-instance adcc state pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
const HP_WORD     address = adcc->address;              /* the port address */
TMLN       *const lptr    = &mptr->ldsc [address];      /* the multiplexer line pointer */
PORT_STATE *const port    = &adcc->ports [address];     /* the port state pointer */
UNIT       *const uptr    = &dptr->units [address];     /* the multiplexer unit pointer */
IMB_DATA          outbound;

switch (adcc->modifier) {                               /* dispatch on the modifier latch value */

    case 000:                                           /* UART Data */
        if (adcc->address_group == BUS_TAG              /* if the port is addressed to talk */
          || not (adcc->running & BIT (address))) {     /*   or the port is not running a channel program */
            outbound = port->uart_rhr;                  /*     then read the receiver holding register */

            if (port->service_masks & S3_DR) {          /* if Receive Buffer Full service is enabled */
                port->uart_rhr = 0;                     /*   then reset the register */
                port->uart_status &= ~S1_BREAK;         /*     which clears break status */

                if (tmxr_rqln (lptr) > 0)               /* if more characters are available */
                    activate_unit (uptr, Receive);      /*   then schedule reception */
                }

            trace_2W (dptr, TRACE_CMD, Trace_Modifier_0_Read, WINDOW_ON (address), address, outbound);

            check_service (adcc_id, address);           /* check for service requests */
            }

        else                                            /* otherwise the port is not talking */
            outbound = 0;                               /*   so return a dummy value */
        break;


    case 001:                                           /* UART Status */
        outbound = port->uart_status;                   /* read the UART status bits */

        if (port->uart_status & S1_BREAK)               /* if a BREAK was received */
            outbound |= S1_TAGGED;                      /*   then tag the status for special handling */
        else                                            /* otherwise a normal character was received */
            outbound |= S1_REGULAR;                     /*   so tag it for regular handling */

        if (port->uart_thr == 0                         /* if both the transmitter holding register */
          && port->uart_tr == 0                         /*   and the transmitter register are empty */
          && port->service_masks & S3_TE)               /*     and Transmitter Empty service is enabled */
            outbound |= S1_EMPTY;                       /*       then set the empty status bit */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_1_Read, WINDOW_ON (address), address, outbound);
        break;


    case 002:                                           /* Modem Status */
        if (not (dptr->flags & DEV_DIAG))               /* if the ADCC is not in diagnostic mode */
            update_status (adcc_id, address);           /*   then update the current line status */

        outbound = port->modem_status;                  /* read the modem status lines */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_2_Read, WINDOW_ON (address), address, outbound);
        break;


    case 020:                                           /* Device-Specified Jump */
        outbound = port->dsj;                           /* read the current jump value */

        trace_2W (dptr, TRACE_CMD, Trace_DSJ, WINDOW_ON (address), address, port->dsj);
        break;


    default:                                            /* unimplemented modifiers */
        outbound = 0;                                   /*   return zero values */
        break;
    }

return outbound;                                        /* return the data value read */
}


/* Write a data word.

   This routine writes the data word indicated by the previously set address and
   modifier latches.  On entry, "adcc_id" contains the instance identifier, and
   "data" contains the data word to write.  Only modifiers 0-7 accept data.
   Writes to other modifier addresses are ignored.

   Setting bit 8 in the UART Control register asserts Master Reset to the UARTs.
   This causes the following actions:

     - Clears the THR, RHR, and TR registers

     - Clears the Framing Error, Overrun Error, Parity Error, and Data Received
       status indicators

     - Sets the Transmitter Holding Register Empty and Transmitter Register
       Empty status indicators

     - Sets the Transmitter Register Output (serial data) to the marking (high)
       level


   Implementation notes:

    1. Asserting the UART Master Reset signal clears the TR (transmitter
       register), THR (transmitter holding register), and RHR (receiver holding
       register), clears the FE (framing error), OE (overrun error), PE (parity
       error), and DR (data received) signals, and sets the TRO (transmitter
       register output), THRE (transmitter holding register empty), and TRE
       (transmitter register empty) signals.

    2. With the diagnostic connector in place, the following modem output
       signals are connected to the modem input signals:

        - CA  (RTS)  to CB  (CTS) and CF (DCD)
        - CD  (DTR)  to CC  (DSR)
        - SCA (SRTS) to SCF (SDCD)
        - CH  (DRS)  to CE  (RI)

       The CH-to-CE connection is observable only if the extender is present.
*/

static void write_data (uint32 adcc_id, IMB_DATA data)
{
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the per-instance adcc state pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
const HP_WORD     address = adcc->address;              /* the port address */
PORT_STATE *const port    = &adcc->ports [address];     /* the per-instance port state pointer */
TMLN       *const lptr    = &mptr->ldsc [address];      /* the per-instance multiplexer line pointer */
TMCKT             control_set;

switch (adcc->modifier) {                               /* dispatch on the modifier latch value */

    case 000:                                           /* UART Data */
        if (port->uart_tr & OCCUPIED)                   /* if the transmitter register is occupied */
            port->uart_thr = S0_DATA (data) | OCCUPIED; /*   then write to the holding register */
        else                                            /* otherwise */
            port->uart_tr = S0_DATA (data) | OCCUPIED;  /*   write to the transmitter directly */

        activate_unit (&dptr->units [address], Send);   /* schedule the transmission */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_0_Write, WINDOW_ON (address), address, data);

        check_service (adcc_id, address);               /* check for service requests */
        break;


    case 001:                                           /* UART Control */
        port->uart_control = data;                      /* set the UART control bits */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_1_Write, WINDOW_ON (address), address, data);

        if (data & S1_RESET) {                          /* if a UART reset is requested */
            port->uart_thr = 0;                         /*   then clear the transmitter holding register */
            port->uart_tr = 0;                          /*     and the transmitter register */
            port->uart_rhr = 0;                         /*     and the receiver holding register */
            port->uart_rr = 0;                          /*     and the receiver register */
            port->uart_status = 0;                      /*     and the UART status bits */

            sim_cancel (&dptr->units [address]);        /* cancel any scheduled event service */
            dptr->units [address].wait = 0;             /*   and any pending ACK wait */
            }

        port->uart_data_mask = S1_DATA_MASK (data);     /* set the new data mask from the character length */
        break;


    case 002:                                           /* Interface Control */
        if (dptr->flags & DEV_DIAG) {                   /* if the diagnostic connector is present */
            port->modem_status = 0;                     /*   then loop the status outputs back to the inputs */

            if (data & S2_CA)                           /* if RTS is asserted */
                port->modem_status |= S2_CB | S2_CF;    /*   then it sets CTS and DCD status */

            if (data & S2_CD)                           /* if DTR is asserted */
                port->modem_status |= S2_CC;            /*   then it sets DSR status */

            if (data & S2_SCA)                          /* if SRTS is asserted */
                port->modem_status |= S2_SCF;           /*   then it sets SDCD status */

            if (data & S2_CH && dptr->flags & DEV_EXT)  /* if DRS is asserted and the extender is present */
                port->modem_status |= S2_CE;            /*   then it sets RI status */
            }

        else {                                          /* otherwise the terminal connector is present */
            control_set = No_Signals;                   /*   so initialize the control set */

            if (data & S2_CA)                           /* if CA is set */
                control_set |= RTS_Control;             /*   then assert the RTS signal */

            if (data & S2_CD)                           /* if CD is set */
                control_set |= DTR_Control;             /*   then assert the DTR signal */

            else if (port->modem_control & S2_CD)       /* setting DTR down will disconnect the port */
                trace_1W (dptr, TRACE_CMD, Trace_DTR_Drop, WINDOW_ON (address), address);

            if (control_set & S2_MODEM)                     /* if a modem control signal is present */
                dptr->units [address].flags |= UNIT_MODEM;  /*   then mark this as a modem port */

            if (dptr->units [address].flags & UNIT_MODEM)   /* if this is a modem port */
                tmxr_control_line (lptr, control_set);      /*   then set the port control lines as indicated */
            }

        port->modem_control = data;                     /* save the modem control signals */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_2_Write, WINDOW_ON (address), address, data);

        check_service (adcc_id, address);               /* check for service requests */
        break;


    case 003:                                                   /* Service Condition Masks */
        if (data & S3_ENABLE)                                   /* if the selected mask is enabled */
            port->service_masks |= BIT (S3_CONDITION (data));   /*   then set the corresponding condition bit */
        else                                                    /* otherwise */
            port->service_masks &= ~BIT (S3_CONDITION (data));  /*   clear the corresponding condition bit */

        if (not (port->service_masks & S3_DR))          /* if the Receiver Buffer Full condition is disabled */
            port->uart_rhr &= ~OCCUPIED;                /*   then it clears the Data Received indication */

        trace_3W (dptr, TRACE_CMD, Trace_Modifier_3_Write, WINDOW_ON (address),
                 address, data, port->service_masks);

        check_service (adcc_id, address);               /* check for service requests */
        break;


    case 004:                                           /* Clear Special Character RAM */
        port->special_ram [S4_DATA (data)] = 0;         /* clear the special flag for the specified character */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_4_Write, WINDOW_ON (address), address, data);
        break;


    case 005:                                           /* Set Special Character RAM */
        port->special_ram [S5_DATA (data)] = 1;         /* set the special flag for the specified character */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_5_Write, WINDOW_ON (address), address, data);
        break;


    case 006:                                               /* Set Baud Rates */
        if (data & S6_RECEIVER)                             /* if the receiver rate is specified */
            port->receive_delay = delays [S6_RATE (data)];  /*   then set it as indicated */
        else                                                /* otherwise */
            port->transmit_delay = delays [S6_RATE (data)]; /*   set the transmitter rate */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_6_Write, WINDOW_ON (address), address, data);
        break;


    case 007:                                           /* Status Reference/Mask */
        if (not (dptr->flags & DEV_EXT))                /* if the extender is not present */
            data &= ~S7_CE;                             /*   then remove the CE bit setting */

        if (data & S7_SET_MASK)                         /* if the mask is being set */
            port->status_mask = S7_VALUES (data);       /*   then save the status mask values */
        else                                            /* otherwise */
            port->status_ref = S7_VALUES (data);        /*   save the status reference values */

        trace_2W (dptr, TRACE_CMD, Trace_Modifier_7_Write, WINDOW_ON (address), address, data);

        check_service (adcc_id, address);               /* check for service requests */
        break;


    default:                                            /* unimplemented modifiers */
        break;                                          /*   have no effect */
    }

return;
}


/* Check for a service condition.

   This routine checks the UART and modem status, sets the DSJ value to reflect
   the status, and returns TRUE if a channel service request is wanted and FALSE
   if not.  Channel requests are enabled only if a channel program is running
   for the addressed port.  The DSJ value is set independently of the program
   state but is prioritized in the presence of multiple service conditions in
   the following order:

     2 = modem status has changed
     0 = receiver register is occupied
     1 = transmitter register is empty
     3 = service is not requested

   The first three conditions may be masked to allow or inhibit their reporting.

   DSJ 2 results from a modem status line differing from its reference state
   with its status mask enabled.  DSJ 2 also results from the UART detecting a
   parity error, overrun error, or break detection.

   DSJ 0 results from the reception of a character into the receiver holding
   register with its service mask enabled.  It persists until the register is
   read.

   DSJ 1 occurs when either the transmitter holding register is empty, or both
   it and the transmitter register are empty, depending on the setting of the
   corresponding service masks.

   If none of the conditions that would set DSJ are present, then DSJ 3 results.

   CSRQ, if enabled, is asserted for all DSJ conditions 0-2, except if DSJ 2 is
   caused by a parity or overrun error.


   Implementation notes:

    1. The UART service request bit for a given port address is (address * 2),
       which produces the correct service priority when extracted with the
       LSB_NUMBER macro.  See the implementation note for the ADCC state
       structure declaration for details.
*/

static t_bool check_service (uint32 adcc_id, uint32 address)
{
DEVICE     *const dptr = adccs [adcc_id].dptr;          /* the per-instance device pointer */
ADCC_STATE *const adcc = adccs [adcc_id].sptr;          /* the per-instance adcc state pointer */
PORT_STATE *const port = &adcc->ports [address];        /* the per-instance port state pointer */
t_bool            csrq;

csrq = ((adcc->running & BIT (address)) != 0);          /* enable CSRQ if the port is active */

if ((port->modem_status ^ port->status_ref) & port->status_mask /* if a masked port line changed */
  || port->uart_status & S1_BREAK)                              /*   or a break was detected */
    port->dsj = DSJ_STATUS;                                     /*     then indicate a status jump */

else if (port->uart_status & (S1_OVERRUN | S1_PARITY)) {    /* otherwise if an overrun or parity error occurred */
    port->dsj = DSJ_STATUS;                                 /*   then indicate a status jump */
    csrq = FALSE;                                           /*     but without asserting CSRQ */
    }

else if (port->uart_rhr & OCCUPIED                      /* otherwise if a character was received */
  && port->service_masks & S3_DR)                       /*   and the condition mask is set */
    port->dsj = DSJ_RECEIVE;                            /*     then indicate a reception jump */

else if (port->uart_thr == 0                            /* otherwise if the transmitter holding register is empty */
  && port->service_masks & S3_THRE                      /*   and the condition mask is set */
  || port->uart_thr == 0                                /* or both the transmitter holding register */
  && port->uart_tr == 0                                 /*   and the transmitter register are empty */
  && port->service_masks & S3_TE)                       /*   and the condition mask is set */
    port->dsj = DSJ_TRANSMIT;                           /*     then indicate a transmission jump */

else {                                                  /* otherwise */
    port->dsj = DSJ_IDLE;                               /*   indicate a port idle jump */
    csrq = FALSE;                                       /*     which does not assert CSRQ */
    }

if (csrq) {                                             /* if a request is active and enabled */
    adcc->service_requests |= UART_BIT (address);       /*   then set the UART service bit */

    trace_5W (dptr, TRACE_CSIRQ, Trace_Service_Condition, WINDOW_ON (address),
              address, port->service_masks, port->modem_status,
              port->status_ref, port->status_mask);

    trace_2W (dptr, TRACE_CSIRQ, Trace_CSRQ_Assert, WINDOW_ON (address), address, port->dsj);
    }

else                                                    /* otherwise */
    adcc->service_requests &= ~UART_BIT (address);      /*   clear the UART service bit */

return csrq;                                            /* return TRUE if service is requested */
}


/* Update the port status.

   This routine updates the stored RS-232 line status for the specified port.
   It obtains the status, either real for a serial connection or simulated for a
   Telnet connection, and updates the states in the port status word.


   Implementation notes:

    1. CE (Ring Indicator) status is available for ports 0-3 only when the
       extender is present.  This is because the extender contains the circuitry
       for all eight ports.

    2. SCF (Secondary Data Carrier Detect) is not available from the PC serial
       port.

    3. In support of serial printers, consideration was given to reporting the
       line state as "connected" if a port log file is attached.  If the port
       is configured as modem-connected in MPE, and output is attempted without
       a log file, then it would be suspended instead of lost.  When the log
       file was attached, output would resume automatically.

       However, if a user terminal session has an attached log file, and the
       Telnet client drops the connection, MPE would not log the user off
       because the port would still be seen as "connected."  Worse, another
       Telnet connection to the same port would resume within the prior session.
       Given that serial printers are a temporary expediency until an HP-IB
       printer simulation is available, this accommodation is rejected.
*/

static void update_status (uint32 adcc_id, uint32 address)
{
DEVICE     *const dptr = adccs [adcc_id].dptr;          /* the per-instance device pointer */
ADCC_STATE *const adcc = adccs [adcc_id].sptr;          /* the per-instance adcc state pointer */
TMXR       *const mptr = adccs [adcc_id].mptr;          /* the per-instance multiplexer pointer */
TMLN       *const lptr = &mptr->ldsc [address];         /* the per-instance multiplexer line pointer */
PORT_STATE *const port = &adcc->ports [address];        /* the per-instance port state pointer */
TMCKT             line_state;

line_state = tmxr_line_status (lptr);                   /* get the current line state */

port->modem_status &= S2_CC;                            /* clear the modem status lines except for DSR */

if (line_state & CTS_Status)                            /* if CTS is set */
    port->modem_status |= S2_CB;                        /*   then set the corresponding status bit */

if (line_state & DCD_Status)                            /* if DCD is set */
    port->modem_status |= S2_CF;                        /*   then set the corresponding status bit */

if (line_state & RI_Status && dptr->flags & DEV_EXT)    /* if RI is set and the extender is present */
    port->modem_status |= S2_CE;                        /*   then set the corresponding status bit */

if (line_state & DSR_Status) {                          /* if DSR is now set */
    if (not (port->modem_status & S2_CC))               /*   but was not previously */
        trace_1W (dptr, TRACE_CMD, Trace_DSR_Raise, WINDOW_ON (address), address);

    port->modem_status |= S2_CC;                        /* set the corresponding status bit */
    }

else {                                                  /* otherwise DSR is now clear */
    if (port->modem_status & S2_CC)                     /*   but if it was previously set */
        trace_1W (dptr, TRACE_CMD, Trace_DSR_Drop, WINDOW_ON (address), address);

    port->modem_status &= ~S2_CC;                       /* clear the corresponding status bit */
    }

return;
}


/* Multiplexer line service.

   The line service routine runs only when there are characters to read from or
   write to the port.  It is scheduled either at a realistic rate corresponding
   to the programmed baud rate of the port to be serviced, or at a somewhat
   faster optimized rate.  It is entered when the transmitter register is
   written, when the poll routine determines that there are characters ready for
   input, while input characters are being echoed, or while waiting for an ACK
   to complete an ENQ/ACK handshake.

   On entry, "uptr" points at the SIMH unit corresponding to the port that
   requires service.  Either the receiver register or the transmitter register
   will be occupied, corresponding to a character received or a character ready
   to transmit.  In both cases, the requisite delay corresponding to the
   established serial rate has elapsed, so both conditions are ready for
   immediate action.


   The routine first checks the receiver register for a character.  If one is
   not already present, then the terminal multiplexer library is called to
   retrieve a possibly waiting character.  This improves read performance by
   not waiting for a poll to occur.  If a valid character is now available, it
   is processed.

   If the receiver holding register is occupied, then an overrun error is
   indicated, as the prior character has not been read.  If the diagnostic mode
   is enabled, then two special conditions are checked.  The diagnostic
   generates a parity error by setting the transmitter baud rate faster than the
   receiver rate, and it generates a break by setting the receiver rate faster
   than the transmitter rate.  If either condition is detected, the appropriate
   status is set.

   If a break (real or simulated) is detected, break reception status is set,
   and reception service completes by clearing the receiver register.
   Otherwise, a valid character was received.  If CAPS LOCK is set, and
   diagnostic mode is disabled, then a lower case character is upshifted.

   If input echoing is enabled, then the received character is stored in the
   transmitter register if it is empty, or in the transmitter holding register
   if it is empty.  If both are occupied, then in hardware the echoed character
   would interfere with the character being transmitted.  In simulation, this is
   simulated by replacing the character in the transmitter register with a DEL
   as an indication to the user that a collision occurred.

   The received character is then stored in the receiver holding register.  If
   the character has been marked as special, the special status field in the
   register is set.  Otherwise, the field reflects reception of a regular
   character.

   Reception is completed by clearing the receiver register, as well as the wait
   for the ACK of an ENQ/ACK handshake.

   If no character was received during this event service call, the routine then
   checks to see if an ACK is expected in reply to an earlier ENQ.  If so, then
   the current wait time is doubled, and the service is rescheduled.  However,
   the new wait time is longer than the current poll time, service rescheduling
   is abandoned in favor of the normal poll for received characters.


   The transmitter register is then checked for a character to output.  If one
   is present, then the character is masked to the indicated character length.
   If diagnostic mode is enabled, the character is looped back to the receiver
   register, and the routine is recursively called to process the reception.
   Otherwise, if the character to be sent is an ENQ, and local acknowledgement
   is enabled, then an ACK is stored in the receiver register, and the reception
   is scheduled.

   Otherwise, the character is converted as directed by the output mode
   specified by the user.  If the conversion does not suppress transmission,
   then the character is sent to the connected terminal, which will be the
   simulation console if the port hosts the MPE system console, or to the user
   terminal on designated multiplexer line.

   If the character is an ENQ, or the transmission reports that the output
   buffer is full, the buffer is flushed to minimize the ENQ delay or to prepare
   for the next transmission.

   If transmission succeeded, then if the character is an ENQ, reception of the
   expected ACK is scheduled, and the ACK wait counter is set.  Otherwise if the
   port is not connected, then the status is reset to indicate a successful
   transmission; if a log file is not attached, the character is reported as
   lost silently (as it would in hardware).

   After a successful transmission, if a character is present in the transmitter
   holding register, then it is moved to the transmitter register, and
   transmission is rescheduled.  Otherwise, the transmitter register is cleared.

   If transmission failed because the multiplexer library transmit buffer was
   full, it is rescheduled to try again, and the error is suppressed.  Any other
   transmission error will cause a simulator stop when it is reported back to
   the caller.

   Finally, a check is made for service requests as a result of reception or
   transmission, and if indicated, CSRQ is asserted for the channel.


   Implementation notes:

    1. The "wait" fields of the port units are not used (i.e., are set to zero)
       except when ports are waiting for ACKs.  Because we want to retrieve the
       ACK as quickly as possible but also minimize the load on the host system,
       we initially wait the normal reception time (fast or realistic) and then
       double the wait each time it expires without reception.  Therefore, a
       non-zero "wait" value indicates that an ACK is expected.

    2. Receipt of any character cancels a pending ACK wait, even though it is
       possible that the character is not an ACK (for example, if the character
       was received but not processed before the ENQ was sent).

    3. During diagnostic operation with the loopback connector in place, and
       assuming that the UART hardware is good, parity errors and breaks cannot
       occur because the UART controls both transmission and reception.  So the
       diagnostic tests the detection ability by operating the transmit and
       receive sides of the UART at different rates and sending a byte whose
       value will cause the indicated conditions on reception.  To test parity
       error detection, the transmitter is set to a faster baud rate (smaller
       delay) than the receiver.  To test break detection, the transmitter is
       set to a slower baud rate (larger delay) than the receiver.  We use the
       rate disparities to fake parity errors and breaks for the diagnostic.

    4. Parity errors cannot occur in normal operation because (a) host serial
       ports strip off the parity bit and provide only the data, and (b) Telnet
       does not use parity.

    5. Characters that are looped back by the diagnostic connector must be
       received by a recursive call to this routine instead of by scheduling the
       event.  This is because loopback reception in hardware occurs
       simultaneously with transmission, and when they complete together,
       "receiver full" service notification has priority over "transmitter
       empty" notification.  If reentry was scheduled, CSRQ assertion with DSJ 1
       for the transmission would occur first.

    6. Output characters are normally buffered by the multiplexer library and
       are sent as a burst by the poll event service routine.  However, if we
       are waiting for the ACK of an ENQ/ACK handshake, then the output buffer
       is immediately flushed to reduce the waiting time for the reply.  The
       buffer is also flushed if the console or port transmission routine
       indicates that the buffer is full.  This buffer flush will not affect a
       stall on the console port, as it has its own internal line structure, but
       it also handles its own buffer flush automatically, so our flush is
       irrelevant.

    7. When rescheduling transmission for a stall condition, the transmit buffer
       has already been flushed, so reentry should find an empty buffer.

    8. Asynchronously asserting CSRQ only makes sense when asserting CSRQ1, as
       that's the only signal to which the CPP responds.  Therefore, we need not
       check the CPP switch setting before asserting.
*/

static t_stat line_service (UNIT *uptr)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated ADCC */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the per-instance ADCC state pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
DIB        *const dibptr  = (DIB *) dptr->ctxt;         /* the pointer to the device context */
const HP_WORD     address = (HP_WORD) (uptr - dptr->units); /* the port address */
TMLN       *const lptr    = &mptr->ldsc [address];      /* the per-instance multiplexer line pointer */
PORT_STATE *const port    = &adcc->ports [address];     /* the per-instance port state pointer */
int32             chr, poll_wait;
uint8             data_byte;
t_stat            status  = SCPE_OK;

trace_1W (dptr, TRACE_SERV, Trace_Port_Service, WINDOW_ON (address), address);


/* Reception service */

if (not (port->uart_rr & OCCUPIED)) {                   /* if the receiver register is empty */
    if (uptr->wait)                                     /*   then if the port is waiting for an ACK */
        tmxr_poll_rx (mptr);                            /*     then poll the line to see if it has arrived */

    port->uart_rr = tmxr_getc_ln (lptr);                /* see if there's now a character ready */

    if (port->uart_rr & TMXR_VALID)                     /* if a character was returned */
        port->uart_rr |= OCCUPIED;                      /*   then mark the register as occupied */
    }

if (port->uart_rr & OCCUPIED) {                         /* if the receiver register is occupied */
    port->uart_status = 0;                              /*   then reset the receiver status */

    if (port->uart_rhr & OCCUPIED) {                    /* if the receiver holding register is occupied */
        port->uart_status |= S1_OVERRUN;                /*   then indicate a reception overrun */

        trace_1W (dptr, TRACE_XFER, Trace_Overrun, WINDOW_ON (address), address);
        }

    if (dptr->flags & DEV_DIAG) {                               /* if diagnostic mode is enabled */
        if (port->transmit_delay < port->receive_delay)         /*   and the transmitter is faster */
            port->uart_status |= S1_PARITY;                     /*     then fake a parity error */

        else if (port->transmit_delay > port->receive_delay)    /* otherwise if the receiver is faster */
            port->uart_rr = SCPE_BREAK;                         /*   then fake a BREAK */
        }

    if (port->uart_rr & SCPE_BREAK) {                   /* if a break is detected */
        port->uart_status |= S1_BREAK;                  /*   then indicate a break */
        port->uart_rhr = S0_SPECIAL;                    /*     and set the special bit */

        trace_1W (dptr, TRACE_XFER, Trace_Break, WINDOW_ON (address), address);
        }

    else {                                              /* otherwise it's a normal reception */
        port->uart_rr = S0_DATA (port->uart_rr);        /*   so mask it to the data character */

//      trace_input (dptr, TRACE_XFER, Trace_Reception, address, port->uart_rr);
        trace_2T (dptr, TRACE_XFER, Trace_Reception,
                  TRIGGER_ON_INPUT (address, port->uart_rr),
                  WINDOW_ON (address),
                  address, port->uart_rr);

        if (uptr->flags & UNIT_CAPSLOCK && not (dptr->flags & DEV_DIAG))    /* if CAPS LOCK and not diagnostic */
            port->uart_rr = (uint32) toupper (port->uart_rr);               /*   then convert to upper case if lower */

        if (port->modem_control & S2_ECHO)                  /* if echoing is enabled */
            if (not (port->uart_tr & OCCUPIED))             /*   then if the transmitter register is empty */
                port->uart_tr = port->uart_rr | OCCUPIED;   /*     then place the echo character there */

            else if (not (port->uart_thr & OCCUPIED))       /*   otherwise if the holding register is empty */
                port->uart_thr = port->uart_rr | OCCUPIED;  /*     then queue the echo character there */

            else                                        /*   otherwise overwrite the transmitter */
                port->uart_tr = DEL | OCCUPIED;         /*     with an error indicator */

        if (port->special_ram [port->uart_rr])              /* if the character is special */
            port->uart_rhr = port->uart_rr | S0_SPECIAL;    /*   then set the special bits */
        else                                                /* otherwise */
            port->uart_rhr = port->uart_rr | S0_REGULAR;    /*   mark it as a regular character */
        }

    port->uart_rr = 0;                                  /* clear the receiver register */
    uptr->wait = 0;                                     /*   and any pending ACK wait */
    }

else if (uptr->wait > 0) {                              /* otherwise if an ACK is expected but has not arrived */
    uptr->wait = uptr-> wait * 2;                       /*   then double the wait time for the next check */

    poll_wait = sim_is_active (&dptr->units [POLL]);    /* get the current remaining poll delay */

    if (uptr->wait < poll_wait) {                       /* if the new wait is shorter than the remaining wait */
        sim_activate (uptr, uptr->wait);                /*   then reschedule the port service */

        trace_2W (dptr, TRACE_SERV, Trace_Port_Rescheduled, WINDOW_ON (address), address, uptr->wait);
        }

    else                                                /* otherwise wait for the remaining poll delay */
        trace_2W (dptr, TRACE_SERV, Trace_Port_Wait, WINDOW_ON (address), address, poll_wait);
    }


/* Transmission service */

if (port->uart_tr & OCCUPIED) {                                 /* if the transmitter register contains a character */
    data_byte = (uint8) (port->uart_tr & port->uart_data_mask); /*   then mask it to the data length */

    if (dptr->flags & DEV_DIAG) {                       /* if the loopback connector is in place */
        port->uart_rr = (uint32) data_byte | OCCUPIED;  /*   then loop the character back to the receiver */
        port->uart_tr = 0;                              /*     and clear the transmitter register */

//      trace_output (dptr, TRACE_XFER, Trace_Loopback, address, data_byte);
        trace_2T (dptr, TRACE_XFER, Trace_Loopback,
                  TRIGGER_ON_OUTPUT (address, data_byte),
                  WINDOW_ON (address),
                  address, data_byte);

        line_service (uptr);                            /* do the loopback reception immediately */
        }

    else if (data_byte == ENQ                           /* if an ENQ is being sent */
      && uptr->flags & UNIT_LOCALACK) {                 /*   and local reply is enabled */
        port->uart_rr = ACK | OCCUPIED;                 /*     then "receive" an ACK on the port */

        trace_1W (dptr, TRACE_XFER, Trace_ENQ_Absorbed, WINDOW_ON (address), address);

        activate_unit (uptr, Receive);                  /* schedule the reception */
        }

    else {                                              /* otherwise */
        chr = sim_tt_outcvt ((int32) data_byte,         /*   convert the output character */
                             TT_GET_MODE (uptr->flags));

        if (chr < 0)                                    /* if the character is to be discarded */
//          trace_output (dptr, TRACE_XFER, Trace_Discard_Output, address, data_byte);
            trace_2T (dptr, TRACE_XFER, Trace_Discard_Output,
                      TRIGGER_ON_OUTPUT (address, data_byte),
                      WINDOW_ON (address),
                      address, data_byte);

        else {                                                  /* otherwise it is printable */
            if (address == 0 && dibptr->channel_number == 1)    /*   so if we are writing to port 0 of channel 1 */
                status = sim_putchar_s (chr);                   /*     then output it to the simulation console */
            else                                                /*   otherwise */
                status = tmxr_putc_ln (lptr, chr);              /*     output it to the multiplexer line */

            if (chr == ENQ || lptr->xmte == 0)          /* if sending an ENQ or the output buffer is full */
                tmxr_poll_tx (mptr);                    /*   then flush the line buffer */

            if (status == SCPE_OK) {                    /* if the character was successfully sent */
//              trace_output (dptr, TRACE_XFER, Trace_Port_Send, address, chr);
                trace_2T (dptr, TRACE_XFER, Trace_Port_Send,
                          TRIGGER_ON_OUTPUT (address, chr),
                          WINDOW_ON (address),
                          address, chr);

                if (chr == ENQ)                                 /* if the character was an ENQ  */
                    uptr->wait = activate_unit (uptr, Receive); /*   then schedule the ACK reception */
                }

            else if (status == SCPE_LOST) {             /* otherwise if the connection was lost */
                status = SCPE_OK;                       /*   then allow the transmission to complete */

                if (lptr->txlog != NULL)                /* report the loss only if logging is disabled */
//                  trace_output (dptr, TRACE_XFER, Trace_Connection_Loss, address, chr);
                    trace_2T (dptr, TRACE_XFER, Trace_Connection_Loss,
                              TRIGGER_ON_OUTPUT (address, chr),
                              WINDOW_ON (address),
                              address, chr);
                }
            }
        }

    if (status == SCPE_OK)                              /* if transmission succeeded */
        if (port->uart_thr & OCCUPIED) {                /*   then if the holding register is occupied */
            port->uart_tr = port->uart_thr;             /*     then move it to the transmitter register */
            port->uart_thr = 0;                         /*       and clear the holding register */

            activate_unit (uptr, Send);                 /* schedule the next transmission */
            }

        else                                            /* otherwise transmission is complete */
            port->uart_tr = 0;                          /*   so clear the transmitter register */

    else if (status == SCPE_STALL) {                    /* otherwise if a stall occurred */
        trace_2W (dptr, TRACE_XFER, Trace_Stall, WINDOW_ON (address), address, chr);

        activate_unit (uptr, Stall);                    /* retry the output again */
        status = SCPE_OK;                               /*   and ignore the error */
        }
    }


if (check_service (adcc_id, address)                    /* check for service requests; if requested */
  && adcc->csrq_enable)                                 /*   and enabled */
    imb_assert_CSRQ (dibptr->channel_number);           /*     then request channel service */

return status;                                          /* return the result of the event service */
}


/* Multiplexer poll service.

   The poll service routine is used to poll for Telnet and serial port
   connections and for incoming characters.  It also polls the simulation
   console for IMB channel 1 port 0 and flushes the output buffers for all
   channels.  Polling starts at simulator startup for channel 1.  For other
   channel assignments, polling starts when the ADCC is attached to a listening
   port or a serial connection is made and stops when it is detached and all
   serial connections are closed.

   On entry, "uptr" points at the poll unit.  If diagnostic mode is not enabled,
   then a check is made for incoming connections.  If a new connection was made,
   the connected device and line number of that device are obtained from the
   returned value.  If more than one ADCC instance exists, and this instance is
   part of a connection group, then the connection might be made on a different
   device than that hosting the poll unit.  In this secondary connection case,
   the connector value will be non-zero, indicating the number of group links
   that must be followed from the current device to find the connecting device.
   Otherwise, the connector value will be zero, indicating the connection was to
   the current device instance.  After determining the connected instance,
   reception is enabled for the connection.

    If diagnostic mode is not enabled, then the connected lines are flushed of
    transmitted characters and polled for received characters.  Then each line's
    modem state is updated.  If the ADCC is hosting the system console, then the
    simulation console is polled for input.  If a character is present, then it
    is stored in the port 0 receiver register, and the reception is scheduled.

    Then each port is polled for input.  If a character is available, it is
    retrieved, stored in the port's receiver register, and port service is
    scheduled.  If the port is waiting for an ACK, the service is scheduled
    immediately; otherwise, the normal receiver rate is used.  If a character is
    not available, a check is made for a service request based on a change in
    modem status.

    When all ports are processed, the poll service is rescheduled.  If channel
    service is requested and enabled, then CSRQ is asserted before returning.


   Implementation notes:

    1. If a character is received on the simulation console, we must call the
       port 0 line service directly.  This is necessary because the poll time
       may be shorter than the port service time, and as the console provides no
       buffering, a second character received before the port service had been
       entered would be lost.

    2. Calling "tmxr_getc_ln" for port 0 when it is the simulation console is
       OK, as reception in the multiplexer line descriptor is disabled by
       default and therefore will return 0.

    3. A port that is waiting for an ACK to complete an ENQ/ACK handshake has
       its unit "wait" field set non-zero.  If the field value is greater than
       the realistic reception time, then the line service is scheduled
       immediately, as the port has already waited the minimum time necessary
       (the "wait" field value is doubled each time the line service is entered
       before the ACK has been received).  Otherwise, the service is scheduled
       using the normal reception time (realistic or optimized).
*/

static t_stat poll_service (UNIT *uptr)
{
const uint32      adcc_id = uptr->Instance;             /* the instance ID of the associated adcc */
DEVICE     *const dptr    = adccs [adcc_id].dptr;       /* the per-instance device pointer */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;       /* the per-instance adcc state pointer */
TMXR       *const mptr    = adccs [adcc_id].mptr;       /* the per-instance multiplexer pointer */
DIB        *const dibptr  = (DIB *) dptr->ctxt;         /* the pointer to the device context */
PORT_STATE        *port;
int32             connector, connection, character;
uint32            line;
UNIT              *pptr, *cuptr;
TMXR              *cmptr;
t_stat            status = SCPE_OK;
t_bool            service = FALSE;

trace_1 (dptr, TRACE_PSERV, Trace_Poll_Service, uptr->wait);

if (not (dptr->flags & DEV_DIAG)) {                     /* if we're not in diagnostic mode */
    connection = tmxr_poll_conn (mptr);                 /* poll for a serial connection only */

    if (connection != -1) {                             /* if a new connection was established */
        connector  = UPPER_WORD (connection);           /*   then get a possible secondary connection */
        connection = LOWER_WORD (connection);           /*     and mask to just the port number */

        cmptr = mptr;                                   /* initialize the connection mux pointer */
        cuptr = uptr;                                   /*   and the connection unit pointer */

        while (connector-- > 0) {                       /* if an secondary connection was made */
            cmptr = (TMXR *) cmptr->exptr;              /*     then point at the next mux descriptor */
            cuptr = cmptr->dptr->units;                 /*       and at the first unit */
            }

        adcc_connect (cuptr, connection, NULL, NULL);   /* enable the connected port to receive */
        }

    tmxr_poll_tx (mptr);                                /* flush the multiplexer output buffers */
    tmxr_poll_rx (mptr);                                /*   and poll the multiplexer connections for input */

    for (line = 0; line < TERM_COUNT; line++)           /* check each port for line state changes */
        update_status (adcc_id, line);                  /*   by updating the current line state */

    if (dibptr->channel_number == 1) {                  /* if the ADCC is on channel 1 */
        status = sim_poll_kbd ();                       /*   then poll the simulation console for input */

        if (status >= SCPE_KFLAG) {                         /* if a character was present */
            adcc->ports [0].uart_rr = status | OCCUPIED;    /*   then mark the register as occupied */
            status = SCPE_OK;                               /*     and then clear the status */

            line_service (&dptr->units [0]);            /* run the system console's I/O service */
            }
        }
    }

for (line = 0; line < TERM_COUNT; line++)               /* check each of the ports for available input */
    if (tmxr_rqln (&mptr->ldsc [line])) {               /* if a character is available on this port */
        character = tmxr_getc_ln (&mptr->ldsc [line]);  /*   then get it */

        if (character & TMXR_VALID) {                   /* if a character was returned */
            pptr = &dptr->units [line];                 /*   then get the pointer to the port unit */
            port = &adcc->ports [line];                 /*     and a pointer to the port state */

            port->uart_rr = character | OCCUPIED;       /* store the character in the receiver register */

            if (dptr->flags & DEV_REALTIME              /* if real-time mode is set */
              && pptr->wait > port->receive_delay) {    /*   and the port has waited long enough for an ACK */
                sim_cancel (pptr);                      /*     then cancel any current wait */
                activate_unit (pptr, Handshake);        /*       and activate the line service */
                }

            else                                        /* otherwise this is a normal input */
                activate_unit (pptr, Receive);          /*   so schedule with the normal receive timing */
            }
        }

    else                                                /* otherwise */
        service |= check_service (adcc_id, line);       /*   check for service requests */

uptr->wait = hp_sync_poll (Service);                    /* maintain poll unit synchronization */
sim_activate (uptr, uptr->wait);                        /*   and continue polling */

if (service && adcc->csrq_enable)                       /* if service is requested and enabled */
    imb_assert_CSRQ (dibptr->channel_number);           /*     then request channel service */

return status;                                          /* return the service status */
}


/* Activate a port unit.

   The specified unit is activated to receive or send a character.  The reason
   for the activation is specified by the "reason" parameter.  If the ADCC is in
   realtime mode, the service time is obtained from the baud rate setting.
   Otherwise, the current value of the optimized timing delay is used.


   Implementation notes:

    1. The loopback time is the difference between the reception and
       transmission times, as the latter event has already occurred when we are
       called.

    2. A stall is rescheduled to synchronize with the poll timer.  This ensures
       that idling is permitted even when a port is stalled for an extended
       length of time (such as when the simulation console is waiting for input
       in concurrent mode).
*/

static int32 activate_unit (UNIT *uptr, ACTIVATOR reason)
{
const uint32      adcc_id = uptr->Instance;                 /* the instance ID of the associated adcc */
DEVICE     *const dptr    = adccs [adcc_id].dptr;           /* the per-instance device pointer */
ADCC_STATE *const adcc    = adccs [adcc_id].sptr;           /* the per-instance adcc state pointer */
UNIT       *const pptr    = &dptr->units [POLL];            /* the per-instance poll unit pointer */
const HP_WORD     address = (HP_WORD) (uptr - dptr->units); /* the port address */
PORT_STATE *const port    = &adcc->ports [address];         /* the per-instance port state pointer */
int32             delay   = 0;

if (dptr->flags & (DEV_DIAG | DEV_REALTIME))            /* if either diagnostic or realtime mode is set */
    switch (reason) {                                   /*   then dispatch the REALTIME activation */

        case Receive:                                   /* reception event */
            delay = port->receive_delay;                /* schedule for the realistic reception time */
            break;

        case Send:                                      /* transmission event */
            delay = port->transmit_delay;               /* schedule for the realistic transmission time */
            break;

        case Handshake:                                         /* handshake reception event */
            delay = port->receive_delay - port->transmit_delay; /* schedule the additional reception overhead */

            if (delay < 0)                              /* if the receive time is less than the send time */
                delay = 1;                              /*   then use a nominal delay */
            break;

        case Stall:                                     /* transmission stall event */
            delay = sim_activate_time (pptr);           /* use the remaining poll time as the wait */
            break;
        }

else                                                    /* otherwise, we are in optimized timing mode */
    switch (reason) {                                   /*   so dispatch the FASTTIME activation */

        case Receive:                                   /* reception event */
        case Send:                                      /* transmission event */
            delay = adcc->fast_time;                    /* use the optimized timing value */
            break;

        case Handshake:                                 /* handshake reception event */
            delay = 1;                                  /* use a nominal delay */
            break;

        case Stall:                                     /* transmission stall event */
            delay = sim_activate_time (pptr);           /* use the remaining poll time as the wait */
            break;
        }

trace_2W (dptr, TRACE_SERV, Trace_Port_Scheduled, WINDOW_ON (address), address, delay);

sim_activate (uptr, delay);                             /* activate the unit */

return delay;                                           /* return the activation delay */
}



/**********************************
 *                                *
 * Interface local trace routines *
 *                                *
 **********************************/


/* Format a register trace.

   This routine is called to trace the value of a register during a read or
   write operation.  On entry, "adcc_id" is the instance ID of the ADCC, "mode"
   is either Read or Write, depending on the mode of access, "Register" is the
   enumerator designating the register to trace, and "data" is the register data
   value.  Entry assumes that the TRACE_CSRW condition has already been
   qualified.
*/

// p0 = mode, p1 = register ID, p2 = register value
// FORMAT_MODE mode, REGISTER Register, IMB_DATA data

static t_stat format_register (TRACE_RECORD *eptr)
{
const FORMAT_MODE mode     = (FORMAT_MODE) eptr->p [0];
const REGISTER    Register = (REGISTER)    eptr->p [1];
const IMB_DATA    data     = (IMB_DATA)    eptr->p [2];
uint32 address;
t_stat status = SCPE_OK;

if (mode == Read)                                       /* if formatting a register read */
    switch (Register) {                                 /*   then dispatch on the register selector */

        case Reg_0:
            trace_format (eptr->dptr, eptr->flags, "Register 0 (FIFO) %sdata in is %03o\n",
                          (R0_MODE (data) == R0_MODE_TAGGED ? "special " : ""),
                          R0_DATA (data));
            break;

        case Reg_2:
            trace_format (eptr->dptr, eptr->flags, "Register 2 (IRQ conditions) status is %s\n",
                          fmt_bitset (data, R2_format));
            break;

        case Reg_E:
            trace_format (eptr->dptr, eptr->flags, "Register E (Channel configuration) status is %schannel ID %u\n",
                          fmt_bitset (data, RE_format), RE_CHANNEL_ID (data));
            break;

        default:                                        /* unimplemented registers */
            trace_format (eptr->dptr, eptr->flags, "Register %X (unimplemented) read as %06o\n",
                          (uint32) Register, data);
            break;
        }

else                                                    /* otherwise a register write is being traced */
    switch (Register) {                                 /*   so dispatch on the register selector */

        case Reg_0:
            if ((data & R0_MODE_MASK) == R0_MODE_COMMAND) { /* if the byte is a command */
                trace_format (eptr->dptr, eptr->flags, "Register 0 (FIFO) command is ");

                address = BUS_ADDRESS (data);           /* get the bus address */

                switch BUS_GROUP (data) {               /* dispatch on the command group */

                    case BUS_PCG:                       /* Primary Command Group */
                        break;

                    case BUS_LAG:                       /* Listen Address Group */
                        if (address == BUS_UNADDRESS)
                            fputs ("Unlisten\n", trace.list);
                        else
                            fprintf (trace.list, "Listen %u\n", address);
                        break;

                    case BUS_TAG:                       /* Talk Address Group */
                        if (address == BUS_UNADDRESS)
                            fputs ("Untalk\n", trace.list);
                        else
                            fprintf (trace.list, "Talk %u\n", address);
                        break;

                    case BUS_SCG:                       /* Secondary Command Group */
                        fprintf (trace.list, "Secondary %02XH\n", address);
                        break;
                    }
                }

            else                                        /* otherwise the byte is a data value */
                trace_format (eptr->dptr, eptr->flags, "Register 0 (FIFO) data out is %02XH\n",
                              R0_DATA (data));
            break;

        case Reg_C:
                trace_format (eptr->dptr, eptr->flags, "Register C (Interrupt) control is %sdevice %u\n",
                              fmt_bitset (data, RC_format), RC_DEVICE (data));
            break;

        default:                                        /* unimplemented registers */
            trace_format (eptr->dptr, eptr->flags, "Register %X (unimplemented) write %06o ignored\n",
                          (uint32) Register, data);
            break;
        }

return status;                                          /* return with the register trace printed */
}


// trigger parse handler
// TRACE TRIGGER ADCC INPUT="<character>[<character>...]"[;PORT=<range>][ OR ...]
// TRACE TRIGGER ADCC OUTPUT="<character>[<character>...]"[;PORT=<range>][ OR ...]

static t_stat parse_trigger (char *cptr, uint32 id, void *vptr)
{
char        *tptr, gbuf [CBUFSIZE];
SET_OPTION  option;
uint32      bitmap;
TRIGGER_SET *sptr = (TRIGGER_SET *) vptr;
t_stat      status = SCPE_OK;

option = (SET_OPTION) id;

switch (option) {
    case Nothing:
        sptr->modes = Default_Mode;
        sptr->last_option = option;
        break;


    case Input:
        cptr = ex_parse_quoted_string (cptr, gbuf, FALSE, TRUE);

        if (cptr == NULL)                               /* if the string is not terminated */
            status = SCPE_ARG;                          /*   then report a bad argument */

        else if (*cptr != '\0')
            status = SCPE_2MARG;

        else {                                          /* otherwise the string is valid */
            memset (sptr->input, FALSE, CHAR_COUNT);    /*   so clear the input test vector */

            for (tptr = gbuf; *tptr != '\0'; tptr++)
                sptr->input [(uint8) *tptr] = TRUE;
            }

        sptr->modes |= Trigger_On_Input;
        sptr->input_ports = BUS_MASK;

        sptr->last_option = option;
        break;


    case Output:
        cptr = ex_parse_quoted_string (cptr, gbuf, FALSE, TRUE);

        if (cptr == NULL)                               /* if the string is not terminated */
            status = SCPE_ARG;                          /*   then report a bad argument */

        else if (*cptr != '\0')
            status = SCPE_2MARG;

        else {                                          /* otherwise the string is valid */
            memset (sptr->output, FALSE, CHAR_COUNT);   /*   so clear the output test vector */

            for (tptr = gbuf; *tptr != '\0'; tptr++)
                sptr->output [(uint8) *tptr] = TRUE;
            }

        sptr->modes |= Trigger_On_Output;
        sptr->output_ports = BUS_MASK;

        sptr->last_option = option;
        break;


    case Port:
        if (sptr->last_option == Port)
            status = SCPE_ARG;

        else if (*cptr == '\0')
            status = SCPE_2FARG;

        else {
            status = hp_parse_range (cptr, BUS_MAX, &bitmap);

            if (status == SCPE_OK)
                if (sptr->last_option == Input)
                    sptr->input_ports = bitmap;
                else
                    sptr->output_ports = bitmap;
            else
                status = SCPE_ARG;
            }
        break;
    }   // end switch

return status;
}


// command              stream  xptr    meaning
//
// Trace_Show_Trigger   file    string  print trigger on file using string prefix
// Trace_Set_Trigger    NULL    string  parse trigger from string
// Trace_Set_Trigger    NULL    NULL    clear trigger
//
// Trace_Show_Window    file    string  print window on file using string prefix
// Trace_Set_Window     NULL    string  parse window from string
// Trace_Set_Window     NULL    NULL    clear window
//
// Trace_Show_List      file    entry   list trace entry on file

static t_stat trace_adcc (TRACE_COMMAND command, uint32 instance_id, FILE *stream, void *xptr)
{
static const char *message [] = {
    "Port %u disconnected by DTR drop\n",
    "Port %d connection established\n",
    "Port %u connection raised DSR\n",
    "Port %u disconnection dropped DSR\n",
    "Port %u service entered\n",
    "Port %u receiver data overrun\n",
    "Port %u break detected\n",
    "Port %u character ENQ absorbed internally\n",
    "Poll delay %d service scheduled\n",
    "Poll delay %d service entered\n"
    };
DEVICE     *const dptr = adccs [instance_id].dptr;      /* the per-instance device pointer */
ADCC_STATE *const adcc = adccs [instance_id].sptr;      /* the per-instance adcc state pointer */
TMXR       *const mptr = adccs [instance_id].mptr;      /* the per-instance multiplexer pointer */
char              *cptr, gbuf [CBUFSIZE];
TRACE_RECORD      *eptr;
TRIGGER_SET       new_trigger;
t_bool            first;
uint32            bitmap, max_port, index;
t_stat            status = SCPE_OK;

switch (command) {

    case Trace_Set_Trigger:
        new_trigger.modes       = adcc->trigger_modes;
        new_trigger.last_option = Nothing;

        memcpy (new_trigger.input,  adcc->trigger_input,  CHAR_COUNT);
        memcpy (new_trigger.output, adcc->trigger_output, CHAR_COUNT);

        new_trigger.input_ports = adcc->trigger_input_ports;
        new_trigger.output_ports = adcc->trigger_output_ports;

        if (xptr == NULL)                           // clear trigger?
            new_trigger.modes = Default_Mode;

        else
            status = trace_parse (xptr, trigger_table, &new_trigger);

        if (status == SCPE_OK) {
            adcc->trigger_modes = new_trigger.modes;

            memcpy (adcc->trigger_input,  new_trigger.input,  CHAR_COUNT);
            memcpy (adcc->trigger_output, new_trigger.output, CHAR_COUNT);

            adcc->trigger_input_ports  = new_trigger.input_ports;
            adcc->trigger_output_ports = new_trigger.output_ports;
            }
        break;


    case Trace_Show_Trigger:
        cptr = (char *) xptr;
        status = SCPE_INCOMP;

        if (adcc->trigger_modes & Trigger_On_Input) {
            fprintf (stream, "%s %s input is ", cptr, sim_dname (dptr));

            first = TRUE;

            for (index = 0; index < CHAR_COUNT; index++) {
                if (adcc->trigger_input [index]) {
                    if (first)
                        first = FALSE;
                    else
                        fputs (" or ", stream);

                    fputs (fmt_char (index), stream);
                    }
                }

            if (adcc->trigger_input_ports == BUS_MAX)
                fputc ('\n', stream);

            else {
                fputs (" and port is ", stream);

                status = hp_print_range (stream, BUS_MAX,
                                         &adcc->trigger_input_ports);
                }

            cptr = "        or";
            status = SCPE_OK;
            }

        if (adcc->trigger_modes & Trigger_On_Output) {
            fprintf (stream, "%s %s output is ", cptr, sim_dname (dptr));

            first = TRUE;

            for (index = 0; index < CHAR_COUNT; index++) {
                if (adcc->trigger_output [index]) {
                    if (first)
                        first = FALSE;
                    else
                        fputs (" or ", stream);

                    fputs (fmt_char (index), stream);
                    }
                }

            if (adcc->trigger_output_ports == BUS_MAX)
                fputc ('\n', stream);

            else {
                fputs (" and port is ", stream);

                status = hp_print_range (stream, BUS_MAX,
                                         &adcc->trigger_output_ports);
                }

            status = SCPE_OK;
            }

        break;


    case Trace_Set_Window:
        cptr = (char *) xptr;

        if (cptr == NULL)                           // clear window?
            adcc->window_ports = BUS_MASK;

        else if (*cptr == '\0')
            status = SCPE_2FARG;

        else {
            cptr = get_glyph (cptr, gbuf, '=');

            if (strncmp (gbuf, "PORT", strlen (gbuf)) == 0) {
                max_port = mptr->lines - 1;
                status = hp_parse_range (cptr, max_port, &bitmap);

                if (status == SCPE_OK)
                    adcc->window_ports = bitmap;
                }

            else
                status = SCPE_ARG;
            }
        break;


    case Trace_Show_Window:
        max_port = mptr->lines - 1;
        cptr = (char *) xptr;

        if (adcc->window_ports == BUS_MASK)     // show
            status = SCPE_NOFNC;        // no window set

        else {     // print window
            fprintf (stream, "%s %s port is ", cptr, sim_dname (dptr));

            status = hp_print_range (stream, max_port,
                                     &adcc->window_ports);
            }
        break;


    case Trace_Set_Store:
        status = trace_store (dptr, (char *) xptr, TRACE_COMMON, TRACE_ALL);
        break;


    case Trace_Set_List:
        break;                                          /* list initialization is not required */


    case Trace_Show_List:
        eptr = (TRACE_RECORD *) xptr;

        switch (eptr->record_id) {
// these must start at 0
            case Trace_DTR_Drop:                        /* p0 = port number */
            case Trace_Connected:                       /* p0 = port number */
            case Trace_DSR_Raise:                       /* p0 = port number */
            case Trace_DSR_Drop:                        /* p0 = port number */
            case Trace_Port_Service:                    /* p0 = port number */
            case Trace_Overrun:                         /* p0 = port number */
            case Trace_Break:                           /* p0 = port number */
            case Trace_ENQ_Absorbed:                    /* p0 = port number */
            case Trace_Poll_Schedule:                   /* p0 = poll delay */
            case Trace_Poll_Service:                    /* p0 = port number */
                trace_format (eptr->dptr, eptr->flags,
                              message [eptr->record_id],
                              eptr->p [0]);
                break;


// 0
            case Trace_Poll_Stopped:
                trace_format (eptr->dptr, eptr->flags, "Poll service stopped\n");
                break;


// 1
            /* p0 = status */

            case Trace_OBSI_Status:
                trace_format (eptr->dptr, eptr->flags,
                              "OBSI status is %schannel %u | device %u\n",
                              (eptr->p [0] & RF_DEVRQ ? "device request | " : ""),
                              RF_CHANNEL (eptr->p [0]), RD_DEVICE (eptr->p [0]));
                break;


            /* p0 = status */

            case Trace_OBII_Status:
                trace_format (eptr->dptr, eptr->flags,
                              "OBII status is channel %u | device %u\n",
                              RD_CHANNEL (eptr->p [0]), RD_DEVICE (eptr->p [0]));
                break;


// 2
            /* p0 = port number, p1 = character */

            case Trace_Modifier_0_Read:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 0 input character is %s\n",
                              eptr->p [0], fmt_char ((uint8) S0_DATA (eptr->p [1])));
                break;


            /* p0 = port number, p1 = status */

            case Trace_Modifier_1_Read:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 1 UART status is %s\n",
                              eptr->p [0], fmt_bitset (eptr->p [1], S1_status_format));
                break;


            /* p0 = port number, p1 = status */

            case Trace_Modifier_2_Read:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 2 interface status is %s\n",
                              eptr->p [0], fmt_bitset (eptr->p [1], S2_status_format));
                break;


            /* p0 = port number, p1 = DSJ value */

            case Trace_DSJ:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u DSJ %u condition is %s\n",
                              eptr->p [0], eptr->p [1], dsj_names [eptr->p [1]]);
                break;


            /* p0 = port number, p1 = character */

            case Trace_Modifier_0_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 0 output character is %s\n",
                              eptr->p [0], fmt_char ((uint8) S0_DATA (eptr->p [1])));
                break;


            /* p0 = port number, p1 = control */

            case Trace_Modifier_1_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 1 UART control is length %u | %s\n",
                              eptr->p [0], S1_LENGTH (eptr->p [1]) + 5,
                              fmt_bitset (eptr->p [1], S1_control_format));
                break;


            /* p0 = port number, p1 = control */

            case Trace_Modifier_2_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 2 interface control is %s\n",
                              eptr->p [0], fmt_bitset (eptr->p [1], S2_control_format));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Modifier_4_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 4 character %s is not special\n",
                              eptr->p [0], fmt_char ((uint8) S4_DATA (eptr->p [1])));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Modifier_5_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 5 character %s is special\n",
                              eptr->p [0], fmt_char ((uint8) S4_DATA (eptr->p [1])));
                break;


            /* p0 = port number, p1 = baud rate modifier */

            case Trace_Modifier_6_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 6 %s is %s\n",
                              eptr->p [0], mode_names [S6_MODE (eptr->p [1])],
                              baud_names [S6_RATE (eptr->p [1])]);
                break;


            /* p0 = port number, p1 = status modifier */

            case Trace_Modifier_7_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 7 is set status %s %s\n",
                              eptr->p [0],
                              (eptr->p [1] & S7_SET_MASK ? "mask" : "reference"),
                              fmt_bitset (eptr->p [1], S7_format));
                break;


            /* p0 = port number, p1 = DSJ value */

            case Trace_CSRQ_Assert:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u asserts CSRQ for DSJ %u condition %s\n",
                              eptr->p [0], eptr->p [1], dsj_names [eptr->p [1]]);
                break;


            /* p0 = port number, p1 = character */

            case Trace_Reception:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u UART received character %s\n",
                              eptr->p [0], fmt_char (LOWER_BYTE (eptr->p [1])));
                break;


            /* p0 = port number, p1 = delay */

            case Trace_Port_Rescheduled:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u delay %d service rescheduled for ACK\n",
                              eptr->p [0], eptr->p [1]);
                break;


            /* p0 = port number, p1 = delay */

            case Trace_Port_Wait:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u delay %d waiting for poll for ACK\n",
                              eptr->p [0], eptr->p [1]);
                break;


            /* p0 = port number, p1 = character */

            case Trace_Loopback:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u UART sent character %s looped back to receiver\n",
                              eptr->p [0], fmt_char (eptr->p [1]));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Discard_Output:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u character %s discarded by output filter\n",
                              eptr->p [0], fmt_char (eptr->p [1]));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Port_Send:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u UART sent character %s\n",
                              eptr->p [0], fmt_char (LOWER_BYTE (eptr->p [1])));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Connection_Loss:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u character %s discarded by connection loss\n",
                              eptr->p [0], fmt_char (LOWER_BYTE (eptr->p [1])));
                break;


            /* p0 = port number, p1 = character */

            case Trace_Stall:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u character %s transmission stalled for full buffer\n",
                              eptr->p [0], fmt_char (LOWER_BYTE (eptr->p [1])));
                break;


            /* p0 = port number, p1 = delay */

            case Trace_Port_Scheduled:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u delay %d service scheduled\n",
                              eptr->p [0], eptr->p [1]);
                break;


// 3
            /* p0 = port number, p1 = condition, p2 = service masks */

            case Trace_Modifier_3_Write:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u modifier 3 %s condition %u %s mask is %s\n",
                              eptr->p [0], (eptr->p [1] & S3_ENABLE ? "enable" : "disable"),
                              S3_CONDITION (eptr->p [1]),
                              fmt_bitset (BIT (S3_CONDITION (eptr->p [1])), S3_format),
                              fmt_bitset (eptr->p [2], S3_format));
                break;


            /* p0 = mode, p1 = register ID, p2 = register value */

            case Trace_Register:
                status = format_register (eptr);
                break;

// 5
            /* p0 = port number, p1 = service masks, p2 = modem status, p3 = status reference, p4 = status mask */

            case Trace_Service_Condition:
                trace_format (eptr->dptr, eptr->flags,
                              "Port %u service condition mask %s modem %s reference %s mask %s\n",
                              eptr->p [0], fmt_bitset (eptr->p [1], S3_format),
                              fmt_bitset (eptr->p [2], S2_status_format),
                              fmt_bitset (eptr->p [3], S7_format),
                              fmt_bitset (eptr->p [4], S7_format));
                break;


            default:
                status = SCPE_IERR;
                break;
            }
        break;
    }

return status;
}
