/* hp3000_cpu.h: HP 3000 CPU declarations

   Copyright (c) 2016-2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   09-Dec-24    JDB     Added UNPAIRED_STACKOP constant
   12-Sep-24    JDB     Added MVBW_FLAGS, memory-reference mode flags
                        Added BR opcode
   03-Jul-24    JDB     Added LSTT accessors
   01-Jul-24    JDB     Added CSMAP to the set of CPU registers
                        Renamed SGT_POINTER to SYSGLOB to match documentation
   23-Mar-24    JDB     Added "trap_SysHalt_CST_Length" macro
   22-Mar-24    JDB     Moved PFARS from unit flags to system configuration
                        Removed MEMSIZE macro and "cpu_unit" global
   21-Mar-24    JDB     Removed "cpu_run_mode_interrupt" global; now local to main
   14-Mar-24    JDB     Removed "cpu_exit_procedure" global; now local to base
   09-Mar-24    JDB     Removed CST_BANK_MASK and DST_BANK_MASK macros
                        Added "cpu_bank_mask" and "cpu_sir_enables" externals
   28-Feb-24    JDB     Added "else" clause to guard PREADJUST_SR use
   04-Jan-24    JDB     Added SS_COLDLOAD simulation stop condition
   28-Dec-23    JDB     Removed "cpu_start_dispatcher" global; now local to base
   02-Nov-23    JDB     Moved firmware presence options to sys_configuration
   07-Oct-23    JDB     Added Amigo clock definitions
   12-Jul-23    JDB     Added PCN_SERIES_4X_5X CPU number, HPIB_IO instruction
   27-Mar-23    JDB     Added the Series 58 CPU model for Amigo I/O instructions
   22-Nov-22    JDB     Added "not system controller" and "not CIC" syshalts
   26-Oct-21    JDB     Renamed DEB_xxx trace flags to TRACE_xxx
   08-Apr-21    JDB     Expanded bank register widths from 4 to 6 bits
   07-Apr-21    JDB     Added the Series IIIX CPU model
   11-Oct-20    JDB     Moved UNIT_OPTS here from hp3000_cpu.c for updating ease
   09-Oct-20    JDB     Renamed trap_Word_Count_Overflow to trap_Invalid_Decimal_Length
   07-Oct-20    JDB     Corrected EXEC trace to omit PSERV events
   05-Sep-20    JDB     Added the EIS dispatcher, corrected EIS_SDEC_MASK value
   18-Feb-19    JDB     Added SS_PAUSE_RESUMED simulation stop condition
   25-Jul-18    JDB     Fixed typo in "cpu_setup_code_segment" declaration
   07-Nov-16    JDB     Added SETR and SETR_X for SETR executor use
                        Renamed cpu_byte_to_word_ea to cpu_byte_ea
   03-Nov-16    JDB     Added LABEL_LOCAL for PARC/XBR/ENDP executor use
   01-Nov-16    JDB     Added debug flag for per-instruction trace capability
   22-Oct-16    JDB     Added "cpu_interrupt_pending" global from cpu_base.c
   07-Oct-16    JDB     Added "cpu_dev" external so executors can trace instructions
   28-Sep-16    JDB     Added CIS definitions
   22-Sep-16    JDB     Added cpu_byte_to_word_ea, STATUS_CS macro
   21-Sep-16    JDB     Added STATUS_CCI, UNIT_CIS, and the CIS dispatcher
   12-Sep-16    JDB     Added the PCN_SERIES_II and PCN_SERIES_III constants
   02-Sep-16    JDB     Added the POWER_STATE enumeration type, the UNIT_PFARS
                        flag, and the "cpu_power_state" external declaration
   24-Aug-16    JDB     Fixed the UNIT_CPU_MODEL test macro
   23-Aug-16    JDB     Added the MOD (module control) register
   12-Jul-16    JDB     Renamed "loading" EXEC_STATE to "waiting"
   21-Mar-16    JDB     Changed cpu_ccb_table type from uint16 to HP_WORD
   14-Feb-16    JDB     First release version
   11-Dec-12    JDB     Created


   This file provides the declarations for interoperation between the CPU and
   its supporting modules.  It provides the symbols that allow direct
   manipulation of the CPU registers and determination of currently installed
   features.
*/



#include <setjmp.h>



/* Supported breakpoint switches */

#define BP_EXEC             (SWMASK ('E'))      /* an execution breakpoint */
#define BP_SUPPORTED        (BP_EXEC)           /* the list of supported breakpoint types */


/* CPU unit flags and accessors.

   UNIT_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
              | - | - | - | - | - | - | - | - | - | - | - | - | - | - | - | T |
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     T = Process clock timing mode is calibrated
*/

#define UNIT_CALTIME        UF_BIT (0)          /* the process clock is calibrated to wall time */


/* CPU tracing formats */

#define BOV_FORMAT          "%02o.%06o  %06o  " /* bank-offset-value trace format string */


/* CPU stop flags */

#define SS_LOOP             BIT (0)             /* stop on infinite loop */
#define SS_PAUSE            BIT (1)             /* stop on PAUS instruction */
#define SS_UNDEF            BIT (2)             /* stop on undefined instruction */
#define SS_UNIMPL           BIT (3)             /* stop on unimplemented instruction */
#define SS_COLDLOAD         BIT (4)             /* stop on a cold-load trap */

#define SS_PAUSE_RESUMED    BIT (30)            /* stop resumes into a PAUS instruction */
#define SS_BYPASSED         BIT (31)            /* stops are bypassed for this instruction */


/* Memory access macros */

#define cpu_read_memory(c,o,v)      mem_read  (&cpu_dev, c, o, v)
#define cpu_write_memory(c,o,v)     mem_write (&cpu_dev, c, o, v)



/* System power state.

   The HP 3000 power supply uses two signals to indicate its state: PON (power
   on) and PFW (power-fail warning).  PON is asserted when the DC power levels
   are within their operating ranges.  PFW is asserted when AC power is lost.
   When a power failure occurs, PFW will be asserted at least three milliseconds
   before PON is denied.  When power is restored, PFW denies immediately, but
   PON does not assert until the DC output voltages have stabilized, and the
   machine is ready to resume execution.

   In simulation, the four states of these two signals are modeled with
   enumeration constants, as follows:

     PON  PFW  State            Simulator Action
     ---  ---  ---------------  ----------------------------
      1    0   power on         executing normally
      1    1   power failing    executing with cpx1_PFINTR
      0    1   power off        will not execute
      0    0   power returning  executing with trap_Power_On
*/

typedef enum {
    Power_On,
    Power_Failing,
    Power_Off,
    Power_Returning
    } POWER_STATE;


/* Micromachine execution state */

typedef enum {
    Running,                                    /* the micromachine is running */
    Paused,                                     /* a PAUS instruction has been executed */
    Waiting,                                    /* a cold load or dump is in progress */
    Halted                                      /* a programmed or front panel HALT has been executed */
    } EXEC_STATE;


/* CPX register flags.

   The CPX1 register contains flags that designate the run-time interrupts.  The
   CPX2 register contains flags for halt-time interrupts.


   Implementation notes:

    1. These are implemented as enumeration types to allow the "gdb" debugger to
       display the CPX register values as bit sets.

    2. The following CPX1 bits are not currently used in simulation: BNDVIOL,
       SYSPAR, DATAPAR, EMULATOR, and OPTION.  The ADDRPAR bit is used, but not
       to indicate bad parity on the address bus.  Instead, it indicates a
       memory write that uses a read-only classification, such as instruction
       fetch.  This should never occur and is an indication of an internal
       simulator error.

    3. The following CPX2 bits are not currently used in simulation: LOADREG,
       LOADADDR, LOADMEM, DISPMEM, SNGLINST, EXECSWCH, INCRADDR, and DECRADDR.
*/

typedef enum {
    cpx1_INTOVFL  = HP_BIT  (0),                /* integer overflow */
    cpx1_BNDVIOL  = HP_BIT  (1),                /* bounds violation */
    cpx1_ILLADDR  = HP_BIT  (2),                /* illegal address */
    cpx1_CPUTIMER = HP_BIT  (3),                /* module access timeout */
    cpx1_SYSPAR   = HP_BIT  (4),                /* system parity error */
    cpx1_ADDRPAR  = HP_BIT  (5),                /* address parity error */
    cpx1_DATAPAR  = HP_BIT  (6),                /* data parity error */
    cpx1_MODINTR  = HP_BIT  (7),                /* module interrupt */
    cpx1_EXTINTR  = HP_BIT  (8),                /* external interrupt */
    cpx1_PFINTR   = HP_BIT  (9),                /* power fail interrupt */
    cpx1_UNUSED   = HP_BIT (10),                /* unused, always 0 */
    cpx1_ICSFLAG  = HP_BIT (11),                /* ICS flag */
    cpx1_DISPFLAG = HP_BIT (12),                /* dispatcher-is-active flag */
    cpx1_EMULATOR = HP_BIT (13),                /* emulator-in-use flag */
    cpx1_IOTIMER  = HP_BIT (14),                /* I/O timeout */
    cpx1_OPTION   = HP_BIT (15)                 /* option present */
    } CPX1FLAG;

#define CPX1_IRQ_SET        HP_BITS (0, 9)      /* the set of CPX1 interrupt sources */


typedef enum {
    cpx2_RUNSWCH  = HP_BIT  (0),                /* RUN switch */
    cpx2_DUMPSWCH = HP_BIT  (1),                /* DUMP switch */
    cpx2_LOADSWCH = HP_BIT  (2),                /* LOAD switch */
    cpx2_LOADREG  = HP_BIT  (3),                /* load register */
    cpx2_LOADADDR = HP_BIT  (4),                /* load address */
    cpx2_LOADMEM  = HP_BIT  (5),                /* load memory */
    cpx2_DISPMEM  = HP_BIT  (6),                /* display memory */
    cpx2_SNGLINST = HP_BIT  (7),                /* single instruction */
    cpx2_EXECSWCH = HP_BIT  (8),                /* EXECUTE switch */
    cpx2_INCRADDR = HP_BIT  (9),                /* increment address */
    cpx2_DECRADDR = HP_BIT (10),                /* decrement address */
/*  cpx2_UNUSED   = HP_BIT (11),                   unused, always 0 */
/*  cpx2_UNUSED   = HP_BIT (12),                   unused, always 0 */
    cpx2_INHPFARS = HP_BIT (13),                /* inhibit power-fail auto-restart */
    cpx2_SYSHALT  = HP_BIT (14),                /* system halt */
    cpx2_RUN      = HP_BIT (15)                 /* run flip-flop */
    } CPX2FLAG;

#define CPX2_IRQ_SET        HP_BITS (0, 8)      /* the set of CPX2 interrupt sources */


/* Status and Interrupt Register.

   The Status and Interrupt Register (SIR) replaces the CPX1 and CPX2 registers
   on the Series 4x machines.  The format is:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | O | K | C | I | R | F | V | M | T | G | A | N | D | S | L | H |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     O = CPU is released from reset
     K = CPU timer tick
     C = IMB CSRQ
     I = IMB IRQ
     R = CPU is up and running
     F = IMB power fail warning
     V = Integer overflow trap (T * O)
     M = Memory parity error
     T = IMB timeout
     G = Toggle RUN/HALT from control panel
     A = Disable Attention FF
     N = IMB DNV
     D = Dispatcher flag (0)
     S = ICS flag (0)
     L = Split-Stack flag (0)
     H = Run/Halt indication to CMP (0/1)

   Note that bits 12-15 are negative logic, i.e., 0 = the active condition.
*/

typedef enum {
    sir_CPUOK      = HP_BIT  (0),               /* (O) CPU is released from reset */
    sir_CPUTIMER   = HP_BIT  (1),               /* (K) CPU timer tick */
    sir_IMBCSRQ    = HP_BIT  (2),               /* (C) IMB Channel Service Request */
    sir_IMBIRQ     = HP_BIT  (3),               /* (I) IMB Interrupt Request */
    sir_CPURUN     = HP_BIT  (4),               /* (R) CPU is up and running */
    sir_PFW        = HP_BIT  (5),               /* (F) IMB power fail warning */
    sir_INTOVF     = HP_BIT  (6),               /* (V) Integer overflow trap */
    sir_MEMPARERR  = HP_BIT  (7),               /* (M) Memory parity error */
    sir_IMBTIMEOUT = HP_BIT  (8),               /* (T) IMB Timeout */
    sir_IMBDNV     = HP_BIT (11),               /* (N) IMB Data Not Valid */
    sir_DISPL      = HP_BIT (12),               /* (D) Dispatcher flag (active low) */
    sir_ICSL       = HP_BIT (13),               /* (S) ICS flag (active low) */
    sir_SPLITSTKL  = HP_BIT (14),               /* (L) Split-Stack flag (active low) */
    sir_RUNL       = HP_BIT (15)                /* (H) Run/Halt indication to CMP (active low) */
    } SIRFLAG;

#define SIR_IRQ_SET         (sir_CPUTIMER | sir_IMBIRQ) /* the set of SIR interrupt sources */


/* Interrupt classifications.

   Interrupts are generated by setting CPX1 or SIR bits, by microcode aborts
   (traps), or by the DISP or IXIT instructions to run the OS dispatcher or in
   response to an external interrupt.  Each interrupt source is serviced by a
   procedure in the Segment Transfer Table of segment 1, and the classification
   values are chosen to match the STT numbers.


   Implementation notes:

    1. The STT numbers are relevant only for interrupts that come from setting
       the CPX1 bits (except for bit 8) and from the SIR bit for the system
       clock tick.  The enumeration values for external, trap, DISP, and IXIT
       interrupts are not significant.
*/

typedef enum {
/*  identifier                STT        Source  Description               */
/*  -----------------------   ---        ------  ------------------------- */
    irq_Integer_Overflow    = 000,    /* CPX1.0  Integer Overflow          */
    irq_Bounds_Violation    = 001,    /* CPX1.1  Bounds Violation          */
    irq_Illegal_Address     = 002,    /* CPX1.2  Illegal Memory Address    */
    irq_Timeout             = 003,    /* CPX1.3  Non-Responding Module     */
    irq_System_Parity       = 004,    /* CPX1.4  System Parity Error       */
    irq_Address_Parity      = 005,    /* CPX1.5  Address Parity Error      */
    irq_Data_Parity         = 006,    /* CPX1.6  Data Parity Error         */
    irq_Module              = 007,    /* CPX1.7  Module Interrupt          */
    irq_External            = 010,    /* CPX1.8  External Interrupt        */
    irq_Power_Fail          = 011,    /* CPX1.9  Power Fail Interrupt      */
    irq_System_Clock        = 014,    /* SIR.2   System Clock Interrupt    */
    irq_Trap                = 015,    /* uABORT  System or user trap       */
    irq_Dispatch            = 016,    /* DISP    Run the dispatcher        */
    irq_IXIT                = 017     /* IXIT    External interrupt        */
    } IRQ_CLASS;


/* Trap classifications.

   Except for the power-on and cold-load traps, all traps result from microcode
   aborts.  In hardware, these result in microcode jumps to the appropriate trap
   handlers.  In simulation, a MICRO_ABORT executes a "longjmp" to the trap
   handler just before the main instruction loop.  As with interrupts, each trap
   is serviced by a procedure in the Segment Transfer Table of segment 1, and
   the classification values are chosen to match the STT numbers.

   Traps are invoked by the MICRO_ABORT macro, which takes as its parameter a
   trap classification value.  Some traps require an additional parameter and
   must be invoked by the MICRO_ABORTP macro, which takes a trap classification
   value and a trap-specific value as parameters.

   Accessors are provided to separate the TRAP_CLASS and the parameter from the
   combined longjmp value.


   Implementation notes:

    1. Trap classifications must be > 0 for longjmp compatibility.

    2. A System Halt trap does not call an OS procedure but instead stops the
       simulator.  The parameter number indicates the reason for the halt.

    3. The User trap is subdivided into traps for a number of arithmetic
       conditions.  These are indicated by their corresponding parameter numbers
       in the upper words of the longjmp values.
*/

typedef enum {
/*  identifier                 STT        Source  Description               */
/*  ------------------------   ---        ------  ------------------------- */
    trap_None                = 000,    /*   --    (none)                    */
    trap_Bounds_Violation    = 001,    /* ucode   Bounds Violation          */
    trap_Unimplemented       = 020,    /* ucode   Unimplemented Instruction */
    trap_STT_Violation       = 021,    /* ucode   STT Violation             */
    trap_CST_Violation       = 022,    /* ucode   CST Violation             */
    trap_DST_Violation       = 023,    /* ucode   DST Violation             */
    trap_Stack_Underflow     = 024,    /* ucode   Stack Underflow           */
    trap_Privilege_Violation = 025,    /* ucode   Privileged Mode Violation */
    trap_Stack_Overflow      = 030,    /* ucode   Stack Overflow            */
    trap_User                = 031,    /* ucode   User Trap                 */
    trap_CS_Absent           = 037,    /* ucode   Absent Code Segment       */
    trap_Trace               = 040,    /* ucode   Trace                     */
    trap_Uncallable          = 041,    /* ucode   STT Entry Uncallable      */
    trap_DS_Absent           = 042,    /* ucode   Absent Data Segment       */
    trap_Power_On            = 043,    /*  hdwe   Power On                  */
    trap_Cold_Load           = 044,    /* ucode   Cold Load                 */
    trap_System_Halt         = 045     /* ucode   System Halt               */
    } TRAP_CLASS;

#define trap_Integer_Overflow       TO_DWORD (001, trap_User)
#define trap_Float_Overflow         TO_DWORD (002, trap_User)
#define trap_Float_Underflow        TO_DWORD (003, trap_User)
#define trap_Integer_Zero_Divide    TO_DWORD (004, trap_User)
#define trap_Float_Zero_Divide      TO_DWORD (005, trap_User)
#define trap_Ext_Float_Overflow     TO_DWORD (010, trap_User)
#define trap_Ext_Float_Underflow    TO_DWORD (011, trap_User)
#define trap_Ext_Float_Zero_Divide  TO_DWORD (012, trap_User)
#define trap_Decimal_Overflow       TO_DWORD (013, trap_User)
#define trap_Invalid_ASCII_Digit    TO_DWORD (014, trap_User)
#define trap_Invalid_Decimal_Digit  TO_DWORD (015, trap_User)
#define trap_Invalid_Word_Count     TO_DWORD (016, trap_User)
#define trap_Invalid_Decimal_Length TO_DWORD (017, trap_User)
#define trap_Decimal_Zero_Divide    TO_DWORD (020, trap_User)

#define trap_SysHalt_STTV_1         TO_DWORD ( 1, trap_System_Halt)
#define trap_SysHalt_Absent_ICS     TO_DWORD ( 2, trap_System_Halt)
#define trap_SysHalt_Absent_1       TO_DWORD ( 3, trap_System_Halt)
#define trap_SysHalt_Overflow_ICS   TO_DWORD ( 4, trap_System_Halt)
#define trap_SysHalt_CST_Length     TO_DWORD ( 5, trap_System_Halt)
#define trap_SysHalt_IO_Timeout     TO_DWORD ( 6, trap_System_Halt)
#define trap_SysHalt_CP_Checksum    TO_DWORD ( 7, trap_System_Halt)
#define trap_SysHalt_CP_Abort       TO_DWORD ( 8, trap_System_Halt)
#define trap_SysHalt_PSEB_Enabled   TO_DWORD ( 9, trap_System_Halt)
#define trap_SysHalt_Not_SysCntlr   TO_DWORD (12, trap_System_Halt)
#define trap_SysHalt_CSTV_1         TO_DWORD (13, trap_System_Halt)
#define trap_SysHalt_Not_CIC        TO_DWORD (17, trap_System_Halt)
#define trap_SysHalt_LOCK_EI        TO_DWORD (23, trap_System_Halt)
#define trap_SysHalt_Trace_1        TO_DWORD (33, trap_System_Halt)

#define PARAM(i)            (uint32)     UPPER_WORD (i)
#define TRAP(i)             (TRAP_CLASS) LOWER_WORD (i)

#define MICRO_ABORT(t)      longjmp (cpu_save_env, (int) (t))
#define MICRO_ABORTP(t,p)   longjmp (cpu_save_env, (int) TO_DWORD ((p),(t)))


/* Central Data Bus module definitions */

#define MODULE_MEMORY_LOWER     0               /* lower memory MCL address */
#define MODULE_MEMORY_UPPER     2               /* upper memory MCL address */
#define MODULE_MEMORY           3               /* upper bound of MCL addresses */
#define MODULE_PORT_CNTLR       4               /* selector channel port controller address */
#define MODULE_CPU              5               /* CPU MCU address */
#define MODULE_UNDEFINED        6               /* addresses 6-7 are unused */

#define MOP_NOP                 0               /* module operation 00 = no operation */
#define MOP_WRITE               1               /* module operation 01 = write */
#define MOP_READ                2               /* module operation 10 = read */
#define MOP_READ_WRITE_ONES     3               /* module operation 11 = read/write ones */


/* Module control register accessors.

   The module control register, MOD, has this format:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0 |  MOP  | 0 |   FROM    | 0   0   0   0 | B | A | 0   0 |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     MOP  = module operation
     FROM = source module address
     B    = this CPU is CPU #2
     A    = this CPU is CPU #1
*/

#define MOD_MOP_MASK        0030000u            /* MOD register MOP field mask */
#define MOD_FROM_MASK       0003400u            /* MOD register FROM field mask */
#define MOD_CPU_2           0000010u            /* (B) CPU number 2 MCU */
#define MOD_CPU_1           0000004u            /* (A) CPU number 1 MCU */

#define MOD_MOP_SHIFT       12                  /* MOD register MOP field alignment shift */
#define MOD_FROM_SHIFT      8                   /* MOD register FROM field alignment shift */

#define TO_MOD_MOP(v)       ((v) << MOD_MOP_SHIFT  & MOD_MOP_MASK)
#define TO_MOD_FROM(v)      ((v) << MOD_FROM_SHIFT & MOD_FROM_MASK)


/* Status register accessors.

   The CPU status register, STA, has this format:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | M | I | T | R | O | C | ccode |  current code segment number  |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     M = user mode/protected mode (0/1)
     I = external interrupts are enabled
     T = user traps are enabled
     R = right-hand stack operation executes next
     O = arithmetic overflow has occurred
     C = arithmetic carry has occurred

   Condition Code:

     00 = CCG (greater than)
     01 = CCL (less than)
     10 = CCE (equal to)
     11 = invalid
*/

#define STATUS_M            HP_BIT  (0)         /* (M) mode flag */
#define STATUS_I            HP_BIT  (1)         /* (I) interrupt flag */
#define STATUS_T            HP_BIT  (2)         /* (T) trap flag */
#define STATUS_R            HP_BIT  (3)         /* (R) right-hand stack op flag */
#define STATUS_O            HP_BIT  (4)         /* (O) overflow flag */
#define STATUS_C            HP_BIT  (5)         /* (C) carry flag */
#define STATUS_CC_MASK      HP_BITS (6,  7)     /* condition code mask */
#define STATUS_CS_MASK      HP_BITS (8, 15)     /* code segment mask */

#define STATUS_CS_WIDTH     8                   /* code segment mask width */

#define STATUS_CC(s)        HP_BITS_TO (6,  7, s)       /* condition code field */
#define STATUS_CS(s)        HP_BITS_TO (8, 15, s)       /* code segment number field */

#define TO_STATUS_CC(n)     TO_HP_BITS (6, 7, n)

#define STATUS_CCG          TO_STATUS_CC (0u)   /* condition code greater than */
#define STATUS_CCL          TO_STATUS_CC (1u)   /* condition code less than */
#define STATUS_CCE          TO_STATUS_CC (2u)   /* condition code equal to */
#define STATUS_CCI          TO_STATUS_CC (3u)   /* condition code invalid */

#define STATUS_OVTRAP       (STATUS_T | STATUS_O)
#define STATUS_NPRV         (STATUS_T | STATUS_O | STATUS_C | STATUS_CC_MASK)


/* Status register mode tests */

#define PRIV                (STA & STATUS_M)            /* current mode is privileged */
#define NPRV                (not PRIV)                  /* current mode is non-privileged */


/* Condition code flags.

   Several instructions define "condition code flags" that specify condition
   code tests to be performed.  These flags are ANDed with the condition code
   from the status register to establish whether the test passes.  CCE and CCL
   are encoded in the status register by bits 6 and 7, respectively, but CCG is
   encoded as 00, causing a direct AND to fail.  This is resolved in microcode
   by the "CC" S-bus micro-order, which copies bits 6 and 7 to bits 8 and 9 of
   the S-bus register and also sets bit 7 if bits 8 and 9 are zero (CCG).  The
   result is a single bit set for each condition code that may be ANDed with the
   CCF field of the instruction.

   The MVBW (Move Bytes While) instruction moves bytes from the source to the
   destination while the bytes are alphabetic, numeric, or either, depending on
   the A and N bits in the instruction.  The CCB classification table is used to
   determine the type of each byte moved.  If both A and N bits are set, then a
   table entry of either CCG or CCE will succeed.  Matching via a single AND
   operation is possible only if the encoding of all three conditions is
   disjoint.

   We implement this scheme without the two-bit right-shift, so that ANDing a
   CCF with STATUS_CC_MASK will produce the correctly aligned status register
   value.  The TO_CCF macro converts the current condition code in the status
   register to a condition code flag.
*/

#define CFL                 0000400u            /* condition code flag less than */
#define CFE                 0001000u            /* condition code flag equal to */
#define CFG                 0002000u            /* condition code flag greater than */

#define TO_CCF(s)           ((s) & STATUS_CC_MASK ? (s) & STATUS_CC_MASK : CFG)


/* Condition code patterns.

   Machine instructions typically set the condition code field of the status
   register to reflect the values of their operands, which may be on the top of
   the stack, in the X register, or in memory.  Each instruction that affects
   the condition code field commonly uses one of the following patterns to set
   the field:

     CCA (arithmetic) sets CC to:
       - CCG (00) if the operand is > 0
       - CCL (01) if the operand is < 0
       - CCE (10) if the operand is = 0

     CCB (byte) sets CC to:
       - CCG (00) if the operand is numeric (byte value is 060-071)
       - CCL (01) if the operand is any other character
       - CCE (10) if the operand is alphabetic (0101-0132 or 0141-0172)

     CCC (comparison) sets CC to:
       - CCG (00) if operand 1 is > operand 2
       - CCL (01) if operand 1 is < operand 2
       - CCE (10) if operand 1 is = operand 2

     CCD (direct I/O) sets CC to:
       - CCG (00) if the device is not ready
       - CCL (01) if the device controller is not responding
       - CCE (10) if the device is ready and controller is responding

   The operand(s) may be a byte, word, double word, triple word, or quadruple
   word, and may be signed (integer) or unsigned (logical).

   Statement macros are provided to set the condition code explicitly, as well
   as to set the code using patterns CCA, CCB, or CCC.  The CCA macro takes two
   parameters: an upper word and a lower word.  The CCB macro takes a single
   byte parameter.  The CCC macro takes four parameters: an upper and lower word
   for each of the two operands.

   No macro is provided for pattern CCD, as that pattern reflects the
   operational status of the target device and interface, rather than the
   condition of an operand.  The I/O instructions use the explicit macros to set
   the codes for pattern D.
*/

#define SET_CCG             STA = STA & ~STATUS_CC_MASK | STATUS_CCG
#define SET_CCL             STA = STA & ~STATUS_CC_MASK | STATUS_CCL
#define SET_CCE             STA = STA & ~STATUS_CC_MASK | STATUS_CCE


/* Condition code pattern CCA (arithmetic).

   Semantics:

     if operand > 0
       then CCG
     else if operand < 0
       then CCL
     else
       CCE

   Macro:

     SET_CCA (upper, lower)

   Implementation:

     if upper = 0 or lower = 0
       then CCE
     else if upper.sign = 1
       then CCL
     else CCG

   Usage:

     SET_CCA (RA, 0)            -- a 16-bit integer value
     SET_CCA (0, RA)            -- a 16-bit logical value
     SET_CCA (RB, RA)           -- a 32-bit integer value
     SET_CCA (0, RB | RA)       -- a 32-bit logical value
     SET_CCA (RC, RB | RA)      -- a 48-bit integer value
     SET_CCA (RD, RC | RB | RA) -- a 64-bit integer value


   Implementation notes:

    1. When either "upper" or "lower" is 0, the corresponding test is optimized
       away.
*/

#define SET_CCA(u,l)        STA = STA & ~STATUS_CC_MASK | \
                              (((u) | (l)) == 0 \
                                ? STATUS_CCE \
                                : (u) & D16_SIGN \
                                    ? STATUS_CCL \
                                    : STATUS_CCG)


/* Condition code pattern CCB (byte).

   Semantics:

     if operand in 060 .. 071
       then CCG
     else if operand in 0101 .. 0132 or operand in 0141 .. 0172
       then CCE
     else
       CCL

   Macro:

     SET_CCB (byte)

   Implementation:

     CCx = ccb_lookup_table [byte]

   Usage:

     SET_CCB (RA)


   Implementation notes:

    1. The byte parameter is not masked before being used as an array index, so
       the caller must ensure that the value is between 0 and 255.

    2. The table value must be masked if it is to be stored in the status
       register because CCG is returned as 100 (base 2) to ensure that all
       values are disjoint for the MVBW and Bcc instructions.
*/

#define SET_CCB(b)          STA = STA & ~STATUS_CC_MASK | \
                              cpu_ccb_table [(b)] & STATUS_CC_MASK


/* Condition code pattern CCC (conditional).

   Semantics:

     if operand_1 > operand_2
       then CCG
     else if operand_1 < operand_2
       then CCL
     else
       CCE

   Macro:

     SET_CCC (upper_1, lower_1, upper_2, lower_2)

   Implementation:

     if upper_1 = upper_2
       then
         if lower_1 = lower_2
           then CCE
         else if lower_1 < lower_2
           then CCL
         else CCG
     else if |upper_1| < |upper_2|
       then CCL
     else CCG

   Usage:

     SET_CCC (RA,  0, RB,  0) -- 16-bit integer comparison
     SET_CCC ( 0, RA,  0, RB) -- 16-bit logical comparison
     SET_CCC (RB, RA, RD, RC) -- 32-bit integer comparison


   Implementation notes:

    1. When any of the parameters are 0, the corresponding tests are optimized
       away.

    2. Complementing the signs of the upper words allows an unsigned comparison
       to substitute for a signed comparison.
*/

#define SET_CCC(u1,l1,u2,l2)  STA = STA & ~STATUS_CC_MASK | \
                                ((u1) == (u2) \
                                  ? ((l1) == (l2) \
                                      ? STATUS_CCE \
                                      : ((l1) < (l2) \
                                          ? STATUS_CCL \
                                          : STATUS_CCG)) \
                                  : (((u1) ^ D16_SIGN) < ((u2) ^ D16_SIGN) \
                                      ? STATUS_CCL \
                                      : STATUS_CCG))

/* Set carry and overflow.

   These macros are used by arithmetic operations to set the carry and overflow
   bits in the status register.  In addition to setting the overflow bit, if
   user traps are enabled, the Integer Overflow bit in the CPX1 register is set,
   which will cause an interrupt
*/

#define SET_CARRY(b)        STA = ((b) ? STA | STATUS_C : STA & ~STATUS_C)

#define SET_OVERFLOW(b)     if (b) { \
                                STA |= STATUS_O; \
                                if (STA & STATUS_T) \
                                    CPX1 |= cpx1_INTOVFL; \
                                } \
                            else \
                                STA &= ~STATUS_O

/* SR preadjust.

   The PREADJUST_SR macro is called to ensure that the correct number of TOS
   registers are valid prior to instruction execution.
*/

#define PREADJUST_SR(n)     if (SR < (n)) \
                                cpu_adjust_sr (n); \
                            else \
                                (void) 0


/* Machine instruction bit-field accessors */

#define BITS_0_3_MASK       0170000u            /* bits 0-3 mask */
#define BITS_0_3_SHIFT      12                  /* bits 0-3 alignment shift */

#define BITS_0_3(v)         (((v) & BITS_0_3_MASK) >> BITS_0_3_SHIFT)

#define BITS_4_5_MASK       0006000u            /* bits 4-5 mask */
#define BITS_4_5_SHIFT      10                  /* bits 4-5 alignment shift */

#define BITS_4_5(v)         (((v) & BITS_4_5_MASK) >> BITS_4_5_SHIFT)

#define BITS_4_7_MASK       0007400u            /* bits 4-7 mask */
#define BITS_4_7_SHIFT      8                   /* bits 4-7 alignment shift */

#define BITS_4_7(v)         (((v) & BITS_4_7_MASK) >> BITS_4_7_SHIFT)

#define BITS_4_9_MASK       0007700u            /* bits 4-9 mask */
#define BITS_4_9_SHIFT      6                   /* bits 4-9 alignment shift */

#define BITS_4_9(v)         (((v) & BITS_4_9_MASK) >> BITS_4_9_SHIFT)

#define BITS_5_9_MASK       0003700u            /* bits 5-9 mask */
#define BITS_5_9_SHIFT      6                   /* bits 5-9 alignment shift */

#define BITS_5_9(v)         (((v) & BITS_5_9_MASK) >> BITS_5_9_SHIFT)

#define BITS_8_11_MASK      0000360u            /* bits 8-11 mask */
#define BITS_8_11_SHIFT     4                   /* bits 8-11 alignment shift */

#define BITS_8_11(v)        (((v) & BITS_8_11_MASK) >> BITS_8_11_SHIFT)

#define BITS_8_12_MASK      0000370u            /* bits 8-12 mask */
#define BITS_8_12_SHIFT     3                   /* bits 8-12 alignment shift */

#define BITS_8_12(v)        (((v) & BITS_8_12_MASK) >> BITS_8_12_SHIFT)

#define BITS_10_15_MASK     0000077u            /* bits 10-15 mask */
#define BITS_10_15_SHIFT    0                   /* bits 10-15 alignment shift */

#define BITS_10_15(v)       (((v) & BITS_10_15_MASK) >> BITS_10_15_SHIFT)

#define BITS_12_15_MASK     0000017u            /* bits 12-15 mask */
#define BITS_12_15_SHIFT    0                   /* bits 12-15 alignment shift */

#define BITS_12_15(v)       (((v) & BITS_12_15_MASK) >> BITS_12_15_SHIFT)

#define BITS_13_15_MASK     0000007u            /* bits 13-15 mask */
#define BITS_13_15_SHIFT    0                   /* bits 13-15 alignment shift */

#define BITS_13_15(v)       (((v) & BITS_13_15_MASK) >> BITS_13_15_SHIFT)


/* Instruction-class accessors */

#define SUBOP_MASK          BITS_0_3_MASK       /* subopcode mask */
#define SUBOP_SHIFT         BITS_0_3_SHIFT      /* subopcode alignment shift */
#define SUBOP(v)            BITS_0_3(v)         /* subopcode accessor */

#define STACKOP_A_MASK      BITS_4_9_MASK       /* stack operation A mask */
#define STACKOP_A_SHIFT     BITS_4_9_SHIFT      /* stack operation A alignment shift */
#define STACKOP_A(v)        BITS_4_9(v)         /* stack operation A accessor */

#define STACKOP_B_MASK      BITS_10_15_MASK     /* stack operation B mask */
#define STACKOP_B_SHIFT     BITS_10_15_SHIFT    /* stack operation B alignment shift */
#define STACKOP_B(v)        BITS_10_15(v)       /* stack operation B accessor */

#define SBBOP_MASK          BITS_5_9_MASK       /* shift/branch/bit operation mask */
#define SBBOP_SHIFT         BITS_5_9_SHIFT      /* shift/branch/bit operation alignment shift */
#define SBBOP(v)            BITS_5_9(v)         /* shift/branch/bit operation accessor */

#define MSFIFROP_MASK       BITS_4_7_MASK       /* move/special/firmware/immediate/field/register operation mask */
#define MSFIFROP_SHIFT      BITS_4_7_SHIFT      /* move/special/firmware/immediate/field/register operation alignment shift */
#define MSFIFROP(v)         BITS_4_7(v)         /* move/special/firmware/immediate/field/register operation accessor */

#define MSSUBOP_MASK        BITS_8_12_MASK      /* move/special suboperation mask */
#define MSSUBOP_SHIFT       BITS_8_12_SHIFT     /* move/special suboperation alignment shift */
#define MSSUBOP(v)          BITS_8_12(v)        /* move/special suboperation accessor */

#define SPECOP_MASK         BITS_12_15_MASK     /* special operation mask */
#define SPECOP_SHIFT        BITS_12_15_SHIFT    /* special operation alignment shift */
#define SPECOP(v)           BITS_12_15(v)       /* special operation accessor */

#define IOCPIMOP_MASK       BITS_4_7_MASK       /* I-O/control/program/immediate/memory operation mask */
#define IOCPIMOP_SHIFT      BITS_4_7_SHIFT      /* I-O/control/program/immediate/memory operation alignment shift */
#define IOCPIMOP(v)         BITS_4_7(v)         /* I-O/control/program/immediate/memory operation accessor */

#define IOCSUBOP_MASK       BITS_8_11_MASK      /* I-O/control suboperation mask */
#define IOCSUBOP_SHIFT      BITS_8_11_SHIFT     /* I-O/control suboperation alignment shift */
#define IOCSUBOP(v)         BITS_8_11(v)        /* I-O/control suboperation accessor */

#define CNTLOP_MASK         BITS_12_15_MASK     /* control operation mask */
#define CNTLOP_SHIFT        BITS_12_15_SHIFT    /* control operation alignment shift */
#define CNTLOP(v)           BITS_12_15(v)       /* control operation accessor */

#define MLBOP_MASK          BITS_0_3_MASK       /* memory/loop/branch operation mask */
#define MLBOP_SHIFT         BITS_0_3_SHIFT      /* memory/loop/branch operation alignment shift */
#define MLBOP(v)            BITS_0_3(v)         /* memory/loop/branch operation accessor */

#define FIRMEXTOP_MASK      BITS_8_11_MASK      /* firmware extension operation mask */
#define FIRMEXTOP_SHIFT     BITS_8_11_SHIFT     /* firmware extension operation alignment shift */
#define FIRMEXTOP(v)        BITS_8_11(v)        /* firmware extension operation accessor */

#define FMEXSUBOP_MASK      BITS_12_15_MASK     /* firmware extension suboperation mask */
#define FMEXSUBOP_SHIFT     BITS_12_15_SHIFT    /* firmware extension suboperation alignment shift */
#define FMEXSUBOP(v)        BITS_12_15(v)       /* firmware extension suboperation accessor */

#define EISOP_MASK          BITS_13_15_MASK     /* Extended Instruction Set operation mask */
#define EISOP_SHIFT         BITS_13_15_SHIFT    /* Extended Instruction Set operation alignment shift */
#define EISOP(v)            BITS_13_15(v)       /* Extended Instruction Set operation accessor */


/* General instruction accessors */

#define IOOP_K_MASK         0000017u            /* I/O K-field mask */
#define IOOP_K_SHIFT        0000000u            /* I/O K-field alignment shift */
#define IO_K(v)             (((v) & IOOP_K_MASK) >> IOOP_K_SHIFT)

#define X_FLAG              0004000u            /* index flag in bit 4 */
#define I_FLAG_BIT_4        0004000u            /* indirect flag in bit 4 */
#define I_FLAG_BIT_5        0002000u            /* indirect flag in bit 5 */
#define M_FLAG              0001000u            /* memory subop flag in bit 6 */

#define START_BIT_MASK      0000360u            /* start bit mask for bit field instructions */
#define START_BIT_SHIFT     4                   /* start bit alignment shift */
#define START_BIT(v)        (((v) & START_BIT_MASK) >> START_BIT_SHIFT)

#define BIT_COUNT_MASK      0000017u            /* bit count mask for bit field instructions */
#define BIT_COUNT_SHIFT     0                   /* bit count alignment shift */
#define BIT_COUNT(v)        (((v) & BIT_COUNT_MASK) >> BIT_COUNT_SHIFT)

#define BIT_POSITION_MASK   0000077u            /* bit position mask for bit test instructions */
#define BIT_POSITION_SHIFT  0                   /* bit position alignment shift */
#define BIT_POSITION(v)     (((v) & BIT_POSITION_MASK) >> BIT_POSITION_SHIFT)

#define SHIFT_COUNT_MASK    0000077u            /* shift count mask for shift instructions */
#define SHIFT_COUNT_SHIFT   0                   /* shift count alignment shift */
#define SHIFT_COUNT(v)      (((v) & SHIFT_COUNT_MASK) >> SHIFT_COUNT_SHIFT)

#define SHIFT_RIGHT_FLAG    0000100u            /* shift instructions left/right (0/1) flag */

#define MODE_DISP_MASK      0001777u            /* memory-reference mode and displacement mask */
#define MODE_MASK           0001700u            /* memory-reference mode mask */
#define MODE_SHIFT          6                   /* memory-reference mode alignment shift */

#define MODE_P_POS          0                   /* P + displacement mode flag */
#define MODE_P_NEG          HP_BIT  (7)         /* P - displacement mode flag */
#define MODE_DB_POS         HP_BIT  (6)         /* DB + displacement mode flag */
#define MODE_Q_POS          HP_BITS (6, 7)      /* Q + displacement mode flag */
#define MODE_Q_NEG          HP_BITS (6, 8)      /* Q - displacement mode flag */
#define MODE_S_NEG          HP_BITS (6, 9)      /* S - displacement mode flag */
#define MODE_DATA           HP_BIT  (6)         /* data access mode flag */

#define DISPL_31_SIGN       0000040u            /* sign bit for 0-31 displacements */
#define DISPL_255_SIGN      0000400u            /* sign bit for 0-255 displacements */

#define DISPL_31_MASK       0000037u            /* mask for 0-31 displacements */
#define DISPL_63_MASK       0000077u            /* mask for 0-63 displacements */
#define DISPL_127_MASK      0000177u            /* mask for 0-127 displacements */
#define DISPL_255_MASK      0000377u            /* mask for 0-255 displacements */

#define DISPL_P_FLAG        0001000u            /* P-relative displacement flag */
#define DISPL_DB_FLAG       0000400u            /* DB-relative displacement flag */
#define DISPL_QPOS_FLAG     0000200u            /* positive Q-relative displacement flag */
#define DISPL_QNEG_FLAG     0000100u            /* negative Q-relative displacement flag */

#define IMMED_MASK          0000377u            /* mask for immediate values */

#define SDEC2_MASK          0000003u            /* two-bit S-decrement mask for move instructions */
#define SDEC3_MASK          0000007u            /* three-bit S-decrement mask for move instructions */
#define SDEC_SHIFT          0                   /* S-decrement alignment shift */

#define SDEC2(v)            (((v) & SDEC2_MASK) >> SDEC_SHIFT)
#define SDEC3(v)            (((v) & SDEC3_MASK) >> SDEC_SHIFT)

#define DB_FLAG             0000020u            /* base set PB/DB base register flag */
#define CIS_DB_FLAG         0000001u            /* CIS set PB/DB base register flag */


/* Explicit instruction opcodes and accessors */

#define UNPAIRED_STACKOP    ~0u                 /* an unpaired stackop was parsed */

#define NOP                 0000000u            /* no operation */
#define QASR                0015700u            /* quadruple arithmetic right shift */
#define MVBL                0020100u            /* move from DB to DL */
#define HPIB_CLK            0020104u            /* first word of the two-word HP-IB clock instructions */
#define RSW                 0020300u            /* read switch register */
#define HPIB_IO             0020302u            /* first word of the two-word HP-IB instructions */
#define DMUL                0020570u            /* double integer multiply */
#define DDIV                0020571u            /* double integer divide */
#define SETR                0027400u            /* set registers (none) */
#define SETR_X              0027404u            /* set registers (index) */
#define SED_1               0030041u            /* set enable interrupt */
#define XCHD                0030060u            /* exchange DB and TOS */
#define HALT_10             0030370u            /* halt 10 */
#define PCAL_0              0031000u            /* procedure call using TOS label */
#define LRA                 0170000u            /* load relative address */

#define MTFDS_MASK          0177730u            /* move to/from data segment mask */
#define MTFDS               0020130u            /* move to/from data segment */

#define EXIT_MASK           0177400u            /* exit procedure mask */
#define EXIT                0031400u            /* exit procedure */

#define PAUS_MASK           0177760u            /* pause mask */
#define PAUS                0030020u            /* pause */

#define BR_MASK             0173000u            /* conditional and unconditional branch mask */
#define BR_DBQS_I           0143000u            /* branch unconditionally DB/Q/S-relative indirect */
#define BCC                 0141000u            /* branch conditionally */
#define BR                  0140000u            /* branch unconditionally */
#define BCC_CCF_SHIFT       2                   /* CCF alignment in BCC instruction */

#define LSDX_MASK           0175000u            /* load/store double-word indexed mask */
#define LDD_X               0155000u            /* load double-word indexed */
#define STD_X               0165000u            /* store double-word indexed */

#define TBR_MASK            0177000u            /* test and branch mask */
#define TBA                 0050000u            /* test and branch, limit in A */
#define MTBA                0052000u            /* modify, test and branch, limit in A */
#define TBX                 0054000u            /* test and branch, limit in X */
#define MTBX                0056000u            /* modify, test and branch, limit in X */

#define CMD_TO_MASK         0000007u            /* CMD command word TO field mask */
#define CMD_MOP_MASK        0000060u            /* CMD command word MOP field mask */

#define CMD_TO_SHIFT        0                   /* CMD command word TO field alignment shift */
#define CMD_MOP_SHIFT       4                   /* CMD command word MOP field alignment shift */

#define CMD_TO(v)           (((v) & CMD_TO_MASK)  >> CMD_TO_SHIFT)
#define CMD_MOP(v)          (((v) & CMD_MOP_MASK) >> CMD_MOP_SHIFT)

#define MVBW_FLAGS          0000034u            /* MVBW flags */
#define MVBW_CCF            0000030u            /* MVBW condition code flags */
#define MVBW_N_FLAG         0000020u            /* MVBW numeric flag */
#define MVBW_A_FLAG         0000010u            /* MVBW alphabetic flag */
#define MVBW_S_FLAG         0000004u            /* MVBW upshift flag */
#define MVBW_CCF_SHIFT      6                   /* CCF alignment in MVBW instruction */

#define NABS_FLAG           0000100u            /* CVDA negative absolute value flag */
#define ABS_FLAG            0000040u            /* CVDA absolute value flag */

#define CVND_SC_MASK        0000016u            /* CVND sign-control mask */
#define CVND_SC_SHIFT       1                   /* CVND sign-control field alignment shift */

#define EIS_SDEC_FLAG       0000020u            /* EIS S-decrement bit 11 */
#define EIS_SDEC_MASK       0000060u            /* EIS S-decrement mask */
#define EIS_SDEC_SHIFT      4                   /* EIS S-decrement alignment shift */

#define EIS_SDEC(v)         (((v) & EIS_SDEC_MASK) >> EIS_SDEC_SHIFT)

#define CIS_SDEC_MASK       0000001u            /* CIS S-decrement mask */
#define CIS_SDEC_SHIFT      0                   /* CIS S-decrement alignment shift */

#define TCCS_CCF_SHIFT      8                   /* CIS TCCS instruction CCF alignment */

#define CIS_XBR_MASK        0177774u            /* XBR/PARC/ENDP mask */
#define CIS_XBR_GROUP       0020474u            /* external branch group */


/* PSHR/SETR instruction accessors */

#define PSR_RL_MASK         0000001u            /* PSHR/SETR register right-to-left mask */
#define PSR_LR_MASK         0000200u            /* PSHR/SETR register left-to-right mask */

#define PSR_SBANK           0000200u            /* Stack bank register */
#define PSR_DB_DBANK        0000100u            /* Data base and data bank registers */
#define PSR_DL              0000040u            /* Data limit register */
#define PSR_Z               0000020u            /* Stack limit register */
#define PSR_STA             0000010u            /* Status register */
#define PSR_X               0000004u            /* Index register */
#define PSR_Q               0000002u            /* Frame pointer */
#define PSR_S               0000001u            /* Stack pointer */

#define PSR_PRIV            (PSR_SBANK | PSR_DB_DBANK | PSR_DL | PSR_Z)


/* PCN instruction result values */

#define PCN_SERIES_II       1                   /* CPU number for the Series II */
#define PCN_SERIES_III      2                   /* CPU number for the Series III */
#define PCN_SERIES_4X_5X    3                   /* CPU number for the Series 4x and 5x */


/* LST and SST instruction accessors.

   The format of the SYSGLOB pointer words differs, depending on whether or not
   the CST Expansion firmware is installed.  If the firmware is not installed,
   then the format is:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                        offset - %1000                         |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   If the firmware is installed, then the format is:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |           (offset - %1000) * 32           |       bank        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The new-format offset uses a 32-word alignment, so to get the actual offset,
   the bank bits are set to zero.
*/

#define LST_OFFSET_MASK     HP_BITS (0, 10)

#define LST_BANK(w)         HP_BITS_TO (11, 15, w)
#define LST_OFFSET(w)       ((w) & LST_OFFSET_MASK)


/* Reserved memory addresses */

#define CSTB_POINTER        0000000u            /* code segment table base pointer */
#define CSTX_POINTER        0000001u            /* code segment table extension pointer */
#define DST_POINTER         0000002u            /* data segment table pointer */
#define ICS_Q               0000005u            /* interrupt control stack marker pointer (QI) */
#define ICS_Z               0000006u            /* interrupt control stack limit (ZI) */
#define INTERRUPT_MASK      0000007u            /* interrupt mask */
#define TOTAL_REGISTER      0000023u            /* time since last soft timer interrupt */
#define LIMIT_REGISTER      0000025u            /* system clock limit register */
#define STATUS_REGISTER     0000026u            /* system clock status register */
#define SYSGLOB             0001000u            /* system global tables pointer */

#define FIRMWARE_SET        (SYSGLOB + 0220u)   /* mapping firmware flags */
#define LSTT_BANK           (SYSGLOB + 0221u)   /* LSTT bank address */
#define LSTT_OFFSET         (SYSGLOB + 0222u)   /* LSTT offset address */
#define CODE_SEG_COUNT      (SYSGLOB + 0223u)   /* count of current process code segments */

#define SINC_REQUEST        HP_BIT (0)          /* the SINC instruction requested a timer interrupt */


/* Code Segment Table accessors.

   Each CST entry from 1-N contains four words, formatted as follows:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | A | M | R | T |               Segment length / 4              |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | - | O | I | - | - | - | S | C |                               |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         Bank address                          |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         Base address                          |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     A = Absent    (1 = segment is absent from memory)
     M = Mode      (1 = segment executes in privileged mode only)
     R = Reference (1 = segment has been referenced; set by microcode)
     T = Trace     (1 = segment is to be traced)

     O = Overlay   (1 = segment is a recoverable overlay candidate)
     I = In        (1 = segment is in motion in)
     S = System    (1 = segment is a system segment)
     C = Core      (1 = segment is core-resident)

     Bank address = memory bank in which segment resides
     Base address = absolute address of PB within bank
*/

#define CST_A_BIT           HP_BIT  (0)         /* code segment is absent */
#define CST_M_BIT           HP_BIT  (1)         /* code segment is privileged */
#define CST_R_BIT           HP_BIT  (2)         /* code segment has been referenced flag */
#define CST_T_BIT           HP_BIT  (3)         /* code segment is to be traced */
#define CST_SEGLEN_MASK     HP_BITS (4, 15)     /* code segment length / 4 */

#define CST_RESERVED        0000300u            /* number of CST entries reserved for the system */


/* Data Segment Table accessors.

   Each DST entry contains four words, formatted as follows:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | A | C | R |                   Segment length / 4              |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | V | O | I | K | M | F | S | C | D |       VM allocation       |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         Bank address                          |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         Base address                          |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     A = Absent    (1 = segment is absent from memory)
     C = Clean bit (1 = data has not been modified)
     R = Reference (1 = segment has been referenced; set by microcode)

     V = Valid     (1 = disc copy is valid)
     O = Overlay   (1 = segment is a recoverable overlay candidate)
     I = In        (1 = segment is in motion in)
     K = Stack     (1 = segment is a stack)
     M = Mod       (1 = a segment change in size or location is requested)
     F = Force     (1 = a forced write of this segment is in progress)
     S = System    (1 = segment is a system segment)
     C = Core      (1 = segment is core-resident)
     D = Disabled  (1 = write is disabled)
     VM allocation    = number of VM pages allocated to this segment

     Bank address = memory bank in which segment resides
     Base address = absolute segment address within bank
*/

#define DST_A_BIT           HP_BIT  (0)         /* data segment is absent */
#define DST_C_BIT           HP_BIT  (1)         /* data segment is clean (not modified) */
#define DST_R_BIT           HP_BIT  (2)         /* data segment has been referenced */
#define DST_SEGLEN_MASK     HP_BITS (3, 15)     /* data segment length / 4 */


/* Segment Transfer Table accessors.

   The format of the STT header differs, depending on whether or not the CST
   Expansion firmware is installed.  If the firmware is not installed, then the
   format is:

     STT Length (entry 0)

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0 | U | 0   0   0   0   0   0 |       Number of labels        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     U = Uncallable (i.e., callable from privileged mode only)

   If the firmware is installed, then the format is:

     STT Length (entry 0)

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |    Number of local labels     |    Total number of labels     |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#define STT_LABELS(h)       HP_BITS_TO (8, 15, h)       /* STT number of labels */
#define STT_LOCALS(h)       HP_BITS_TO (0,  7, h)       /* STT number of local labels */


/* Stack marker accessors.

   PCAL and interrupts push a four-word marker onto the stack:

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                       X register value                        |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | T | M |          PB-relative return address                   |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                     Status register value                     |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | D |                     Delta Q value                         |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     T = a trace or control-Y interrupt is pending
     M = the code segment is physically mapped
     D = the dispatcher was interrupted

   The T and D bits are set by the operating system, if applicable, after the
   stack marker was originally written.  The M bit is set when the marker is
   written and indicates the CST mapping mode for the segment number in the
   Status Register value; it is significant only if CST Expansion firmware is
   installed.
*/

#define STMK_D              HP_BIT  (0)         /* dispatcher flag */
#define STMK_T              HP_BIT  (0)         /* trace flag */
#define STMK_M              HP_BIT  (1)         /* physical mapping flag */

#define STMK_OFFSET(p)      HP_BITS_TO (2, 15, p)       /* PB-relative return address */
#define STMK_RETURN(p)      TO_HP_BITS (2, 15, p)       /* PB-relative return address */


/* Program label accessors.

   The format of program labels differs, depending on whether or not the CST
   Expansion firmware is installed.  If the firmware is not installed, then the
   formats are:

     Local Program Label

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0 | U |                  PB-relative address                  |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     External Program Label

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1 |        STT Number         |        Segment Number         |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   If the firmware is installed, then the formats are:

     Local Program Label

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0 | U |                  PB-relative address                  |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     External Program Label

       0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | M |        STT Number         |        Segment Number         |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     U              = Uncallable (1 = procedure callable from privileged mode only)
     M              = Mapped     (1 = segment number is physically mapped)
     Address        = PB-relative address of procedure entry
     STT Number     = STT entry within target segment
     Segment Number = number of target segment


   Implementation notes:

    1. With CST expansion firmware installed, local labels and logically mapped
       external labels are indistinguishable without knowing the locations of
       the labels in the STT and the count of local labels in the STT header.
       To avoid having to handle the format differences at every use, we save
       the internal/external state in a program label descriptor structure when
       the label is obtained from the STT.

       In places where labels are created on-the-fly, such as the LLBL
       conversion of local labels to external labels, firmware presence tests
       are still required.
*/

#define ISR_SEGN            1                           /* interrupt service routines segment number */
#define LABEL_STTN_MAX      127                         /* STT number maximum value */

#define LABEL_PHYSICAL      HP_BIT (0)                  /* external program physically mapped flag */
#define LABEL_LOGICAL       0                           /* external program logically mapped flag */

#define TO_LABEL_MAP(m)     ((m) << 1 & LABEL_PHYSICAL) /* convert stack mapping bit to label mapping bit */
#define TO_STMK_MAP(l)      ((l) >> 1 & STMK_M)         /* convert label mapping bit to stack mapping bit */

#define LABEL_EXTERNAL      HP_BIT (0)                  /* external program label flag */
#define LABEL_STTN(l)       HP_BITS_TO (1,  7, l)       /* external program label STT number */
#define LABEL_SEGN(l)       HP_BITS_TO (8, 15, l)       /* external program label segment number */

#define TO_LABEL(s,n)       (TO_HP_BITS (1,  7, n) | TO_HP_BITS (8, 15, s))

#define LABEL_LOCAL         0                           /* local program label flag */
#define LABEL_UNCALLABLE    HP_BIT (1)                  /* local program label uncallable flag */
#define LABEL_OFFSET(l)     HP_BITS_TO (2, 15, l)       /* local program label address */

#define TO_ISR_LABEL(n)     (LABEL_PHYSICAL | TO_LABEL (ISR_SEGN, n))

typedef enum {                                  /* label accessing modes */
    Local    = 0,                               /*   local access */
    External = 1                                /*   external access */
    } LABEL_MODE;

typedef struct {                                /* program label descriptor */
    HP_WORD     label;                          /*   the program label */
    LABEL_MODE  mode;                           /*   the label accessing mode */
    } PLABEL;


/* Trace identifier definitions */

typedef enum {
    Trace_CPU_Instruction,
    Trace_CPU_Stack,
    Trace_CPU_Segment,
    Trace_CPU_Trap,
    Trace_CPU_Interrupt,
    Trace_CPU_STT_Entry,
    Trace_CPU_STT_Lookup,
    Trace_CPU_STT_Resolver,
    Trace_CPU_Service,
    Trace_CPU_Speed,
    Trace_CPU_Separator,
    Trace_CPU_Stop,

    Trace_CPU_Dispatch,
    Trace_CPU_LRA,
    Trace_CPU_LST,
    Trace_CPU_SST,
    Trace_CPU_IXIT,

    Trace_CPU_CIS_Operand,
    Trace_CPU_CIS_Edit,

    Trace_CPU_EIS_Decimal,
    Trace_CPU_EIS_Binary
    } CPU_TRACE_ID;


/* CPU registers */

extern HP_WORD CIR;                             /* Current Instruction Register */
extern HP_WORD NIR;                             /* Next Instruction Register */
extern HP_WORD PB;                              /* Program Base Register */
extern HP_WORD P;                               /* Program Counter */
extern HP_WORD PL;                              /* Program Limit Register */
extern HP_WORD PBANK;                           /* Program Segment Bank Register */
extern HP_WORD DL;                              /* Data Limit Register */
extern HP_WORD DB;                              /* Data Base Register */
extern HP_WORD DBANK;                           /* Data Segment Bank Register */
extern HP_WORD Q;                               /* Stack Marker Register */
extern HP_WORD SM;                              /* Stack Memory Register */
extern HP_WORD SR;                              /* Stack Register Counter */
extern HP_WORD Z;                               /* Stack Limit Register */
extern HP_WORD SBANK;                           /* Stack Segment Bank Register */
extern HP_WORD TR [4];                          /* Top of Stack Registers */
extern HP_WORD X;                               /* Index Register */
extern HP_WORD STA;                             /* Status Register */
extern HP_WORD SWCH;                            /* Switch Register */
extern HP_WORD CPX1;                            /* Run-Mode Interrupt Flags Register */
extern HP_WORD CPX2;                            /* Halt-Mode Interrupt Flags Register */
extern HP_WORD MOD;                             /* Module Control Register */
extern HP_WORD PCLK;                            /* Process Clock Register */
extern HP_WORD CNTR;                            /* Microcode Counter */
extern HP_WORD CR;                              /* Timer Count Register */
extern HP_WORD SP0;                             /* Scratch Pad Register 0 */
extern HP_WORD SIR;                             /* Status and Interrupt Register */
extern HP_WORD CSMAP;                           /* Code Segment Mapping Register */

/* Top of stack register names */

#define RA                  TR [0]
#define RB                  TR [1]
#define RC                  TR [2]
#define RD                  TR [3]


/* CPU device */

extern DEVICE cpu_dev;                          /* Central Processing Unit */


/* CPU global state */

extern jmp_buf     cpu_save_env;                /* saved environment for microcode aborts */
extern uint32      cpu_stop_flags;              /* set of simulation stop flags */
extern HP_WORD     cpu_bank_mask;               /* mask for the bank field of CST and DST */
extern HP_WORD     cpu_sir_irq_set;             /* mask for the set of SIR interrupt sources */

extern POWER_STATE cpu_power_state;             /* power supply state */
extern EXEC_STATE  cpu_micro_state;             /* micromachine execution state */
extern t_bool      cpu_base_changed;            /* TRUE if any base register has been changed */


/* Condition Code B mapping table */

extern const HP_WORD cpu_ccb_table [256];       /* byte-value to condition-code map */


/* CPU global instruction support routines */

extern void cpu_push       (void);
extern void cpu_pop        (void);
extern void cpu_queue_up   (void);
extern void cpu_queue_down (void);
extern void cpu_flush      (void);
extern void cpu_adjust_sr  (uint32 target);
extern void cpu_mark_stack (void);

extern void   cpu_ea      (HP_WORD mode_disp, ACCESS_CLASS *class, HP_WORD *offset, BYTE_SELECTOR *selector);
extern uint32 cpu_byte_ea (ACCESS_CLASS class, uint32 byte_offset, uint32 block_length);

extern void   cpu_setup_irq_handler  (IRQ_CLASS class, HP_WORD parameter);
extern void   cpu_setup_ics_irq      (IRQ_CLASS class, TRAP_CLASS trap);
extern void   cpu_setup_code_segment (PLABEL plab, HP_WORD *status, HP_WORD *entry_0);
extern void   cpu_setup_data_segment (HP_WORD segment_number, HP_WORD *bank, HP_WORD *address);
extern PLABEL cpu_read_label         (HP_WORD stt_index);
extern void   cpu_call_procedure     (PLABEL plab, HP_WORD offset);
extern void   cpu_update_pclk        (void);


/* CPU global instruction execution routines */

extern t_bool cpu_interrupt_pending (t_stat *status);
extern t_stat cpu_branch_short      (t_bool check_loop);

extern HP_WORD cpu_add_16 (HP_WORD augend,       HP_WORD addend);
extern HP_WORD cpu_sub_16 (HP_WORD minuend,      HP_WORD subtrahend);
extern HP_WORD cpu_mpy_16 (HP_WORD multiplicand, HP_WORD multiplier);

extern t_stat cpu_stack_op                      (void);
extern t_stat cpu_shift_branch_bit_op           (void);
extern t_stat cpu_move_spec_fw_imm_field_reg_op (void);
extern t_stat cpu_io_cntl_prog_imm_mem_op       (void);
extern t_stat cpu_eis_fp_op                     (void);
extern t_stat cpu_eis_dec_op                    (void);
extern t_stat cpu_cis_op                        (void);
extern uint32 cpu_cis_edit_count                (uint8 opcode, uint8 operand);


/* CPU global subsidiary tracing routines */

extern t_stat cpu_base_trace (TRACE_RECORD *eptr);
extern t_stat cpu_cis_trace  (TRACE_RECORD *eptr);
extern t_stat cpu_eis_trace  (TRACE_RECORD *eptr);
