/* hp3000_trace.c: HP 3000 system tracing interface

   Copyright (c) 2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   13-Aug-24    JDB     Created


   This module supports memory-buffered tracing as a replacement for the earlier
   trace operations that wrote directly to the debug log file.  Direct writing
   is simpler and worked well.  However, it imposed a severe performance penalty
   on execution, sometimes to the point that extensive tracing was not a viable
   means of capturing the internal execution process.  It also tended to
   generate huge log files, while often only the last part of which was
   significant.

   Buffered tracing separates the capture of trace data with the formatting and
   logging of that data.  Trace data is now captured to a circular memory buffer
   with a user-specified capacity.


   Conditions:

     - stopped
     - executing, waiting for trigger
     - executing, triggered, in process
     - stopped, triggered, complete
     - halted

     base  limit  current  trigger  start      end        Condition
     ----  -----  -------  -------  ---------  ---------  -------------------------------
     NULL  NULL      X        X       X         X         Memory deallocated
     set   set      NULL     NULL   NULL       NULL       Memory allocated, stopped
     set   set    base-1     NULL   NULL       NULL       Executing, waiting for trigger
     set   set    base-1     base   base       limit      Executing, triggered (always)
     set   set    -> NULL  current  base       base-1     Halted, trace empty
     set   set    -> NULL  current  base       current    Halted, not wrapped
     set   set    -> NULL  current  current+1  current    Halted, wrapped
     set   set       X     current  trigger    limit      Start triggered, not wrapped
     set   set       X     current  trigger    trigger-1  Start triggered, wrapped
     set   set       X     current  base       limit      Center triggered, not wrapped
     set   set       X     current  trigger/2  start-1    Center triggered, wrapped
     set   set       X     current  base       trigger    End triggered, not wrapped
     set   set       X     current  trigger+1  trigger    End triggered, wrapped
     set   set     NULL      set       set       set      Trace complete


   Commands:

     TRACE MEMORY=<state-count>

     TRACE [ -N ] LIST
     TRACE [ -N ] LIST=<filename>
     TRACE LIST NOTHING

     TRACE TRIGGER [ MODIFY ] <dev> <type>=<condition>[;<type>...][ OR <dev>...]
     TRACE TRIGGER NOTHING
     TRACE TRIGGER ANYTHING
     TRACE TRIGGER POSITION={ START | CENTER | END }

     TRACE STORE [ MODIFY ] <dev>[ OR <dev>...]
     TRACE STORE [ MODIFY ] <dev> ANYTHING[ OR <dev>...]
???  TRACE STORE [ MODIFY ] <dev> NOTHING[ OR <dev>...]
     TRACE STORE [ MODIFY ] <dev> <state>[;<state>...][ OR <dev>...]
     TRACE STORE NOTHING

     TRACE WINDOW [ MODIFY ] <dev> <type>=<condition>[;<type>...][ OR <dev>...]
     TRACE WINDOW NOTHING

     TRACE EXECUTE
     TRACE HALT
     SHOW TRACE

   Specific commands:

     TRACE TRIGGER CPU INSTRUCTION=<symbol>[;IGNORE=<dont-care-value>[.<dont-care-value>]]
     TRACE TRIGGER CPU TRAP=<trap-number>[;PARAMETER=<parameter>]
     TRACE TRIGGER CPU INTERRUPT=<device-number>
     TRACE TRIGGER MCL ADDRESS=<addr>[;DATA=<data>][;ACTION=<status>]

   Changes (maybe):

     TRACE STATES=<count>
       "Memory" is confusing -- is the value in bytes or states [the latter]?

     TRACE WINDOW <dev>=<condition>[,<dev>=<condition>...]
       Change TRACE_EXEC to a CPU window condition; remove EXEC as a debug
       option.  When TRACE WINDOW CPU-<instruction>[;<mask>] is present at
       instruction prelude, CPU diag setting is removed and saved, to be
       temporarily reinstated when execution condition trips, and permanently
       reinstated in instruction postlude.

       One problem with CPU conditions is specifying a stack instruction pair,
       as they use commas between stackops.  Could either restrict WINDOW
       commands to a single spec, or could require comma to be replaced by a
       dash or a slash, or could check remainder of line; if next glyph is a
       device name, then comma ends spec, else append until the following comma
       or EOL.

       Also use for IOP=<filter> and CPP=<filter>, except that specified
       condition enables storing instead of disables it.  Can specify
       ~<condition> to reverse.  Maybe extend expanded MTAB to handle ranges >
       32 bits.  Could also extend, e.g., to ADCC port or DC bus address.

       Specifying dummy condition "DISABLE" clears all windows.

     TRACE STORE
       Specifying <dev> with no flags sets all on except TRACE_PSERV,
       TRACE_IOBUS, and TRACE_IMBUS.  These latter three are rarely wanted.  Add
       a dummy flag "ALL" that sets all flags for the device.

     Add TRACE_HANDLER function pointer to the DIB.  Format ID values can be
     local, as the call to the formatter will simply be dptr->ctxt->trace.
     Also, the local module can parse or print the window spec when passed the
     stream pointer (NULL if parse else print) and the character pointer (points
     at the string to parse or the prefix to print).

     Be sure that logical device names are being picked up (using find_dev)!


   Trace operational states (acquisition and memory):

     State               Indicator
     ------------------  ---------------------------------------------------
     halted/unallocated  base = NULL
     halted/empty        current = NULL and start = NULL
     running/empty       current /= NULL and start = NULL
     running/filling     current /= NULL and start /= NULL
     halted/filled       current = NULL and start /= NULL and listed = FALSE
     halted/listed       current = NULL and start /= NULL and listed = TRUE

   State transitions:

     Simulator startup:
       -> halted/unallocated

     TRACE MEMORY=<count>
       if count = 0 then -> halted/unallocated
       else -> halted/empty

     TRACE EXECUTE
       if halted/unallocated then error
       else -> running/empty

     trace state
       if running/empty then -> running/filling
       if running/filling then store state
       else ignore state

     trace last state after trigger
       -> halted/filled

     TRACE HALT
       if halted/unallocated then error
       elsif running/empty then -> halted/empty
       elsif running/filling then -> halted/filled
       else no state change

     TRACE LIST
       if halted/unallocated then error
       elsif running/empty then -> halted/empty
       elsif running/filling then -> halted/filled
       if halted/filled then -> halted/listed

     trace exit
       if running/filling or halted/filled then ask for confirmation to discard
       else exit


     State               Indicator
     ------------------  -----------------
     halted/unallocated  Trace_Unallocated
     halted/empty        Trace_Empty
     running/empty       Trace_Empty
     running/filling     Trace_Filling
     halted/filled       Trace_Complete
     halted/not-empty    Trace_Halted

*/


#include <ctype.h>
#include <stdint.h>

#include "hp3000_defs.h"
#include "hp3000_cpu.h"
#include "hp3000_mem.h"
#include "hp3000_io.h"      // DIB
#include "hp3000_imb.h"
#include "hp3000_trace.h"


// temp
#if 0
static const char *status_name [] = {
    "Unallocated",      // no mem
    "Empty",            // mem, no states
    "Filling",          // mem, filling, triggered
    "Complete",         // mem, full, complete
    "Halted",           // mem, partial, stopped
    "Listed"            // mem, complete, printed
    };
#endif


// defs

typedef enum {
    Start_Of_Trace,
    Center_Of_Trace,
    End_Of_Trace
    } TRIGGER_POSITION;

typedef enum {
    Before_Trigger,
    At_Trigger,
    After_Trigger
    } TRIGGER_LOCATION;


// local command handlers

static t_stat memory_cmd      (int32 flag, char *cptr);
static t_stat trigger_cmd     (int32 flag, char *cptr);
static t_stat uncondition_cmd (int32 flag, char *cptr);
static t_stat position_cmd    (int32 flag, char *cptr);
static t_stat store_cmd       (int32 flag, char *cptr);
static t_stat window_cmd      (int32 flag, char *cptr);
static t_stat execute_cmd     (int32 flag, char *cptr);
static t_stat halt_cmd        (int32 flag, char *cptr);
static t_stat list_cmd        (int32 flag, char *cptr);


// local utility

static void initialize_trace (void);
static char *split_command   (char *cptr);


// can't be const because find_ctab doesn't declare

static CTAB trace_table [] = {
    { "MEMORY",  memory_cmd,  0 },
    { "TRIGGER", trigger_cmd, 0 },
    { "STORE",   store_cmd,   0 },
    { "WINDOW",  window_cmd,  0 },
    { "EXECUTE", execute_cmd, 0 },
    { "HALT",    halt_cmd,    0 },
    { "LIST",    list_cmd,    0 },
    { NULL,      NULL,        0 }
    };

static CTAB trigger_table [] = {
    { "NOTHING",  uncondition_cmd, Trigger_On_Nothing  },
    { "ANYTHING", uncondition_cmd, Trigger_On_Anything },
    { "POSITION", position_cmd,    0                   },
    { "MODIFY",   NULL,            0                   },
    { NULL,       NULL,            0                   }
    };

static CTAB position_table [] = {
    { "START",  NULL, Start_Of_Trace  },
    { "CENTER", NULL, Center_Of_Trace },
    { "END",    NULL, End_Of_Trace    },
    { NULL,     NULL, 0               }
    };

// local state

static size_t device_size = 0;                  /* maximum device name size */
static size_t flag_size   = 0;                  /* maximum debug flag name size */

static TRIGGER_LOCATION trigger_location;
static TRIGGER_POSITION trigger_position  = Start_Of_Trace;


// global

TRACE_DESCRIPTOR trace = {
    Trigger_On_Anything,    // trace modes
    Trace_Unallocated,      // trace state
    NULL,                   // trace list file
    NULL,                   // trace list filename
    (time_t) -1,            // trace acquisition time
    0,                      // count of entries
    NULL,                   // trace entry array start pointer
    NULL,                   // trace entry array end pointer
    NULL,                   // trane entry current state pointer
    NULL,                   // trace entry trigger state pointer
    NULL,                   // trace entry start state pointer
    NULL,                   // trace entry end state pointer
    FALSE,                  // trace buffer wraparound flag
    FALSE                   // trace buffer has been listed
    };


// global routines

// process termination
// if trace buffer allocated
//   and trace is active

t_stat trace_shutdown (int32 flag, char *cptr)
{
static char confirm [] = "Exiting will lose trace memory contents; OK to exit [N]?";

if (trace.status == Trace_Unallocated
  || trace.status == Trace_Empty
  || trace.listed)
    return SCPE_EXIT;                                   /* then it is OK to exit */

else if (get_yn (confirm, FALSE) == FALSE)                  /*     and the user denies confirmation */
    return SCPE_INCOMP;                                 /*       then abort the command */

else
    return SCPE_EXIT;
}


// process entry condition

TRACE_RECORD *trace_enter (void)
{
register TRACE_RECORD *tptr;

if (trace.current == trace.limit) {             // if pointing at the last entry
    trace.current = trace.base;                 //   then point at the first entry

    if (not trace.wrapped)
        if (trace.status == Trace_Empty)
            trace.status = Trace_Filling;
        else
            trace.wrapped = TRUE;                       //     and set the wraparound flag
    }

else                                            // otherwise
    trace.current++;                            //   point at the next entry

if (trace.wrapped                               // if wraparound has occurred
  && trace.current->xptr != NULL)               //   and the block pointer is in use
    free (trace.current->xptr);                 //     then free its target

tptr = trace.current;                           // set the temporary

if (trace.current == trace.end) {               // if this is state completes the list
    trace.current = NULL;                       //   then stop tracing
    trace.modes &= ~Trace_Active;
    trace.status = Trace_Complete;

    trace.acq_time = time (NULL);

    cprintf ("Trace complete\n");
    }

return tptr;
}


// process trigger condition

TRACE_RECORD *trace_trigger (void)
{
trace.trigger = trace_enter ();
trace.modes &= ~Trigger_Waiting;

if (trigger_position == Start_Of_Trace) {       // if the trigger position is the start of the trace
    trace.start = trace.trigger;                //   then the list starts at the trigger point

    if (trace.start == trace.base)              // if the list starts at the buffer start
        trace.end = trace.limit;                //   then the list ends at the buffer end
    else                                        // otherwise
        trace.end = trace.trigger - 1;          //   the list ends at the prior entry after wraparound

    cprintf ("Trace in process\n");
    }

else if (trigger_position == Center_Of_Trace) {     // otherwise if the position is the center
    if (not trace.wrapped)
        trace.start = trace.base;

    else if (trace.trigger > trace.base + trace.count / 2)    // if the current position is more than halfway
        trace.start = trace.trigger - trace.count / 2;        //   then the list starts halfway back
    else                                                // otherwise
        trace.start = trace.trigger + trace.count / 2;        //   the list starts halfway ahead

    if (trace.start == trace.base)              // if the list starts at the buffer start
        trace.end = trace.limit;                //   then the list ends at the buffer end
    else                                        // otherwise
        trace.end = trace.start - 1;            //   the list ends at the prior entry after wraparound

    cprintf ("Trace in process\n");
    }

else {                                          // otherwise the trigger position is the end of the trace
    trace.end = trace.trigger;                  //   so the list ends with the current position

    if (trace.end == trace.limit                // if the list ends at the buffer end
      || not trace.wrapped)                     //   or the trace has not wrapped around yet
        trace.start = trace.base;               //   then the list starts at the buffer start
    else                                        // otherwise
        trace.start = trace.end + 1;            //   the list starts at the next entry after wraparound

    trace.current = NULL;                       // all states are already in the buffer
    trace.modes &= ~Trace_Active;
    trace.status = Trace_Complete;

    trace.acq_time = time (NULL);

    cprintf ("Trace complete\n");
    }

return trace.trigger;
}


// process trace command
//
//     TRACE MEMORY=<count>
//     TRACE TRIGGER ANYTHING
//     TRACE TRIGGER NOTHING
//     TRACE TRIGGER CPU INSTRUCTION=<symbol>[;IGNORE=<value>]
//     TRACE TRIGGER CPU ADDRESS=<addr>[;DATA=<data>][;STATUS=<status>]
//     TRACE TRIGGER CPU TRAP=<trap-number>[;PARAM=<parameter>]
//     TRACE TRIGGER CPU IRQ=<device-name>
//     TRACE TRIGGER CPU POSITION={ START | CENTER | END }
//     TRACE STORE <dev>=<debug-flag>[;<debug-flag>...][,<dev>=<debug-flag>...]
//     TRACE WINDOW <dev>=<condition>
//     TRACE WINDOW DISABLE
//     TRACE EXECUTE
//     TRACE HALT
//     TRACE [-N] LIST[=<filename>]

t_stat trace_cmd (int32 flag, char *cptr)
{
char   gbuf [CBUFSIZE];
CTAB   *cmdptr;
t_stat status;

if (cptr == NULL || *cptr == '\0')                      /* if nothing follows */
    status = SCPE_2FARG;

else {
    cptr = get_sim_sw (cptr);

    if (cptr == NULL)
        status = SCPE_INVSW;

    else if (*cptr == '\0')                      /* if nothing follows */
        status = SCPE_2FARG;

    else {
        cptr = get_glyph (cptr, gbuf, '=');                 /* get the command in upper case */

        cmdptr = find_ctab (trace_table, gbuf);

        if (cmdptr == NULL)
            status = SCPE_NXPAR;

        else {

    //printf ("Entry status is %s\n", status_name [trace.status]);    // test

            status = cmdptr->action (cmdptr->arg, cptr);

    //printf ("Exit status is %s\n", status_name [trace.status]);     // test
            }
        }
    }

return status;
}


// SHOW TRACE
/* Show the instruction execution trace criteria.

   This display routine is called to show the criteria that select instruction
   execution tracing.  The "st" parameter is the open output stream.  The other
   parameters are not used.

   This routine services an extended modifier entry, so it must add the trailing
   newline to the output before returning.
*/

t_stat trace_show (FILE *stream, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
{
static const char *position_names [] = {
    "start of trace",
    "center of trace",
    "end of trace"
    };
char   *prefix;
uint32 dev;
t_stat status;
DEBTAB *tptr;
DIB    *dibptr;
t_bool debug, first;

fputs ("\nTrace Specification\n\n", stream);


// MEMORY

fputs ("   MEMORY\n      ", stream);

if (trace.status == Trace_Unallocated)
    fputs ("is not allocated\n\n", stream);

else if (trace.count % 1000000 == 0)
    fprintf (stream, "is %uM states\n\n",
             trace.count / 1000000);

else if (trace.count % 1000 == 0)
    fprintf (stream, "is %uK states\n\n",
             trace.count / 1000);

else
    fprintf (stream, "is %u states\n\n",
             trace.count);


// TRIGGER

fputs ("   TRIGGER\n      ", stream);

if (trace.modes & Trigger_On_Anything)
    fputs ("on anything\n", stream);

else {
    first = TRUE;

    for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
        dptr = sim_devices [dev];                           /*   and get a pointer to the DEVICE structure */
        dibptr = (DIB *) dptr->ctxt;

        if (dibptr != NULL && dibptr->trace != NULL) {
            if (first)
                prefix = "on";
            else
                prefix = "        or";

            status = dibptr->trace (Trace_Show_Trigger, dibptr->card_index, stream, prefix);  // valid?

            if (status == SCPE_OK)      // if trigger was printed
                first = FALSE;
            }
        }

    if (first)                          // if no triggers were printed
        fputs ("on nothing\n", stream);
    }

fprintf (stream, "      position is %s\n\n",
         position_names [trigger_position]);


// STORE

fputs ("   STORE\n      ", stream);

if (not (trace.modes & Store_Active))
    fputs ("nothing\n\n", stream);

else {
    first = TRUE;

    for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
        dptr = sim_devices [dev];                           /*   and get a pointer to the DEVICE structure */

        if (dptr->dctrl != 0) {
            if (first)
                prefix = "on";
            else
                prefix = "        or";

            fprintf (stream, "%s %s state is ", prefix, sim_dname (dptr));

            debug = FALSE;

            for (tptr = dptr->debflags; tptr->name != NULL; tptr++)
                if (dptr->dctrl & tptr->mask) {
                    if (debug)
                        fputc (';', stream);
                    else
                        debug = TRUE;

                    fputs (tptr->name, stream);
                    }

            fputc ('\n', stream);   // must be at least one
            first = FALSE;
            }
        }

    fputc ('\n', stream);   // must be at least one
    }


// WINDOW

fputs ("   WINDOW\n      ", stream);

first = TRUE;

for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
    dibptr = (DIB *) sim_devices [dev]->ctxt;

    if (dibptr != NULL && dibptr->trace != NULL) {
        if (first)
            prefix = "enabled on";
        else
            prefix = "        or";

        status = dibptr->trace (Trace_Show_Window, dibptr->card_index, stream, prefix);  // valid?

        if (status == SCPE_OK)      // if window was printed
            first = FALSE;
        }
    }

if (first)
    fputs ("disabled\n", stream);

fputc ('\n', stream);


// LIST

fputs ("   LIST\n      ", stream);

if (trace.listname == NULL)
    fputs ("nothing\n", stream);
else
    fprintf (stream, "on %s\n", trace.listname);

fputs ("      ", stream);

if (trace.modes & Trace_Active)
    if (trace.modes & Trigger_Waiting)
        fputs ("is waiting for trigger\n\n", stream);
    else if (trace.status == Trace_Filling)
        fputs ("trace is in process\n\n", stream);
    else
        fputs ("is empty\n\n", stream);

else if (trace.status == Trace_Complete)
    fputs ("trace is complete\n\n", stream);

else if (trace.status == Trace_Halted)
    fputs ("trace is halted\n\n", stream);

else
    fputs ("is empty\n\n", stream);


return SCPE_OK;
}


// trace parser

t_stat trace_parse (char *cptr, TRACE_PARSER_TABLE *ptab, void *vptr)
{
char   *nptr, gbuf [CBUFSIZE];
t_stat status = SCPE_OK;

if (*cptr == '\0')     // nothing follows device name?
    status = SCPE_2FARG;

else while (cptr != NULL && status == SCPE_OK) {    // set trigger(s)
    nptr = strchr (cptr, ';');  // find the end of the current option

    if (nptr != NULL)               // if found
        *nptr++ = '\0';             // terminate at the next option

    cptr = get_glyph (cptr, gbuf, '=');         /* get the next option keyword */

    while (ptab->name != NULL)
        if (strncmp (gbuf, ptab->name, strlen (gbuf)) == 0)
            break;
        else
            ptab++;

    if (ptab->name == NULL) {
        status = SCPE_ARG;
        break;
        }

    else
        status = ptab->parser (cptr, ptab->id, vptr);

    cptr = nptr;    // point at next
    }   // end while

return status;
}


// trace store

t_stat trace_store (DEVICE *dptr, char *cptr, uint32 common_states, uint32 all_states)
{
DEBTAB *tptr;
char   gbuf [CBUFSIZE];
uint32 control;
t_stat status = SCPE_OK;

if (*cptr == '\0')
    dptr->dctrl = common_states;

else {
    get_glyph (cptr, gbuf, ';');

    if (strncmp (gbuf, "ANYTHING", strlen (gbuf)) == 0)
        dptr->dctrl = all_states;

    else {
        control = 0;

        do {
            for (tptr = dptr->debflags; tptr->name != NULL; tptr++)
                if (strcmp (tptr->name, gbuf) == 0) {             /* match? */
                    control |= tptr->mask;
                    break;
                    }

            if (tptr->name == NULL)                      /* no match? */
                status = SCPE_ARG;
            else
                cptr = get_glyph (cptr, gbuf, ';');                 /* get debug flag */
            }
        while (gbuf [0] != '\0' && status == SCPE_OK);

        if (status == SCPE_OK)
            dptr->dctrl = control;
        }
    }

return status;
}


// trace formatter

#define FLAG_SIZE           32                          /* sufficiently large to accommodate all flag names */
#define FORMAT_SIZE         1024                        /* sufficiently large to accommodate all format strings */

void trace_format (DEVICE *dptr, uint32 flag, const char *format, ...)
{
static const char *location_names [] = { "--", ">>", "++" };
const char *nptr;
va_list    argptr;
DEBTAB     *debptr;
char       *fptr;
char       flag_name [FLAG_SIZE];               /* desired size is [sys_flag_size + 1] */
char       header_fmt [FORMAT_SIZE];            /* desired size is [device_size + flag_size + format_size + 6] */

if (trace.list != NULL && dptr != NULL) {               /* if the output stream and device pointer are valid */
    debptr = dptr->debflags;                            /*   then get a pointer to the trace flags table */

    if (debptr != NULL)                                 /* if the trace table exists */
        while (debptr->name != NULL)                    /*   then search it for an entry with the supplied flag */
            if (debptr->mask & flag) {                  /* if the flag matches this entry */
                nptr = debptr->name;                    /*   then get a pointer to the flag name */
                fptr = flag_name;                       /*     and the buffer */

                do
                    *fptr++ = (char) tolower (*nptr);   /* copy and downshift the flag name */
                while (*nptr++ != '\0');

                sprintf (header_fmt, "%s%-*s %*s: ",            /* format the prefix and store it */
                         location_names [trigger_location],     /*  relative to the trigger position */
                         (int) device_size, sim_dname (dptr),   /*   while padding the device and flag names */
                         (int) flag_size, flag_name);           /*     as needed for proper alignment */

                strcat (header_fmt, format);            /* append the supplied format */

                if (trace.list == stdout) {                         /* if debug output is to the (raw) console */
                    fptr = header_fmt + strlen (header_fmt) - 1;    /*   then find the end of the string */

                    if (*fptr == '\n') {                /* if the format ends with a LF */
                        *fptr++ = '\r';                 /*   then replace it */
                        *fptr++ = '\n';                 /*     with a CR LF sequence */
                        *fptr   = '\0';                 /*       and terminate with a NUL */
                        }
                    }

                va_start (argptr, format);                  /* set up the argument list */
                vfprintf (trace.list, header_fmt, argptr);  /*   and format and print to the debug stream */

                va_end (argptr);                        /* clean up the argument list */
                break;                                  /*   and exit with the job complete */
                }

            else                                        /* otherwise */
                debptr++;                               /*   look at the next trace table entry */
    }

return;
}



// locals


//  - TRACE MEMORY=<count>
// note: any trailing "K" or "M" may be in lower case

static t_stat memory_cmd (int32 flag, char *cptr)
{
char   *tptr;
size_t max_count, multiplier;

if (trace.modes & Trace_Active)        // if still tracing
    return SCPE_NOFNC;              // then not allowed

else if (cptr == NULL || *cptr == '\0')                 /* if the buffer size is missing */
    return SCPE_MISVAL;                                 /*   then report the error */

else
    trace.count = (size_t) strtotv (cptr, &tptr, 10);   /* parse the new buffer size */

if (cptr == tptr)                                       /* if parsing failed */
    return SCPE_ARG;                                    /*   then report it */

else if (*tptr == '\0') {
    multiplier = 1;
    max_count = SIZE_MAX;
    }

else if (toupper (*tptr) == 'K') {
    multiplier = 1000;
    max_count = SIZE_MAX / multiplier;
    }

else if (toupper (*tptr) == 'M') {
    multiplier = 1000000;
    max_count = SIZE_MAX / multiplier;
    }

else
    return SCPE_ARG;

if (trace.count > max_count)
    return SCPE_ARG;

else if (multiplier > 1 && tptr [1] != '\0')
    return SCPE_2MARG;

else if (trace.status != Trace_Unallocated) {      // deallocate request?
    initialize_trace ();                                // free parameters

    free (trace.base);
    trace.base = trace.limit = NULL;
    trace.status = Trace_Unallocated;
    }

if (trace.count > 0) {
    trace.count = trace.count * multiplier;
    trace.base = malloc (trace.count * sizeof (TRACE_RECORD));       //   tps points at first element

    if (trace.base == NULL)                            // if out of memory
        return SCPE_MEM;               //   then the command fails

    else {
        trace.limit = trace.base + trace.count - 1;    //   and tpe points at last element
        trace.status = Trace_Empty;
        }
    }

trace.current = NULL;                   // no current entry
trace.trigger = NULL;                   //   the trigger has not occurred
trace.start   = NULL;                   //   and the trace list has not started
trace.end     = NULL;                   //   and the trace list has not ended

trace.wrapped = FALSE;                                 // wraparound has not occurred
trace.listed = FALSE;

return SCPE_OK;
}


//     TRACE TRIGGER [ MODIFY ] <dev> <type>=<condition>[;<type>...][ OR <dev>...]
//     TRACE TRIGGER NOTHING
//     TRACE TRIGGER ANYTHING
//     TRACE TRIGGER POSITION={ START | CENTER | END }
//
// entry with cptr pointing after TRIGGER

static t_stat trigger_cmd (int32 flag, char *cptr)
{
char   *optr, *nptr, gbuf [CBUFSIZE];
CTAB   *cmdptr;
DEVICE *dptr;
DIB    *dibptr;
uint32 dev;
t_stat status = SCPE_OK;

if (trace.modes & Trace_Active)        // if still tracing
    status = SCPE_NOFNC;              // then not allowed

else if (*cptr == '\0')                      /* if nothing follows TRIGGER */
    status = SCPE_2FARG;

else {
    nptr = get_glyph (cptr, gbuf, '=');

    cmdptr = find_ctab (trigger_table, gbuf);

    if (cmdptr != NULL && cmdptr->action != NULL)        // valid and action defined?
        status = cmdptr->action (cmdptr->arg, nptr);

    else {
        if (cmdptr != NULL)     // is MODIFY so get device
            if (*nptr == '\0')      // nothing follows
                status = SCPE_2FARG;
            else
                cptr = nptr;                                        //   skip clear and point at device

        else                    // not MODIFY, clear all triggers
            for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
                dibptr = (DIB *) sim_devices [dev]->ctxt;

                if (dibptr != NULL && dibptr->trace != NULL)
                    dibptr->trace (Trace_Set_Trigger, dibptr->card_index, NULL, NULL);  // clear trigger
                }

        while (status == SCPE_OK && cptr != NULL) {
            optr = split_command (cptr);    // split the command at the "OR"

            cptr = get_glyph (cptr, gbuf, '\0');    // get next dev

            if (*cptr == '\0')                              /* if nothing follows the device name */
                status = SCPE_2FARG;                        /*   then report a missing argument */

            else {
                dptr = find_dev (gbuf);

                if (dptr == NULL || dptr->flags & DEV_DIS)           // not found or disabled
                    status = SCPE_NXDEV;

                else {
                    dibptr = (DIB *) dptr->ctxt;

                    if (dibptr == NULL || dibptr->trace == NULL)
                        status = SCPE_NOFNC;

    // must either find " OR " (case insensitive!), so next string start location is known
    //   or return updated cptr from device trace parser (so can get_glyph and compare to "OR")!
    //   BUT parse_sym does no return the updated cptr!
    //
    // cptr -> <type>=<condition>[;<subtype>=<condition>...][ OR <dev>...]

                    else {
                        status = dibptr->trace (Trace_Set_Trigger, dibptr->card_index, NULL, cptr);

                        if (status == SCPE_OK) {
                            trace.modes &= ~Trigger_On_Anything;
                            cptr = optr;
                            }
                        }
                    }
                }
            }
        }
    }

return status;
}


//  - TRACE TRIGGER ANYTHING
//  - TRACE TRIGGER NOTHING

static t_stat uncondition_cmd (int32 flag, char *cptr)
{
uint32 dev;
DIB    *dibptr;

// clear existing triggers

for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
    dibptr = (DIB *) sim_devices [dev]->ctxt;

    if (dibptr != NULL && dibptr->trace != NULL)
        dibptr->trace (Trace_Set_Trigger, dibptr->card_index, NULL, NULL);  // clear trigger
    }

trace.modes = trace.modes & ~Trigger_On_Anything | (TRACE_MODES) flag;  /* set the new mode */

if (flag == Trigger_On_Nothing)
    trigger_position = End_Of_Trace;
else                                             // On_Anything
    trigger_position = Start_Of_Trace;

return SCPE_OK;
}


//  - TRACE TRIGGER POSITION={ START | CENTER | END }

static t_stat position_cmd (int32 flag, char *cptr)
{
CTAB *cmdptr;
char gbuf [CBUFSIZE];

if (cptr == NULL || *cptr == '\0')                      /* if nothing follows */
    return SCPE_2FARG;

else {
    get_glyph (cptr, gbuf, 0);     // get 1st value into buf

    cmdptr = find_ctab (position_table, gbuf);

    if (cmdptr == NULL)
        return SCPE_NXPAR;

    else
        trigger_position = (TRIGGER_POSITION) cmdptr->arg;
    }

return SCPE_OK;
}


//     TRACE STORE [ MODIFY ] <dev>[ OR <dev>...]
//     TRACE STORE [ MODIFY ] <dev> ANYTHING[ OR <dev>...]
//     TRACE STORE [ MODIFY ] <dev> NOTHING[ OR <dev>...]
//     TRACE STORE [ MODIFY ] <dev> <state>[;<state>...][ OR <dev>...]
//     TRACE STORE NOTHING

typedef enum {
    Option_Modify,
    Option_Nothing,
    Option_Anything
    } TRACE_OPTION;

static CTAB option_table [] = {
    { "NOTHING",  NULL, Option_Nothing },
    { "MODIFY",   NULL, Option_Modify  },
    { NULL,       NULL, 0              }
    };

static t_stat store_cmd (int32 flag, char *cptr)
{
char   *optr, *nptr, gbuf [CBUFSIZE];
CTAB   *cmdptr;
DEVICE *dptr;
DIB    *dibptr;
DEBTAB *tptr;
uint32 dev;
size_t device_length, flag_length;
t_stat status = SCPE_OK;

if (trace.modes & Trace_Active)        // if still tracing
    status = SCPE_NOFNC;              // then not allowed

else if (*cptr == '\0')                      /* if nothing follows */
    status = SCPE_2FARG;

else {
    nptr = get_glyph (cptr, gbuf, '\0');

    cmdptr = find_ctab (option_table, gbuf);

    if (cmdptr != NULL && cmdptr->arg == Option_Modify)    // if MODIFY
        if (*nptr == '\0')      // nothing follows
            return SCPE_2FARG;
        else
            cptr = nptr;                                        //   skip clear and point at device

    else {
        for (dev = 0; sim_devices [dev] != NULL; dev++) /* loop through the devices */
            sim_devices [dev]->dctrl = 0;               /*   and clear all current trace flags */

        trace.modes &= ~Store_Active;

        if (cmdptr != NULL && cmdptr->arg == Option_Nothing)  // is NOTHING
            if (*nptr == '\0')
                return SCPE_OK;
            else
                return SCPE_2MARG;
        }

    initialize_trace ();

    while (status == SCPE_OK && cptr != NULL) {
        optr = split_command (cptr);    // split the command at the "OR"

        cptr = get_glyph (cptr, gbuf, '\0');    // get next dev

        dptr = find_dev (gbuf);

        if (dptr == NULL || dptr->flags & DEV_DIS)           // not found or disabled
            status = SCPE_NXDEV;

        else if (not (dptr->flags & DEV_DEBUG)     // no debug
          || dptr->debflags == NULL)               // no table
            status = SCPE_NOFNC;

    //  - TRACE STORE <dev>[ OR <dev>...]
    //  - TRACE STORE <dev> <debug-flag>[;<debug-flag>...][ OR <dev>...]
    //  - TRACE STORE <dev> ANYTHING[ OR <dev>...]

        else {    // <dev>
            dibptr = (DIB *) dptr->ctxt;

            if (dibptr == NULL || dibptr->trace == NULL)
                status = SCPE_NOFNC;
            else
                status = dibptr->trace (Trace_Set_Store, dibptr->card_index, NULL, cptr);

            if (status == SCPE_OK) {
                trace.modes |= Store_Active;
                cptr = optr;
                }
            }
        }

    if (status == SCPE_OK) {     // accum state name sizes
        device_size = 0;
        flag_size   = 0;

        for (dev = 0; sim_devices [dev] != NULL; dev++) { /* loop through the devices */
            dptr = sim_devices [dev];                       /*   and get the next device pointer */

            if (not (dptr->flags & DEV_DIS)           // not disabled
              && dptr->debflags != NULL                 // debug table exists
              && dptr->dctrl != 0) {                    // one or more flags set
                device_length = strlen (sim_dname (dptr));      /*   then get the length of the device name */

                if (device_length > device_size)                /* if it's greater than the current maximum */
                    device_size = device_length;                /*   then reset the size */

                for (tptr = dptr->debflags; tptr->name != NULL; tptr++)
                    if (dptr->dctrl & tptr->mask) {             /* if the trace state is active */
                        flag_length = strlen (tptr->name);  /*     then get the flag name length */

                        if (flag_length > flag_size)        /* if it's greater than the current maximum */
                            flag_size = flag_length;        /*   then reset the size */
                        }
                }
            }
        }
    }

return status;
}


/* Change the instruction execution trace criteria.

   This validation routine is called to configure the criteria that select
   instruction execution tracing.  The routine processes commands of the form:

     SET CPU EXEC=<match>[;<mask>]
     SET CPU NOEXEC

   If the <mask> value is not supplied, a mask of 177777 octal is used.  The
   values are entered in the current CPU data radix, which defaults to octal,
   unless an override switch is present on the command line.

   On entry, the "option" parameter is 0 to clear and 1 to set the criteria, and
   "cptr" points to the first character of the match value to be set.  The unit
   and description pointers are not used.

//

     TRACE WINDOW <dev>=<condition>
     TRACE WINDOW DISABLE

       Change TRACE_EXEC to a CPU window condition; remove EXEC as a debug
       option.  When TRACE WINDOW CPU-<instruction>[;<mask>] is present at
       instruction prelude, CPU diag setting is removed and saved, to be
       temporarily reinstated when execution condition trips, and permanently
       reinstated in instruction postlude.

       Also use for IOP=<filter> and CPP=<filter>, except that specified
       condition enables storing instead of disables it.  Can specify
       ~<condition> to reverse.  Maybe extend expanded MTAB to handle ranges >
       32 bits.  Could also extend, e.g., to ADCC port or DC bus address.


   Implementation notes:

//  1. Parsing separate device conditions in the same window command presents a
       problem -- the <condition> may include a space and one or more commas,
       which are also used to delimit the device conditions.  Examples are:
       "CPU=FDIV,NOP,IMB=1,2" and "CPU=SETR SBANK,DB,DL,Z,STATUS,X,Q,S,IMB=1,2".

       There are three possible solutions.  First, restrict the window command
       to a single condition.  Second, use a different delimiter, such as a
       colon, for commands (the space, comma, semicolon, and dash are used
       within the condition).  Third, retain the comma and determine the end of
       the condition by searching for the <dev> that follows.  Option 2 would be
       difficult for the user to remember, as it would be special for this
       command only.  Option 3 is possible -- search for the next equals sign
       and backtrack to the preceding comma; the condition either ends there or
       at the end of the line if no equals sign is found.  But the plethora of
       commas would be confusing for the user (if not the parser).  Option 1
       requires a separate DISABLE parameter, and the user must specify
       multiple WINDOW command if needed.   Still, it is notably simpler, so we
       implement option 1.

*/

static t_stat window_cmd (int32 flag, char *cptr)
{
char   gbuf [CBUFSIZE];
DEVICE *dptr;
DIB    *dibptr;
uint32 dev;
t_stat status = SCPE_OK;

if (trace.modes & Trace_Active)        // if still tracing
    status = SCPE_NOFNC;              // then not allowed

else if (*cptr == '\0')                      /* if nothing follows */
    status = SCPE_2FARG;

else {
    cptr = get_glyph (cptr, gbuf, '='); // parse <dev> or keyword

    if (strncmp (gbuf, "DISABLE", 7) == 0) {
        for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
            dibptr = (DIB *) sim_devices [dev]->ctxt;

            if (dibptr != NULL && dibptr->trace != NULL)
                dibptr->trace (Trace_Set_Window, dibptr->card_index, NULL, NULL);  // clear window
            }

//        trace.modes &= ~Window_Active;
        }

    else if (*cptr == '\0')      // no param
        status = SCPE_MISVAL;

    else {
        dptr = find_dev (gbuf);

        if (dptr == NULL)
            status = SCPE_NXDEV;

        else {
            dibptr = (DIB *) dptr->ctxt;

            if (dibptr == NULL
              || dibptr->trace == NULL)
                status = SCPE_NOFNC;

            else {
                status = dibptr->trace (Trace_Set_Window, dibptr->card_index, NULL, cptr);

//                if (status == SCPE_OK)
//                    trace.modes |= Window_Active;
                }
            }
        }
    }

return status;
}


//  - TRACE EXECUTE

static t_stat execute_cmd (int32 flag, char *cptr)
{
if (trace.status == Trace_Unallocated   // if the trace buffer has not been allocated
  || trace.modes & Trace_Active         // or still tracing
  || not (trace.modes & Store_Active))  // or no store states
    return SCPE_NOFNC;                  // then not allowed

initialize_trace ();

if (trace.modes & Trigger_On_Anything) {     // if triggering immediately
    trace.trigger = trace.base;             //   then the trigger has occurred
    trace.start   = trace.base;             //   and the trace list starts at the buffer start
    trace.end     = trace.limit;            //   and ends at the buffer end

    cprintf ("Trace in process\n");
    }

else {
    trace.modes |= Trigger_Waiting;

    cprintf ("Trace executing\n");
    }

trace.modes |= Trace_Active;
trace.current = trace.limit;                        // points at last element

return SCPE_OK;
}


//  - TRACE HALT

static t_stat halt_cmd (int32 flag, char *cptr)
{
if (trace.status == Trace_Unallocated)      // if the trace buffer has not been allocated
    return SCPE_NOFNC;                      //   then the command fails

else if (trace.modes & Trace_Active) {        // if still tracing then halt
    trace.modes &= ~Trace_Active;

    trace.acq_time = time (NULL);

    if (trace.status == Trace_Filling)      // if filling
        trace.status = Trace_Halted;        // then mark the halt

    if (trace.wrapped)                          // if wraparound has occurred
        if (trace.current == trace.limit)       //   then if pointing at the last entry
            trace.start = trace.base;           //     then the list starts with the first entry
        else                                    //   otherwise
            trace.start = trace.current + 1;    //      the list starts with the next entry
    else                                        // otherwise with no wraparound
        trace.start = trace.base;               //   the list starts with the first entry

    trace.end = trace.current;                  // the list ends with the current pointer
    trace.current = NULL;                       //   and tracing stops

    if (trace.modes & Trigger_Waiting) {        // if not yet triggered
        trace.modes &= ~Trigger_Waiting;
        trace.trigger = trace.end;              //   then the trigger is the last state
        }

    cprintf ("Trace halted\n");
    }

return SCPE_OK;
}


//  - TRACE LIST
//  - TRACE LIST NOTHING
//  - TRACE [ -N ] LIST
//  - TRACE [ -N ] LIST=<filename>

static t_stat list_cmd (int32 flag, char *cptr)
{
struct tm    *acquisition_time;
TRACE_RECORD *entry;
DIB          *dibptr;
char         date_stamp [81], separator [81];
uint32       dev;
size_t       date_stamp_size;
t_stat       status;

if (cptr != NULL && *cptr != '\0') {
    if (trace.list != NULL) {
        fclose (trace.list);
        free (trace.listname);

        trace.list     = NULL;
        trace.listname = NULL;
        }

    if (strcasecmp (cptr, "NOTHING") == 0)
        return SCPE_OK;

    else if (sim_switches & SWMASK ('N'))                    /* if a new list file is requested */
        trace.list = fopen (cptr, "w");                 /*   then open an empty file for writing */

    else                                                /* otherwise */
        trace.list = fopen (cptr, "a");                 /*   open an existing file for appending */

    if (trace.list == NULL)                             /* if the open failed */
        return SCPE_OPENERR;                            /*   then report the error */

    trace.listname = malloc (strlen (cptr) + 1);

    if (trace.listname == NULL) {
        fclose (trace.list);
        return SCPE_MEM;
        }

    else
        strcpy (trace.listname, cptr);

    if (trace.status == Trace_Unallocated)              /* if the trace buffer has not been allocated */
        return SCPE_OK;                                 /*   then return quietly with the file created */
    }

else if (sim_switches & SWMASK ('N')) {                 /* if a new list file is requested */
    if (trace.list == NULL)                             /*   but the list file is not defined */
        return SCPE_NOFNC;                              /*     then the command is not allowed */
    else                                                /*   otherwise */
        trace.list = freopen (trace.listname, "w",      /*     reopen the file to clear the contents */
                              trace.list);

    if (trace.status == Trace_Unallocated)              /* if the trace buffer has not been allocated */
        return SCPE_OK;                                 /*   then return quietly with the file cleared */
    }

if (trace.status == Trace_Unallocated               /* if the trace buffer has not been allocated */
  || trace.list == NULL)                    // or no log file
    return SCPE_NOFNC;                      //   then the command fails

else if (trace.modes & Trace_Active)             // if still tracing then halt
    halt_cmd (0, "");

if (trace.status == Trace_Empty) {           // never started
    cprintf ("Trace memory is empty\n");
    return SCPE_OK;
    }

// TRACE -N LIST without filename rewinds and writes EOF

memset (separator, '-', sizeof separator - 1);

separator [sizeof separator - 1] = '\0';

acquisition_time = localtime (&trace.acq_time);

if (acquisition_time == NULL)
    date_stamp_size = 0;
else
    date_stamp_size = strftime (date_stamp, sizeof date_stamp,
//                              "%a, %b %d, %Y at %I:%M %p",
                                "%a, %d %b, %Y at %H:%M",
                                acquisition_time);

if (date_stamp_size == 0)
    fprintf (trace.list, "%s\n             %s simulator trace acquisition time unknown\n%s\n",
             separator, sim_name, separator);
else
    fprintf (trace.list, "%s\n         %s simulator trace acquired on %s\n%s\n",
             separator, sim_name, date_stamp, separator);

for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through the devices */
    dibptr = (DIB *) sim_devices [dev]->ctxt;

    if (dibptr != NULL && dibptr->trace != NULL)
        dibptr->trace (Trace_Set_List, dibptr->card_index, NULL, NULL);  // initialize list
    }

trigger_location = Before_Trigger;

entry = trace.start;

while (TRUE) {
    if (entry->flags & (TRACE_DATA | TRACE_FETCH))
        dibptr = (DIB *) mem_dev.ctxt;
    else if (entry->flags & (TRACE_IMBRQ | TRACE_IMBUS))
        dibptr = (DIB *) imb_dev_ctxt;
    else
        dibptr = (DIB *) entry->dptr->ctxt;

    if (entry == trace.trigger)
        trigger_location = At_Trigger;

    if (dibptr == NULL || dibptr->trace == NULL)
        status = SCPE_IERR;

    else
        status = dibptr->trace (Trace_Show_List, dibptr->card_index, trace.list, entry);

    if (status == SCPE_IERR)
        cprintf ("Unhandled %s trace record ID %u\n",
                 sim_dname (entry->dptr), entry->record_id);

    if (entry == trace.end)       // if done
        break;

    else if (entry == trace.limit)             // if pointing at the last entry
        entry = trace.base;                 //   then point at the first entry

    else                                            // otherwise
        entry++;                            //   point at the next entry

    if (trigger_location == At_Trigger)
        trigger_location = After_Trigger;
    }

fflush (trace.list);                                    /* flush the trace list file */

trace.listed = TRUE;

return status;
}


// utility

static void initialize_trace (void)
{
TRACE_RECORD *entry;

if (trace.status == Trace_Unallocated)
    return;

else if (trace.status != Trace_Empty) {                     // populated?
    entry = trace.start;

    while (TRUE) {
        if (entry->xptr != NULL)
            free (entry->xptr);

        if (entry == trace.end)       // if done
            break;

        else if (entry == trace.limit)             // if pointing at the last entry
            entry = trace.base;                 //   then point at the first entry

        else                                            // otherwise
            entry++;                            //   point at the next entry
        }

    trace.status = Trace_Empty;
    }

trace.current = NULL;                   // no current entry
trace.trigger = NULL;                   //   the trigger has not occurred
trace.start   = NULL;                   //   and the trace list has not started
trace.end     = NULL;                   //   and the trace list has not ended

trace.wrapped = FALSE;                                 // wraparound has not occurred
trace.listed = FALSE;

return;
}


// split the "cptr" command line at an " OR " string with a NUL
// return a pointer to tne remainder of the line.
//
// cptr -> <type>=<condition>[;<subtype>=<condition>...][ OR <dev>...]

static char *split_command (char *cptr)
{
char *bptr = cptr;
char *optr = NULL;

while (*bptr != '\0') {
    if (*bptr != ' ')
        bptr++;

    else {
        if ((bptr [1] == 'O' || bptr [1] == 'o')
          && (bptr [2] == 'R' || bptr [2] == 'r')
          && bptr [3] == ' ') {
            *bptr = '\0';
            optr = bptr + 4;

            sim_trim_endspc (cptr);

            while (isspace (*optr))                                 /* absorb spaces */
                optr++;

            break;
            }

        else
            bptr++;
        }
    }

return optr;
}
