/* hp3000_trace.h: HP 3000 system tracing interface definitions

   Copyright (c) 2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   13-Aug-24    JDB     Created


   This module supports memory-buffered tracing....

*/



#include <time.h>



typedef enum {
    Trace_Unallocated,      // no mem
    Trace_Empty,            // mem, no states
    Trace_Filling,          // mem, filling, triggered
    Trace_Complete,         // mem, full, complete
    Trace_Halted            // mem, partial, stopped
    } TRACE_STATUS;

typedef enum {
    Trigger_On_Nothing  = 0,
    Trace_Active        = BIT ( 0),
    Store_Active        = BIT ( 1),
//    Window_Active       = BIT ( 2),
    Trigger_Waiting     = BIT ( 3),
    Trigger_On_Anything = BIT ( 4)
    } TRACE_MODES;

typedef struct {
    TRACE_MODES   modes;     // trace mode flags
    TRACE_STATUS  status;    // trace current status
    FILE          *list;     // trace list file
    char          *listname; // trace list filename
    time_t        acq_time;  // trace acquisition time
    size_t        count;     // count of entries in memory
    TRACE_RECORD  *base;     // trace entry array start pointer
    TRACE_RECORD  *limit;    // trace entry array end pointer
    TRACE_RECORD  *current;  // trane entry current state pointer
    TRACE_RECORD  *trigger;  // trace entry trigger state pointer
    TRACE_RECORD  *start;    // trace entry start state pointer
    TRACE_RECORD  *end;      // trace entry end state pointer
    t_bool        wrapped;   // trace buffer wraparound flag
    t_bool        listed;    // trace buffer has been listed
    } TRACE_DESCRIPTOR;

typedef t_stat TRACE_PARSER
    (char    *cptr,
     uint32  id,
     void    *vptr);

typedef const struct {
    const char    *name;
    const uint32  id;
    TRACE_PARSER  *parser;
    } TRACE_PARSER_TABLE;


#undef TRACING
#undef TRACINGP

#define TRACING(ptr,flg)    (trace.modes & Trace_Active && ((ptr)->dctrl & (flg)))


// constant comparisons are eliminated, even in no-optimization cases (gcc -O0)
// need to include the trigger state even if TRACE_state not defined

// Trace_All_States bypasses only the debug flag check.
// The trigger state is always included, even if the flag check and window
//   check fail.

#define Trace_All_States    TRUE
#define Trace_Set_States    FALSE
#define Trace_No_Trigger    FALSE
#define Trace_No_Window     TRUE

#define trace_internal(n, state_match, trigger_match, window_match, \
                       dp, flg, id, v1, v2, v3, v4, v5, v6, v7, v8, v9, xp) \
    do                                          \
        if (trace.modes & Trace_Active) {       \
            TRACE_RECORD *tpx;                  \
                                                \
            if (trace.modes & Trigger_Waiting   \
              && (trigger_match))               \
                tpx = trace_trigger ();         \
                                                \
            else if (((dp)->dctrl & (flg)       \
              || state_match)                   \
              && (window_match))                \
                tpx = trace_enter ();           \
                                                \
            else                                \
                break;                          \
                                                \
            tpx->record_id = id;                \
            tpx->dptr = dp;                     \
            tpx->flags = flg;                   \
            if (n >= 1)                         \
                tpx->p [0] = v1;                \
            if (n >= 2)                         \
                tpx->p [1] = v2;                \
            if (n >= 3)                         \
                tpx->p [2] = v3;                \
            if (n >= 4)                         \
                tpx->p [3] = v4;                \
            if (n >= 5)                         \
                tpx->p [4] = v5;                \
            if (n >= 6)                         \
                tpx->p [5] = v6;                \
            if (n >= 7)                         \
                tpx->p [6] = v7;                \
            if (n >= 8)                         \
                tpx->p [7] = v8;                \
            if (n >= 9)                         \
                tpx->p [8] = v9;                \
            tpx->xptr = xp;                     \
            }                                   \
                                                \
        else                                    \
            (void) 0;                           \
    while (0)

// standard calls

#define trace_0(dp, flg, id) \
    trace_internal(0, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_1(dp, flg, id, v1) \
    trace_internal(1, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, 0, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_2(dp, flg, id, v1, v2) \
    trace_internal(2, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_3(dp, flg, id, v1, v2, v3) \
    trace_internal(3, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, 0, 0, 0, 0, 0, 0, NULL)

#define trace_4(dp, flg, id, v1, v2, v3, v4) \
    trace_internal(4, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, 0, 0, 0, 0, 0, NULL)

#define trace_5(dp, flg, id, v1, v2, v3, v4, v5) \
    trace_internal(5, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, v5, 0, 0, 0, 0, NULL)

#define trace_6(dp, flg, id, v1, v2, v3, v4, v5, v6) \
    trace_internal(6, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, v5, v6, 0, 0, 0, NULL)

#define trace_7(dp, flg, id, v1, v2, v3, v4, v5, v6, v7) \
    trace_internal(7, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, v5, v5, v7, 0, 0, NULL)

#define trace_8(dp, flg, id, v1, v2, v3, v4, v5, v6, v7, v8) \
    trace_internal(8, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, v5, v6, v7, v8, 0, NULL)


// multiple trace flag calls

#define trace_3M(dp, flg, id, v1, v2, v3) \
    trace_internal(3, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, (dp)->dctrl & (flg), id, v1, v2, v3, 0, 0, 0, 0, 0, 0, NULL)


// any trace flag calls

#define trace_4A(dp, flg, id, v1, v2, v3, v4) \
    trace_internal(4, Trace_All_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, 0, 0, 0, 0, 0, NULL)

#define trace_9A(dp, flg, id, v1, v2, v3, v4, v5, v6, v7, v8, v9) \
    trace_internal(8, Trace_All_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, v5, v6, v7, v8, v9, NULL)


// buffer parameter calls

#define trace_3B(dp, flg, id, v1, v2, v3, bptr) \
    trace_internal(3, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, 0, 0, 0, 0, 0, 0, bptr)

#define trace_4B(dp, flg, id, v1, v2, v3, v4, bptr) \
    trace_internal(4, Trace_Set_States, Trace_No_Trigger, Trace_No_Window, \
        dp, flg, id, v1, v2, v3, v4, 0, 0, 0, 0, 0, bptr)


// trigger parameter calls

#define trace_2T(dp, flg, id, trigger, window, v1, v2) \
    trace_internal(2, Trace_Set_States, \
        (trigger), (window), \
        dp, flg, id, v1, v2, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_3T(dp, flg, id, trigger, window, v1, v2, v3) \
    trace_internal(3, Trace_Set_States, \
        (trigger), (window), \
        dp, flg, id, v1, v2, v3, 0, 0, 0, 0, 0, 0, NULL)

#define trace_4T(dp, flg, id, trigger, window, v1, v2, v3, v4) \
    trace_internal(4, Trace_Set_States, \
        (trigger), (window), \
        dp, flg, id, v1, v2, v3, v4, 0, 0, 0, 0, 0, NULL)


// window parameter calls

#define trace_0W(dp, flg, id, window_test) \
    trace_internal(0, Trace_Set_States, Trace_No_Trigger, window_test, \
        dp, flg, id, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_1W(dp, flg, id, window_test, v1) \
    trace_internal(1, Trace_Set_States, Trace_No_Trigger, window_test, \
        dp, flg, id, v1, 0, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_2W(dp, flg, id, window_test, v1, v2) \
    trace_internal(2, Trace_Set_States, Trace_No_Trigger, window_test, \
        dp, flg, id, v1, v2, 0, 0, 0, 0, 0, 0, 0, NULL)

#define trace_3W(dp, flg, id, window_test, v1, v2, v3) \
    trace_internal(3, Trace_Set_States, Trace_No_Trigger, window_test, \
        dp, flg, id, v1, v2, v3, 0, 0, 0, 0, 0, 0, NULL)

#define trace_5W(dp, flg, id, window_test, v1, v2, v3, v4, v5) \
    trace_internal(5, Trace_Set_States, Trace_No_Trigger, window_test, \
        dp, flg, id, v1, v2, v3, v4, v5, 0, 0, 0, 0, NULL)


// globals

extern TRACE_DESCRIPTOR trace;


// global trace functions

extern t_stat       trace_shutdown (int32 flag, char *cptr);
extern t_stat       trace_cmd      (int32 flag, char *cptr);
extern t_stat       trace_show     (FILE *stream, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr);
extern TRACE_RECORD *trace_enter   (void);
extern TRACE_RECORD *trace_trigger (void);
extern t_stat       trace_parse    (char *cptr, TRACE_PARSER_TABLE *ptab, void *vptr);
extern t_stat       trace_store    (DEVICE *dptr, char *cptr, uint32 common_states, uint32 all_states);
extern void         trace_format   (DEVICE *dptr, uint32 flag, const char *format, ...);
